/*

  fairness.c

  Created by Ivaylo Dobrikov on 1/9/14.

  The model comprises functions for checking LTL fairness.

 */

#include "ltl_fairness.h"
#include "fairness.h"

#include <stdio.h>
#include <assert.h>

static int check_fairness(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                          SP_term_ref scc_state_ids, SP_term_ref strong_fairness_assumption,
                          tarjandata *tres, int *tres_scc_size, fairness_data *fair);

/**** 
 Utility functions for handling data used for the fairness algorithm.
 ****/

static int* add_to_array(int* atomids, int len, int* atomids_nonfair, int *size) {
    
    int i;
    int* atomids_nonfair_new = 0;
    if ((*size) + len > 0) {
        atomids_nonfair_new = (int*) ltl_malloc(sizeof(int)*((*size) + len));
    }
    for (i=0; i<(*size); i++) {
        atomids_nonfair_new[i] = atomids_nonfair[i];
    }
    
    for (i=0; i<len; i++) {
        atomids_nonfair_new[i+(*size)] = atomids[i];
    }
    *size = (*size) + len;
    
    if (*size > 0) ltl_free(atomids_nonfair);
    if (len > 0) ltl_free(atomids);
    
    return atomids_nonfair_new;
}

static void copy_data(tarjandata* tres, tarjandata *t, int scc_size) {
    ltlatom *atom;
    int n = 0;
    while (n<scc_size) {
        atom = peek_stack(t,scc_size-1-n);
        push_to_stack(tres,atom,0);
        n++;
    }

    int i, lowlink;
    n=0;
    int rec_size = t->rec_stacksize;
    while (n<rec_size) {
        atom = peek_tarjan_stack(t,&i,&lowlink,rec_size-1-n);
        push_tarjan_stack(tres,atom,i,lowlink);
        n++;
    }
    //assert(tres->rec_stacksize == t->rec_stacksize);
    
}

static void create_tarjan_data_result(tarjandata* t, int scc_size, tarjandata* main_t, int main_scc_size,
                                      tarjandata* tres, int *tres_scc_size) {
        int path_length = 0;
        ltlatom *atom;
        ltlatom *entry_atom_main_scc;
        ltlatom *entry_atom_scc;
        ltlatom **main_scc = save_stack(main_t,main_scc_size,&entry_atom_main_scc,1,0);
        ltlatom **scc = save_stack(t,scc_size,&entry_atom_scc,1,0);
        
        ltlatom **path = get_path_to_scc(main_t->f, main_scc, main_scc_size, entry_atom_main_scc, scc, scc_size, entry_atom_scc, &path_length);
        ltl_free(main_scc);
        ltl_free(scc);
        main_scc = 0;
        scc = 0;

        // Copy the rest of the tarjan data
        int i, lowlink; // Daniel, do we need here the correct values of i and lowlink?
        int n=0;
        int rec_size = main_t->rec_stacksize;
        // saving the path to the main scc
        while (n<rec_size) {
            atom = peek_tarjan_stack(main_t,&i,&lowlink,rec_size-1-n);
            //ltl_printnode_with_info("push node of tarjan path to entry node of the main scc: ",atom->node);
            push_tarjan_stack(tres,atom,i,lowlink);
            n++;
        }
        
        // saving the path from the entry of the main scc to the entry of the found scc 
        int j = 0;
        while (j<path_length) {
            atom = path[j];
            //ltl_printnode_with_info("create_tarjan_data_result: push node (tarjan path): ",atom->node);
            push_tarjan_stack(tres,atom,0,1);
            j++;
        }
        if (path_length > 0) ltl_free(path);

        // saving the found scc component
        j = 0;
        while (j<scc_size) {
            atom = peek_stack(t,scc_size-1-j);
            push_to_stack(tres,atom,0);
            j++;
        }
        *tres_scc_size = scc_size;
}

void print_given_scc(tarjandata *t,int scc_size) {
    printf("SCC size: %d\n",scc_size);
    int i;
    for (i=0; i<scc_size; i++) {
        ltl_printnode_with_info("node (SCC-saved): ",peek_stack(t, i)->node);
    }
}

static int* get_unfair_atom_ids(tarjandata *t, int scc_size, SP_term_ref sf_argument, int *len,
                                SP_term_ref scc_state_ids) {
    
    ltlatom** runfair;
    int* atomids = 0;
    int rsize = 0;
    SP_pred_ref predicate = t->f->get_transition_ids_predicate;
    int msize = 0;
    SP_integer* matching_transitions = get_matching_transitions(predicate,sf_argument,scc_state_ids,&msize, t->f,1);
    if (!check_if_scc_contains_transitions(t,scc_size,matching_transitions,msize)) {
        extract_unfair_atoms(t,scc_size,matching_transitions,msize,&runfair,&rsize);
        int i;
        if (rsize > 0) atomids = (int*) ltl_malloc(sizeof(int)*rsize);
        for (i=0; i<rsize; i++) {
            atomids[i] = find_atom_id(t->f,runfair[i]);
        }
        *len = rsize;
        if (rsize > 0) ltl_free(runfair);
    } else {
        *len = 0;
    }
    if (msize > 0) ltl_free(matching_transitions);
    return atomids;
}

/********** Functions for checking weak and strong fairness **********/

// function for checking recursively the resulted strong components after the decomposition of the main SCC
static int cfair(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                 Graph component, tarjandata *tres, int *tres_scc_size, fairness_data *fair) {

    int result;
    bitfield *evaluation;
    ltlatom *atom;
    
    ltl_formula *f = t->f;
    
    vertex_list vertices = component.vertices;
    
    tarjandata new_t;
    
    init_tarjan_data(&new_t, f);
    // Save new SCC to tarjan stack
    int new_scc_size = 0;
    int atomid;
    while (vertices != NULL) {
        atomid = (vertices->vertex).atomid;
        get_atom(&f->astore, atomid, &atom, &evaluation);
        push_to_stack(&new_t,atom,0);
        vertices=vertices->next;
        new_scc_size++;
    }
    // print_given_scc(&new_t,new_scc_size);
    // assert(new_t.scc_stacksize == new_scc_size );
    if (check_all_until_goals_present(&new_t, new_scc_size)
        && is_nontrivial_scc(&new_t, new_scc_size)) {
                
        // call recursively is_fair_scc_check
        result = is_fair_scc_check(&new_t,new_scc_size,main_t,main_scc_size,tres,tres_scc_size,fair);
    } else {
        result = 0;
    }

    delete_tarjan_data(&new_t);
    return result;
    
}

// The function creates the new SCC-subgraph after removing all impartial nodes and
// decomposes it in SCC's in case the graph is not empty. After that each new SCC will
// be further checked whether it is self-fulfilling SCC and satisfies the strong fairness condition.
static int check_fairness(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                          SP_term_ref scc_state_ids, SP_term_ref strong_fairness_assumption,
                          tarjandata *tres, int *tres_scc_size, fairness_data *fair) {

    int sfargs_size = 0;
    SP_term_ref *sf_arguments = prolog_list_to_array(strong_fairness_assumption, &sfargs_size);
    int result;
    
    int i;
    int size = 0, len = 0;
    int* atomids_nonfair = 0; int* atomids = 0;
    // get all transition ids that are enabled in the SCC and
    // such that are copy of an event/action in the strong fair arguments list
    for (i=0; i<sfargs_size; i++) {
        atomids = get_unfair_atom_ids(t,scc_size,sf_arguments[i],&len,scc_state_ids);
        atomids_nonfair = add_to_array(atomids,len,atomids_nonfair,&size);
    }

    ltl_free(sf_arguments);
    len = size;

    // In case of an empty list, the SCC was fair and we return 1.
    // Otherwise, we need to decompose the current SCC B into SCCs B1,...,Bn and check
    // whether one of these is self-fulfilling, non-trivial and fair.

    if (len == 0) {
        // no impartial vertices found in the SCC component => found SCC is fair
        IFFAIRNESSDEBUG( ltlmsg("No unfair nodes found => SCC component is fair!\n"); );
        create_tarjan_data_result(t, scc_size, main_t, main_scc_size, tres, tres_scc_size);
        SP_put_term(fair->strong,strong_fairness_assumption);
        fair->strong_size = sfargs_size;
        //copy_data(tres,t,scc_size);
        //*tres_scc_size = scc_size;
        result = 1;
    } else {

        IFFAIRNESSTIMING(clock_t now; clock_t timer_start = clock(); );

        // construct an SCC graph without adding all non-fair vertices "atomids_nonfair"
        Graph g = get_scc_graph(t, atomids_nonfair, &len, scc_size);

        IFFAIRNESSTIMING(now = clock(); t->f->strong_fairness_time += (now - timer_start); );
    
        ltl_free(atomids_nonfair);

        if(empty_graph(g)) {
            IFFAIRNESSDEBUG( ltlmsg("Graph is empty => SCC is not fair\n"); );
            delete_graph_data(&g);
            *tres_scc_size = 0;
            result = 0;
            IFLTLTIMING(t->f->rejected_sccs += 1);

        } else {
            IFFAIRNESSDEBUG( ltlmsg("Decompose SCC!\n"); );

            int result_rec_check = 0;
            Graph* scc_components;
            int num_components; // number of all components in g (trivial and non-trivial)
           
            IFFAIRNESSTIMING(clock_t now; clock_t timer_start = clock(); );
            scc_components = decompose_scc(&g,&num_components);
            IFFAIRNESSTIMING(now = clock(); t->f->cpp_time_spent += (now - timer_start); );
            delete_graph_data(&g);
            
            int component_idx;
            for (component_idx=0; component_idx<num_components; component_idx++) {
                if (cfair(t,scc_size,main_t,main_scc_size,scc_components[component_idx],tres,tres_scc_size,fair)) {
                    // there is a fair SCC component in the main SCC
                    result_rec_check = 1;
                    break;
                }
                IFLTLTIMING(t->f->rejected_sccs += 1);
            }

            delete_scc_components_data(scc_components,num_components);
            result = result_rec_check;
        }
    }
    return result;
}

int check_strong_fairness_all(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                              SP_term_ref scc_state_ids, tarjandata *tres, int *tres_scc_size,
                              fairness_data *fair) {
   // need to check SCC for strong fairness in regard to all enabled operations

    SP_pred_ref predicate = t->f->get_enabled_actions_predicate;
    SP_term_ref enabled_actions = get_enabled_actions(predicate,scc_state_ids,t->f,1);
    return check_fairness(t, scc_size, main_t, main_scc_size, scc_state_ids, enabled_actions, tres, tres_scc_size,fair);
}

// The main function for checking strong fairness.
static int check_strong_fairness(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                                 SP_term_ref scc_state_ids, tarjandata *tres, int *tres_scc_size,
                                 fairness_data *fair) {
    int sf_size;
    SP_term_ref *sf_assumptions = prolog_list_to_array(t->f->strong_fairness_arguments,&sf_size);
    
    if (t->f->strong_fairness_all_operations) {
        return check_strong_fairness_all(t,scc_size,main_t,main_scc_size,scc_state_ids,tres,tres_scc_size,fair);
    } else if (sf_size>0) {

        int sf_check_result = 0;
        int i;
        for (i=0; i<sf_size; i++) {
            if (check_fairness(t, scc_size, main_t, main_scc_size, scc_state_ids, sf_assumptions[i], tres, tres_scc_size,fair)) {
                sf_check_result = 1;
                break;
            }
        }
        
        ltl_free(sf_assumptions);
        
        return sf_check_result;
        
    } else {
        copy_data(tres,t,scc_size);
        *tres_scc_size = scc_size;
        return 1;
    }
}

// The function checks a conjunction of fairness actions, e.g. for "SF(a) & SF(b) => f".
// For the example, it will be checked if the SCC violating f is fair regarding actions a and b
int check_weak_and_strong_fairness(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                                   SP_term_ref scc_state_ids, SP_term_ref weak_fairness_arguments,
                                   tarjandata *tres, int *tres_scc_size, fairness_data *fair) {
    IFFAIRNESSTIMING(clock_t now; clock_t timer_start = clock(); );

    int wfargs_size, i;
    int wfair_scc = 1, msize = 0;
    SP_integer* matching_transitions;
    SP_term_ref *wf_arguments = prolog_list_to_array(weak_fairness_arguments,&wfargs_size);
    
    SP_pred_ref predicate = t->f->get_transition_ids_predicate;
    for (i=0; i<wfargs_size; i++) {
        
        matching_transitions = get_matching_transitions(predicate,wf_arguments[i],scc_state_ids,&msize,t->f,1);
        
        // check weak fairness for each operation
        if (    msize > 0
             && !check_if_scc_contains_transitions(t, scc_size, matching_transitions, msize)
             && enabled_in_each_node(t, scc_size, matching_transitions, msize)) {
            wfair_scc = 0;
            ltl_free(matching_transitions);
            IFLTLTIMING(t->f->rejected_sccs += 1);
            break;
        }
        if (msize>0) ltl_free(matching_transitions);
    }
    
    ltl_free(wf_arguments);

    IFFAIRNESSTIMING(now = clock(); t->f->weak_fairness_time += (now - timer_start); );

    if (!wfair_scc) {
        copy_data(tres,t,scc_size);
        *tres_scc_size = scc_size;
        return 0;
    } else {
        // if the SCC fulfills the weak fair constraints then we check for strong fairness
        if (check_strong_fairness(t, scc_size, main_t, main_scc_size, scc_state_ids, tres, tres_scc_size,fair)) {
            SP_put_term(fair->weak,weak_fairness_arguments);
            fair->weak_size = wfargs_size;
            return 1;
        } else {
            return 0;
        }
    }
}

int check_weakall_and_strong_fairness(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                                      SP_term_ref scc_state_ids, tarjandata *tres, int *tres_scc_size,
                                      fairness_data *fair) {
   // need to check SCC for weak fairness in regard to all enabled operations

   SP_pred_ref predicate = t->f->get_enabled_actions_predicate;
   SP_term_ref enabled_actions = get_enabled_actions(predicate,scc_state_ids,t->f,1);
      
   return check_weak_and_strong_fairness(t,scc_size,main_t,main_scc_size,scc_state_ids,enabled_actions,tres,tres_scc_size,fair);
}

/* 
 The main function of this module. is_fair_scc_check returns 1 if the given SCC component (in the t)
  is fair in regard to the weak and strong fairness assumptions, otherwise 0 is returned.
 Arguments:
   t: the tarjan data that comprises the found SCC for the LTL formula t->f
   scc_size: the size of the SCC (the number of atoms on the SCC stack)
   scc_state_ids: is the prolog list of the state IDs in the original state space of the model
   Note that in case the SCC is checked for strong fairness the component can be decomposed
   in various SCCs which probably are also checked in regard to the fairness assumptions at some point.
   main_t: keeps track of the currently checked SCC;
   main_scc_size: the size of the SCC in main_t;
   tres: the SCC which is fair in regard to the fairness assumptions;
   tres_scc_size: the size of the resulted SCC in tres
 */
int is_fair_scc_check(tarjandata *t, int scc_size, tarjandata *main_t, int main_scc_size,
                      tarjandata *tres, int *tres_scc_size, fairness_data *fair) {
  
    SP_term_ref scc_state_ids = create_scc_nodes_list(t,scc_size);

    IFFAIRNESSDEBUG(print_given_scc(t,scc_size));
    
    if (t->f->strong_fairness_all_operations) {
        /* case: SEF => f
         In this case checking the SCC for general strong fairness makes other fairness checks redundant.
         If the SCC is generally strongly fair, then it is also weakly fair. Otherwise, if the SCC is not
         generally strong fair, it will be rejected as a valid counter-example.
         */
        return check_strong_fairness_all(t,scc_size,main_t,main_scc_size,scc_state_ids,tres,tres_scc_size,fair);
    } else if (t->f->weak_fairness_all_operations) {
      /* case: WEF => f */
        return check_weakall_and_strong_fairness(t,scc_size,main_t,main_scc_size,scc_state_ids,tres,tres_scc_size,fair);
    } else {
        /* 
           First we check all weak fairness assumptions. There is a possibility that the user
           may want to check the SCC for several weak fairness assumptions. These can be given
           in form of disjunctions. For istance, "(WF(a) & WF(b)) or WF(c) => f" sets two weak
           fairness assumptions. Each time an SCC C is reported as a counter-example for f,
           it should be checked if C is weakly fair in regard to the events a and b, or weakly
          fair in regard to the event c.
         */
        int wf_size;
        SP_term_ref *wf_assumptions = prolog_list_to_array(t->f->weak_fairness_arguments,&wf_size);
        
        if (wf_size>0) {
            
            int wf_check_result = 0;
            int i;
            for (i=0; i<wf_size; i++) {
                if (check_weak_and_strong_fairness(t,scc_size,main_t,main_scc_size,scc_state_ids,wf_assumptions[i],tres,tres_scc_size,fair)) {
                    wf_check_result = 1;
                    break;
                }
            }
            ltl_free(wf_assumptions);
            return wf_check_result;
            
        } else {
            // only strong fairness constraints have been set
            return check_strong_fairness(t,scc_size,main_t,main_scc_size,scc_state_ids,tres,tres_scc_size,fair);
        }
    }
}

