/*  
 ltl_fairness.cpp
 
 Created by Ivaylo Dobrikov on 12/9/13.
 
 Functions for determining SCCs of a given graph. Used for checking LTL fairness.

*/

#include "ltl_fairness.h"

/* boost library includes */
#include "boost/graph/adjacency_list.hpp"
#include "boost/graph/strong_components.hpp"
#include "boost/graph/graph_utility.hpp"

/* standard C++ library includes */
#include <iostream>
#include <assert.h>

enum vertex_atom_id_t { vertex_atom_id };
namespace boost {
    BOOST_INSTALL_PROPERTY(vertex, atom_id);
}

Graph* decompose_scc(Graph *g, int* components_size) {
    
    //Pre-condition: the function should be called only when g is non-empty graph
    assert(g->vertices != NULL);
    
    using namespace boost;
    typedef adjacency_list < vecS, vecS, directedS, property<vertex_atom_id_t,unsigned int> > BoostGraphType;
    
    BoostGraphType G;
    // defining a property map for mapping vertices to their atom ids
    // (vs->vertex).graphid --> (vs->vertex).atomid (see calls of the put-function below)
    property_map<BoostGraphType, vertex_atom_id_t>::type atomid = get(vertex_atom_id, G);

    // Construct SCC graph
    
    // Add vertices
    GraphVertex vertex;
    vertex_list current_vlist = g->vertices;
    while (current_vlist != NULL) {
        vertex = current_vlist->vertex;
        add_vertex(vertex.graphid,G);
        current_vlist = current_vlist->next;
    }

    // Add the atomid-property to each vertex
    current_vlist = g->vertices;
    while (current_vlist != NULL) {
        vertex = current_vlist->vertex;
        put(atomid, vertex.graphid, vertex.atomid);
        current_vlist = current_vlist->next;
    }
    
    // the graph must not be empty at this point
    assert(num_vertices(G) > 0);
    
    // Add edges
    GraphEdge edge;
    edge_list current_elist = g->edges;
    while (current_elist != NULL) {
        edge = current_elist->edge;
        add_edge(edge.src.graphid, edge.dst.graphid, G);
        current_elist = current_elist->next;
        IFFAIRNESSDEBUG( std::cout << "GraphEdge added: " << edge.src.graphid << " --> " << edge.dst.graphid << std::endl);
    }
    
    IFFAIRNESSDEBUG(print_graph(G));
    
    // Determine the remained SCCs
    typedef graph_traits<adjacency_list<vecS, vecS, directedS> >::vertex_descriptor Vertex;
    
    // component is a vector array where each vertex j is mapped to the respective component id (component[j])
    std::vector<int> component(num_vertices(G));
    std::vector<Vertex> root(num_vertices(G));
    int num = strong_components(G, &component[0]);
    
    // in a non-empty graph there is at least one strong component
    assert(num > 0);
    
    // picking the strong components and representing those as graphs from type Graph
    IFFAIRNESSDEBUG(std::cout << "Total number of components: " << num << std::endl);

    // iterating through all vertices
    std::vector<int>::size_type j;
    
    // visited is used to guarantee that each vertex is saved in a unique strong component
    int visited[component.size()];
    for (j=0; j<component.size(); j++) visited[j]=0;
    
    Graph *connected_components = (Graph*) ltl_malloc(sizeof(Graph)*num);
    int component_idx = 0;
    for (j = 0; j != component.size(); ++j) {
        
        int cmp_id = component[j];
        std::vector<int>::size_type k;
        int sz = 0;
        
        if (! visited[j]) {
            // determine size of the strong component
            for (k=0; k != component.size(); ++k) {
                if (cmp_id == component[k]) {
                    sz++;
                    visited[k] = 1;
                }
            }
            
            IFFAIRNESSDEBUG(std::cout << "Components Size is " << sz << std::endl);

            vertex_list vs = NULL;

            int idx = 0;
            for (k=0; k != component.size(); ++k) {
                // add vertices to vs* with same component ids
                if (cmp_id == component[k]) {
                    GraphVertex v;
                    v.graphid = k;
                    v.atomid = get(atomid, k);
                    vs = insert_vertex(vs,v);
                    idx++;
                }
            }
            
            // initialize the respective SCC as graph from type Graph
            Graph new_graph;
            new_graph.vertices= vs;
            // no edges need to be stored in new_graph,
            // since we are interested only in the atom ids of the componenets
            new_graph.edges = NULL;

            connected_components[component_idx] = new_graph;
            component_idx++;
        }
    }
    
    *components_size = component_idx;
    return connected_components;
}
