
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "ltlc.h"

#include "tarjan.h"
#include "fairness.h"
#include "graph_utilities.h"

static int is_fair_scc(tarjandata *t, int scc_size, tarjandata *tres, int *tres_scc_size, fairness_data *fair);
static int tarjan(tarjandata *t, ltlatom *from, found_scc_type **result);
static found_scc_type* get_found_scc(tarjandata *t, int scc_size, int sort_scc, fairness_data fair);
const int update_countdown_init = 100000;
const clock_t ltl_update_interval = (clock_t) (CLOCKS_PER_SEC * 1.0);
void sortscc(int atomnumber, ltlatom **scc);
static found_scc_type *is_valid_scc(tarjandata *t, int scc_size);

void push_tarjan_stack(tarjandata *s, ltlatom *from, int i, int lowlink) {
  assure_arraysize(&s->rec_stack, s->rec_stacksize+1);
  tarjanstate *ts = (tarjanstate*) get_array_element(&s->rec_stack, s->rec_stacksize);
  ts->from = from;
  ts->iterator = i;
  ts->lowlink = lowlink;
  s->rec_stacksize++;
}

static void pop_tarjan_stack(tarjandata *s, ltlatom **from, int *i, int *lowlink) {
  if (s->rec_stacksize > 0) {
    s->rec_stacksize--;
    tarjanstate *ts = (tarjanstate*) get_array_element(&s->rec_stack, s->rec_stacksize);
    *from = ts->from;
    *i = ts->iterator;
    *lowlink = ts->lowlink;
  } else {
    fprintf(stderr, "pop_tarjan_stack(): empty stack!");
    exit(1);    
  }
}

ltlatom *peek_tarjan_stack(tarjandata *s, int *i, int *lowlink, int pos) {
    tarjanstate *ts = (tarjanstate*) get_array_element(&s->rec_stack, s->rec_stacksize-1-pos);
    ltlatom *atom = ts->from;
    *i = ts->iterator;
    *lowlink = ts->lowlink;
    return atom;
}

/* Note: function used in tarjan.c and fairness.c modules. For using this function
         for a tarjan SCCs search set_on_tarjan_stack should be set to 1, otherwise to 0. */
void push_to_stack(tarjandata *t, ltlatom *atom, int set_on_tarjan_stack) {
  ltlatom **stackpointer;
  assure_arraysize(&t->scc_stack, t->scc_stacksize+1);
  stackpointer = (ltlatom**) get_array_element(&t->scc_stack, t->scc_stacksize);
  *stackpointer = atom;
  atom->tarjan_on_stack = set_on_tarjan_stack;
  (t->scc_stacksize)++;
}

ltlatom *peek_stack(tarjandata *t, int pos) {
  return *((ltlatom **) get_array_element(&t->scc_stack, t->scc_stacksize-1-pos));
}

static int inspect_scc_stack(tarjandata *t, int number) {
  ltlatom *atom;
  int n = 0;
  if (t->scc_stacksize > 0) {
    atom = peek_stack(t, 0);
    while (n<t->scc_stacksize && atom->tarjan_number >= number) {
      n++;
      if (n<t->scc_stacksize) atom = peek_stack(t,n);
    }
  } else {
    fprintf(stderr, "inspect_scc_stack(): empty scc stack!");
    exit(1);
  }
  return n;
}

ltlatom **save_stack(tarjandata *t, int size, ltlatom **entry, int sort_scc, int debug) {
  ltlatom **saved;
  int i;
  IFLTLDEBUG( ltlmsg("save scc stack (%i elements)\n", size); );
  if (t->scc_stacksize >= size) {
    saved = (ltlatom **) ltl_malloc(sizeof(ltlatom*) * size);
    for (i=0; i<size; i++) {
      saved[size-1-i] = peek_stack(t, i);
      if(debug) IFFAIRNESSDEBUG( ltl_printnode_with_info("node (SCC-saved): ",saved[size-1-i]->node););
    }
    *entry = saved[0];
  } else {
    fprintf(stderr, "save_stack(): scc stack underflow!!");
    exit(1);
  }
  /* sort scc to enable a fast binary search to check
     if an atom is in the scc */
  if (sort_scc) {
     sortscc(size, saved);
  }
  return saved;
}

static void decrease_stack(tarjandata *t, int dec) {
  int i;
  if (t->scc_stacksize >= dec) {
    for (i=0; i<dec; i++) {
      peek_stack(t,i)->tarjan_on_stack = 0;
    }
    t->scc_stacksize -= dec;
  } else {
    fprintf(stderr, "decrease_stack(): scc stack underflow!!");
    exit(1);
  }
}

static int update_call2(tarjandata *t) {
  int status, result;
  SP_term_ref states, gen_atoms, gen_trans, vis_atoms, vis_trans;
  ltl_formula *f = t->f;
  clock_t now = clock();
  if (f->update_countdown < 0 || now >= f->last_update_call + ltl_update_interval) {
    states = SP_new_term_ref();
    gen_atoms = SP_new_term_ref();
    gen_trans = SP_new_term_ref();
    vis_atoms = SP_new_term_ref();
    vis_trans = SP_new_term_ref();
    SP_put_integer(states, (long) f->nstore.size);
    SP_put_integer(gen_atoms, (long) f->number_of_atoms);
    SP_put_integer(gen_trans, (long) f->number_of_atomtrans);
    SP_put_integer(vis_atoms, (long) t->current_number);
    SP_put_integer(vis_trans, (long) t->checked_transitions);
    
    status = SP_query(f->update_predicate, states, gen_atoms, gen_trans, vis_atoms, vis_trans);
    switch (status) {
    case SP_SUCCESS: /* nothing to do */ 
      result=1; 
      break;
    case SP_FAILURE: /* cancel */
      IFLTLDEBUG( ltlmsg("cancel!\n"); )
      f->canceled = 1;
      result = 0; 
      break;
    default: /* exception */
      IFLTLDEBUG( ltlmsg("update raised exception!\n");
                  ltldeb_dec(); );
      ltl_reraise_exception();
      result = 0;
      break;
    }
    f->last_update_call = now;    
  } else {
    result = 1;
  }
  f->update_countdown = update_countdown_init;
  return result;
}

static int update_call(tarjandata *t) {
  if (t->f->update_predicate_activated) {
    return update_call2(t);
  } else {
    return 1;
  }
}

/* returns lowlink of node */
int tarjan(tarjandata *t, ltlatom *from, found_scc_type **result) {
  found_scc_type *found_scc;
  int i;
  ltlatom *to;
  ltl_formula *formula = t->f;
  int recdepth = 0;
  int lowlink = 0, seen_lowlink = -1;
  int abort = 0;

  IFLTLDEBUG(int debug_depth = ltldeb_get(););

  found_scc = 0;
  i=0;

  IFLTLDEBUG(ltlmsg("tarjan: start with atom %i\n", find_atom_id(formula, from));
             ltldeb_inc(););

  while (!found_scc && !abort && recdepth >= 0) {
      
      /* printing out some LTL statistics */
      IFLTLTIMING( clock_t elapsed_time = clock() - formula->timer_current;
                  if (elapsed_time / CLOCKS_PER_SEC > 20) { // every 20 seconds
                      formula->timer_current = clock();
                      printf("LTL (current statistics): %i atoms, %i transitions generated, and  %i callbacks needed.\n",
                             formula->number_of_atoms, formula->number_of_atomtrans, formula->callback_calls);
                      
                      if (formula->check_found_scc_for_fairness_activated) {
                          printf("LTL (fairness): %i strongly connected components were rejected, %i callbacks needed.\n\n",
                                 formula->rejected_sccs, formula->fairness_callback_calls);
                      }
                  });
      
      if (i==0) {
      /* recursion step to new atom */

      if (explore_atom(formula, from)) {
        from->tarjan_number = t->current_number;
        t->current_number++;

        push_to_stack(t, from, 1);
        lowlink = from->tarjan_number;

        IFLTLDEBUG(ltlmsg("tarjan: atom %i (num %i) has %i transitions\n",
                          find_atom_id(formula, from),
                          from->tarjan_number,
                          from->number_of_transitions,
                          from->node->number_of_transitions);
                   ltldeb_inc();ltldeb_inc(););
      } else {
        IFLTLDEBUG(ltlmsg("tarjan: callback for atom %i failed, abort\n",
                          find_atom_id(formula, from)););
        abort = 1;
      }
    } else {
      /* after returning from a recursion step */
      IFLTLDEBUG(ltlmsg("tarjan: atom %i (num %i): returned from recursion.\n",
                        find_atom_id(formula, from),
                        from->tarjan_number););
      
      if(seen_lowlink != -1 && lowlink > seen_lowlink) {
          lowlink = seen_lowlink;
      }
    }

    if (!abort) {
      /* check if we should call the update callback */
      if (formula->update_countdown <= 0 && !update_call(t)) {
        abort = 1;
      }
      formula->update_countdown--;
    }

    if(!abort) {
      if(i < from->number_of_transitions) {
        /* next iteration step */

        IFLTLDEBUG(ltlmsg("tarjan: atom %i (num %i): iteration %i of %i\n",
                          find_atom_id(formula, from),
                          from->tarjan_number,
                          i,
                          from->number_of_transitions););

        to = get_atom_trans(formula, from->first_transition_id + i);
        /* increase number of checked transitions (for statistics only) */
        t->checked_transitions++;

        if (to->tarjan_number < 0) {
          IFLTLDEBUG(ltlmsg("tarjan: from atom %i found new atom %i \n", 
                            find_atom_id(formula, from),
                            find_atom_id(formula, to)); );
          IFLTLDEBUG(ltlmsg("tarjan:  newly found"); );
          /* unvisited atom, evaluate recursively  */
          push_tarjan_stack(t, from, i, lowlink);
          recdepth++;
          from=to;
          i=0;
          seen_lowlink = -1;
        } else {
          if (to->tarjan_on_stack && to->tarjan_number < lowlink) {
            IFLTLDEBUG(ltlmsg("tarjan: from atom %i found known atom %i on stack \n", 
                              find_atom_id(formula, from),
                              find_atom_id(formula, to)); );
            lowlink = to->tarjan_number;
            /* in POR case we need to expand fully state 'from' */

          } else {
            IFLTLDEBUG(ltlmsg("tarjan: from atom %i found known atom %i (not on stack) \n", 
                              find_atom_id(formula, from),
                              find_atom_id(formula, to)); );
          }
          i++;
        }
      } else {
        /* end of iteration reached */
        if (lowlink == from->tarjan_number) {
          /* SCC detected */
          int scc_size = inspect_scc_stack(t, from->tarjan_number);
          IFLTLDEBUG(ltlmsg("SCC found, size=%i\n",scc_size););
          IFLTLDEBUG( ltldebug_store_scc(t,scc_size); );
          found_scc = is_valid_scc(t,scc_size);
          decrease_stack(t, scc_size);
        }

        if (!found_scc) {
#ifdef LTLSTORELOWLINK
          from->tarjan_lowlink = lowlink;
#endif

          IFLTLDEBUG(/*ltlmsg("tarjan(atom %i, num %i) finished%s.\n",
                       find_atom_id(formula, from),
                       from->tarjan_number,
                       found_scc ? ", counterexample found" : "");*/
                     ltldeb_dec();ltldeb_dec(); );
  
          seen_lowlink = lowlink;

          /* get back one recursion level */
          recdepth--;
          if (recdepth >= 0) {
            pop_tarjan_stack(t, &from, &i, &lowlink);
            i++;
          }
        }
      }
    }
  }

  IFLTLDEBUG(ltldeb_set(debug_depth););

  *result = found_scc;

  return abort ? 0 : 1;
}

static found_scc_type* get_found_scc(tarjandata *tres, int scc_size, int sort_scc, fairness_data fair) {
  int i,lowlink;
  ltlatom *from;

  found_scc_type *found_scc;
  int j;
  ltlatom *entry;

  int recdepth = tres->rec_stacksize;
  found_scc = (found_scc_type *) ltl_malloc(sizeof(found_scc_type));
  found_scc->scc = save_stack(tres, scc_size, &entry, sort_scc, 1);
  found_scc->scc_entry = entry;
  IFFAIRNESSDEBUG(ltl_printnode_with_info("entry node of SCC (tarjan): ",found_scc->scc_entry->node););
  found_scc->number_of_scc_atoms = scc_size;
  found_scc->path = (ltlatom **) ltl_malloc(sizeof(ltlatom*) * recdepth);
  found_scc->number_of_path_elements = recdepth;
  for (j=recdepth-1; j>=0; j--) {
    pop_tarjan_stack(tres, &from, &i, &lowlink);
    IFFAIRNESSDEBUG(ltl_printnode_with_info("node (Path): ",from->node););
    found_scc->path[j] = from;
  }

  found_scc->fairness_assumptions = fair;
  return found_scc;
}
    
static int all_yesterdays_false(ltl_formula *f, bitfield *evaluation) {
  int i;
  for (i=0; i<f->number_of_yesterdays; i++) {
    if (getbit(evaluation, f->index_yesterdays + i)
        || getbit(evaluation, f->index_yesterdaynots + i)) {
      return 0;
    }
  }
  return 1;
}

void init_tarjan_data(tarjandata *t, ltl_formula *f) {
    
  t->f = f;

  init_array(&t->scc_stack, sizeof(ltlatom*), 0, "scc_stack");
  t->scc_stacksize = 0;

  init_array(&t->rec_stack, sizeof(tarjanstate), 0, "rec_stack");
  t->rec_stacksize = 0;

  t->current_number = 0;
  t->checked_transitions = 0;

  t->scccheck_bf_used = (bitfield *) ltl_malloc(get_bitfield_memsize(f->closure_size));
  t->scccheck_bf_goal = (bitfield *) ltl_malloc(get_bitfield_memsize(f->closure_size));
}

void delete_tarjan_data(tarjandata *t) {
  ltl_free(t->scccheck_bf_used);
  ltl_free(t->scccheck_bf_goal);

  delete_array(&t->scc_stack);
  delete_array(&t->rec_stack);
}

int explore_initialisation_nodes(ltl_formula *f, ltlnode **starts, int number_of_starts) {
  int i,failure = 0;
  for (i=0; !failure && i<number_of_starts; i++) {
    failure = !explore_node(f, starts[i], 1);
  }
  return failure;
}

/*
 * Returns the atom if it's an initial atom, otherwise returns NULL
 */
ltlatom *get_initial_atom(ltl_formula *f, unsigned int atomid) {
  ltlatom *atom;
  ltl_pform_ref *mainref = &(f->mainref);
  bitfield *evaluation;
  get_atom(&f->astore, atomid, &atom, &evaluation);
  if (!eval(evaluation, mainref)
      || !all_yesterdays_false(f, evaluation)) {
    atom = NULL;
  }
  return atom;
}

int start_tarjan(ltl_formula *f,
                 ltlnode **starts,
                 int number_of_starts,
                 size_t *used_memory,
                 found_scc_type **result) {
  unsigned int j,i,firstatomid;
  ltlatom *atom;
  int failure = 0;
  tarjandata t;

  init_tarjan_data(&t,f);

  /* first explore the initialisation nodes, 
     (i.e. create their atoms)
     they must be treated differently in the sence that
     all "yesterdays" are false */
  failure = explore_initialisation_nodes(f,starts,number_of_starts);
  
  /* now, iterate through all initialisation atoms and
     start a tarjan search from each. Stop when we
     encounter an SCC or failure.
  */
  found_scc_type *found_scc = 0;
  for (j=0; !found_scc && !failure && j<number_of_starts; j++) {
    ltlnode *start = starts[j];
    firstatomid = start->first_atom_id;
    for (i=0; !found_scc && !failure && i<start->number_of_atoms; i++) {
      atom = get_initial_atom(f,firstatomid+i);
      /* tarjan_number<0 means that the atom is not yet explored */
      if (atom && atom->tarjan_number < 0) {
        failure = !tarjan(&t, atom, &found_scc);
      }
    }
  }

  /* record used memory for stacks */
  *used_memory =  ltl_array_memory_usage(&t.scc_stack) + 
    ltl_array_memory_usage(&t.rec_stack);

  delete_tarjan_data(&t);

  *result = found_scc;
  return !failure;
}

/*
 * in a self-fulfilling SCC, for each true until, there
 * must be an atom in the SCC where the goal of the until
 * is true.
 */
int check_all_until_goals_present(tarjandata *t, int scc_size) {
  int j, i;
  unsigned int until;
  ltl_formula *formula = t->f;
  clear_bitfield(t->scccheck_bf_used, formula->closure_size);
  clear_bitfield(t->scccheck_bf_goal, formula->closure_size);
  ltl_part_formula *cl = formula->closure;
  
  /* iterate through all atoms and store for which
   * untils there is an atom where the until is true
   * and an atom where its goal is true
   */
  for (i=0; i<scc_size; i++) {
    ltlatom *atom = peek_stack(t,i);
    bitfield *evaluation = get_atom_evaluation(atom);
    for (j=0; j < formula->number_of_untils; j++) {
      until = formula->untils[j];
      /* is the until true? */
      if (getbit(evaluation,until)) {
        setbit(t->scccheck_bf_used,until,1);
      }
      /* is the goal of the until true? */
      if (!getbit(t->scccheck_bf_goal,until) &&
          eval(evaluation, &(cl[until].args.case_until.goal))) {
        setbit(t->scccheck_bf_goal,until,1);
      }
    }
  }
  /* now check for each until if "until==true => goal(until)==true" */
  for (j=0; j < formula->number_of_untils; j++) {
    until = formula->untils[j];
    if (getbit(t->scccheck_bf_used,until) &&
        !getbit(t->scccheck_bf_goal,until)) {
      IFLTLDEBUG( ltlmsg("not all until goals present.\n") );
      return 0;
    }
  }
  return 1;
}

/*
 * an SCC is trivial if it contains of just a single
 * atom without transitions and its node is not a 
 * deadlock
 */
int is_nontrivial_scc(tarjandata *t, int scc_size) {
    if (scc_size > 1) {
        /* non-trivial */
        IFLTLDEBUG(ltlmsg("SCC is non-trivial (card > 0)\n");)
        return 1;
    } else {
        if (scc_size == 1) {
            ltlatom *atom = peek_stack(t,0);
            if (atom->node->number_of_transitions == 0) {
                IFLTLDEBUG(ltlmsg("SCC is deadlock -> non-trivial\n");)
                /* deadlock, ok */
                return 1;
            } else {
                /* check if there is a reference to the atom itself */
                unsigned int i, first = atom->first_transition_id;
                for (i=0; i<atom->number_of_transitions; i++) {
                    if(get_atom_trans(t->f, first+i) == atom) {
                        IFLTLDEBUG(ltlmsg("SCC contains self-reference -> non-trivial\n");)
                        return 1;
                    }
                }
                IFLTLDEBUG(ltlmsg("SCC is trivial\n");)
                return 0;
            }
        } else {
            fprintf(stderr,"assertion error: found SCC is empty.");
            exit(1);
        }
    }
}

/* TODO: this is copy&pasted from ltl_interface.c */
SP_term_ref get_node_id(ltlnode *node) {
    SP_term_ref result;
    switch (node->type) {
        case term_node: result = node->id.ref; break;
        case int_node: result = SP_new_term_ref(); SP_put_integer(result, node->id.id); break;
        default: fprintf(stderr, "fatal: unexpected node type\n"); exit(1);
    }
    return result;
}

void print_fairness_conditions(SP_term_ref fairness_agrs, char const *type) {
    SP_term_ref
        list = SP_new_term_ref(),
        elem = SP_new_term_ref();
    printf("Printing out %s fairness arguments.\n", type);
    SP_put_term(list,fairness_agrs);
    
    SP_pred_ref pred;
    /* Look up pp_tp/2 */
    if (!(pred = SP_predicate("pp_tp2",2,"ltl_translate"))) {
        fprintf(stderr, "Could not find ltl_translate:pp_tp/2\n");
        exit(1);
    }
    SP_term_ref arg, res;
    SP_qid goal;
    char const *text=NULL;
    printf("[");
    while (SP_get_list(list,elem,list)) {
        SP_put_term(arg = SP_new_term_ref(), elem);
        SP_put_variable(res = SP_new_term_ref());
        if (!(goal = SP_open_query(pred,arg,res))) {
            fprintf(stderr, "Failed to open query.\n");
            exit(1);
        }
        if (text) {
            printf(", ");
        }
        while (SP_next_solution(goal)==SP_SUCCESS) {
            SP_get_string(res,&text);
            printf("%s", text);
        }
        SP_close_query(goal);
    }
    printf("]\n");
}

/*
 * check if an SCC is valid by using a callback
 * This way fairness conditions are implemented.
 * At this point it is already checked that the SCC
 * is non-trivial and self-fulfilling.
 */
static int is_fair_scc(tarjandata *t, int scc_size, tarjandata *tres, int *tres_scc_size,
                       fairness_data *fair) {
  
    init_mallocs_graph_counter();
  
    int result = is_fair_scc_check(t, scc_size, t, scc_size, tres, tres_scc_size, fair);
    
    // assure that there are no memory leaks after fairness check
    assert_open_mallocs_graph_are_zero();
  
    return result;
}

static found_scc_type *is_valid_scc_with_fairness(tarjandata *t, int scc_size, fairness_data *fair) {
    found_scc_type *found_scc;
    tarjandata tres;
    int tres_scc_size;
    init_tarjan_data(&tres, t->f);
    
    if( is_fair_scc(t, scc_size, &tres, &tres_scc_size, fair) ) {
        found_scc = get_found_scc(&tres, tres_scc_size, 1, *fair);
    } else {
        found_scc = 0;
    }
    delete_tarjan_data(&tres);
    return found_scc;
}

static found_scc_type *is_valid_scc(tarjandata *t, int scc_size) {
  found_scc_type *found_scc;
  if (check_all_until_goals_present(t, scc_size) && is_nontrivial_scc(t, scc_size)) {
      
      fairness_data fair;
      fair.weak = SP_new_term_ref();
      fair.weak_size = 0;
      fair.strong = SP_new_term_ref();
      fair.strong_size = 0;

    if (must_check_fairness(t)) {
      clock_t now;

      IFLTLTIMING ( clock_t timer_start = clock(); );
      found_scc = is_valid_scc_with_fairness(t,scc_size,&fair);
      IFFAIRNESSDEBUG(print_fairness_conditions(fair.weak,"weakly");
                      print_fairness_conditions(fair.strong,"strongly"););

      now = clock();
      IFLTLTIMING( t->f->fairness_checking_time += (now - timer_start););
    } else {
      found_scc = get_found_scc(t,scc_size,1,fair);
    }
  } else {
    found_scc = 0;
  }
  return found_scc;
}

static int atomcmp(const void *va, const void *vb) {
  const ltlatom *a = *((const ltlatom **)va);
  const ltlatom *b = *((const ltlatom **)vb);
  return a<b ? -1 : (a>b ? 1 : 0);
}

/* sort the atoms in the SCC. This is done to 
 * enable a binary search in the SCC, see
 * function is_in_scc(). */
void sortscc(int atomnumber, ltlatom **scc) {
  qsort(scc, atomnumber, sizeof(ltlatom*), atomcmp);
}

/* check if the given atom is in the SCC. This function
 * relies on the fact that the atoms in the SCC are
 * sorted w.r.t. atomcmp(). */
int is_in_scc(ltlatom *atom, int atomnumber, ltlatom **scc) {
  return bsearch(&atom, scc, atomnumber, sizeof(ltlatom*), atomcmp) != 0;
}
