// (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
// Heinrich Heine Universitaet Duesseldorf
// This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


// see http://blog.demofox.org/2013/07/06/fast-lightweight-random-shuffle-functionality-fixed/
// for the basic idea and explanation

#include <sicstus/sicstus.h>
#include "random_permutations_glue.h"

#include <stdint.h>

/* MurMur Hash 3 32-bit variant and 64-bit variant depending on system */
/* conditionals are copied from sicstus.h */
#if SP_SIZEOF_VOID_P == 4 /* 32-bit */
SP_integer murmur3(SP_integer* key, uint32_t len, SP_integer seed) {
    static const uint32_t c1 = 0xcc9e2d51;
    static const uint32_t c2 = 0x1b873593;
    static const uint32_t r1 = 15;
    static const uint32_t r2 = 13;
    static const uint32_t m = 5;
    static const uint32_t n = 0xe6546b64;

    uint32_t hash = seed;

    const int nblocks = len / 4;
    const uint32_t *blocks = (const uint32_t *) key;
    int i;
    for (i = 0; i < nblocks; i++) {
        uint32_t k = blocks[i];
        k *= c1;
        k = (k << r1) | (k >> (32 - r1));
        k *= c2;

        hash ^= k;
        hash = ((hash << r2) | (hash >> (32 - r2))) * m + n;
    }

    const uint8_t *tail = (const uint8_t *) (key + nblocks * 4);
    uint32_t k1 = 0;

    switch (len & 3) {
    case 3:
        k1 ^= tail[2] << 16;
    case 2:
        k1 ^= tail[1] << 8;
    case 1:
        k1 ^= tail[0];

        k1 *= c1;
        k1 = (k1 << r1) | (k1 >> (32 - r1));
        k1 *= c2;
        hash ^= k1;
    }

    hash ^= len;
    hash ^= (hash >> 16);
    hash *= 0x85ebca6b;
    hash ^= (hash >> 13);
    hash *= 0xc2b2ae35;
    hash ^= (hash >> 16);

    return hash;
}
#elif SP_SIZEOF_VOID_P == 8 /* 64-bit */
SP_integer murmur3(SP_integer* key, int len, SP_integer seed) {
  const uint64_t m = 0xc6a4a7935bd1e995LLU;
  const int r = 47;

  uint64_t h = seed ^ (len * m);

  const uint64_t * data = (const uint64_t *)key;
  const uint64_t * end = (len >> 3) + data;

  while(data != end)
  {
    uint64_t k = *data++;

    k *= m;
    k ^= k >> r;
    k *= m;

    h ^= k;
    h *= m;
  }

  const unsigned char * data2 = (const unsigned char *)data;

  switch(len & 7)
  {
  case 7: h ^= (uint64_t)(data2[6]) << 48;
  case 6: h ^= (uint64_t)(data2[5]) << 40;
  case 5: h ^= (uint64_t)(data2[4]) << 32;
  case 4: h ^= (uint64_t)(data2[3]) << 24;
  case 3: h ^= (uint64_t)(data2[2]) << 16;
  case 2: h ^= (uint64_t)(data2[1]) << 8;
  case 1: h ^= (uint64_t)(data2[0]);
          h *= m;
  };

  h ^= h >> r;
  h *= m;
  h ^= h >> r;

  return h;
}
#endif /* 64-bit */

void get_num_bits(SP_integer length, SP_integer *nextPower, SP_integer *numBits) {
    // find (next) larger power of 4 to build up a
    // balanced feistel network
    // the next power of 4 is the maximal index of the permutation
    *nextPower = 4;
    while (length > *nextPower) {
        *nextPower = *nextPower * 4;
    }

    *numBits = 0;
    int mask = *nextPower - 1;

    while(mask) {
        mask = mask >> 1;
        *numBits = *numBits + 1;
    }

    // algorithm relies on exchanging left and right
    // therefor we need half as many bits (in each group)
    *numBits = *numBits / 2;
}

// symmetric feistel network. the "randomness" comes from xoring in
// the hash/seed of the bits on the right side
void feistel_rounds(SP_integer seed, SP_integer rightMask, SP_integer* left, SP_integer* right) {
    // stop after five feistel rounds.
    // this can be increased to provide
    // more/better randomness
    int index;
    for (index = 0; index < 5; index++) {
        SP_integer newLeft = *right;
        SP_integer newRight = *left ^ (murmur3(right, sizeof(*right), seed) & rightMask);

        *left = newLeft;
        *right = newRight;
    }
}

// we return a term random_element(TheElement,NextIndex)
void draw_index(SP_integer idx, SP_integer maxIdx, SP_integer seed, SP_integer halfNumBits, SP_integer leftMask, SP_integer rightMask, SP_integer from, SP_integer to, SP_integer *rnd, SP_integer *nextIdx) {
    SP_integer left = 0;
    SP_integer right = 0;
    do {
        // split index in left and right
        left =  (idx & leftMask) >> halfNumBits;
        right = idx & rightMask;
        feistel_rounds(seed,rightMask,&left,&right);
        *rnd = (left << halfNumBits) | right;
        idx = idx + 1;
    } while (*rnd > to - from && idx <= maxIdx);

    if(idx > maxIdx) {
      SP_fail();
    }

    *nextIdx = idx;
}
