% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_trace_checking, [tcltk_check_state_sequence_from_file/1,
                             tcltk_check_sequence_from_file/1,
                             check_default_trace_for_specfile/1, get_default_trace_file/2,
                             print_trace_as_fdr_check/0,
                             perform_initialisation/0,
                             find_successor_state/4
                            ]).

:- use_module(library(lists)).

:- use_module(tools).
:- use_module(module_information,[module_info/2]).
:- module_info(group,testing).
:- module_info(description,'Replay saved traces; raise error if not possible. Used in regression testing.').

:- use_module(bmachine,[b_machine_name/1]).
:- use_module(b_global_sets,[lookup_global_constant/2]).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(specfile,[b_mode/0]).

:- use_module(specfile,[csp_mode/0]).

:- use_module(translate,[translate_event/2]).

:- use_module(state_space,[current_options/1, current_state_id/1, current_expression/2]).

:- use_module(junit_tests).

% --------------------------------

% replay trace with state predicates; one predicate per line

% b_trace_checking:tcltk_check_state_sequence_from_file('~/Desktop/scheduler6.ptrace').

tcltk_check_state_sequence_from_file(File) :-
   open(File,read,Stream,[encoding('UTF-8')]),
   state_space:current_state_id(ID),
   format(user_output,'Opened predicate list file: ~w~n Starting from state ~w.~n',[File,ID]),
   call_cleanup(tcltk_check_state_sequence_from_file_aux(File,0,Stream,ID),
                (format(user_output,'Finished processing list file: ~w~n',[File]),
                 close(Stream))).

tcltk_check_state_sequence_from_file_aux(File,LineNr,Stream,ID) :- read_line(Stream,Codes),
   Codes \= end_of_file,
   L1 is LineNr+1,
   !, %print(codes(Codes)),nl,
   process_line(File,L1,Stream,Codes,ID).
tcltk_check_state_sequence_from_file_aux(_,LineNr,_,_) :- print('reached EOF at line: '), print(LineNr),nl.

:- use_module(bmachine,[b_parse_wo_type_machine_predicate_from_codes_to_raw_expr/2,
                    b_type_check_raw_expr/4]).
:- use_module(translate_keywords).

% TO DO: also allow operation names, parameters to be specified?
process_line(File,LineNr,Stream,Codes,ID) :-
   % insert newlines to ensure error message correct:
    insert_nl_prefix(LineNr,Codes,CodesNL),
    bmachine:b_parse_wo_type_machine_predicate_from_codes_to_raw_expr(CodesNL,Parsed),
    % in case we did translate keywords (e.g., in Event-B mode); translate them back:
    translate_keywords:raw_backtranslate_keyword_ids(Parsed,Parsed2),
    % now type check the raw expression:
    bmachine:b_type_check_raw_expr(Parsed2,[prob_ids(all),variables],Pred,closed),
    % bmachine:b_parse_machine_predicate_from_codes(CodesNL,[prob_ids(all),variables],Pred),
    !,
    format('Line ~w : ~w --?--> ',[LineNr,ID]),translate:print_bexpr(Pred),nl,
    find_successor_state(ID,Pred,SuccID,LineNr),
    tcltk_check_state_sequence_from_file_aux(File,LineNr,Stream,SuccID).
process_line(_File,LineNr,_Stream,Codes,_ID) :-
    atom_codes(Atom,Codes),
    add_error(tcltk_check_state_sequence_from_file,'Could not parse state predicate: ',Atom),
    add_error(tcltk_check_state_sequence_from_file,'Line: ',LineNr),
    %add_error(tcltk_check_state_sequence_from_file,'File: ',File),
    fail.

insert_nl_prefix(N,Acc,Acc) :- N =< 1,!.
insert_nl_prefix(N,Acc,[10,13|Res]) :- N1 is N-1, insert_nl_prefix(N1,Acc,Res).

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2]).
:- use_module(state_space,[visited_expression/2,
    max_reached_for_node/1, max_reached_or_timeout_for_node/1, time_out_for_node/1]).
:- use_module(library(codesio),[format_to_codes/3]).
:- use_module(eventhandling, [announce_event/1]).

find_successor_state(_ID,Pred,NewID,_) :-
  % transition(ID,ActionOpAsTerm,OpID,SuccID),
   user:tcltk_get_options(_),
   current_options(Options), % print(Options),nl,
   member( (_,Action,ActionOpAsTerm,NewID), Options),
   state_space:visited_expression(NewID,State),
   expand_const_and_vars_to_full_store(State,EState),
   % print(try(NewID,State)),nl,
   add_prob_deferred_set_elements_to_store(EState,DEState,all),
   announce_event(start_solving),
   b_interpreter:b_test_boolean_expression_wf(Pred,[],DEState), % deferred
   announce_event(end_solving),
   print(successor_found(NewID)),nl,
   print(' --> '),my_print_event(ActionOpAsTerm,_),nl,nl,
   my_perform_action_string(Action,ActionOpAsTerm,NewID),
   !.
find_successor_state(ID,Pred,NewID,_LineNr) :-
   max_reached_or_timeout_for_node(ID), % we have not computed all options
   get_op_name_for_state(ID,OpName), print(try_providing_custom_predicate(OpName)),nl,
   user:tcltk_add_user_executed_operation_typed(OpName,Pred,NewID), % The Pred can only talk about parameters and the state before; thus this only works for setup_constants and initialise_machine
   print(successor_found(NewID)),nl,
   print(' --> '),print(OpName),nl,
   !.
find_successor_state(ID,Pred,_,LineNr) :-
   % we could try and execute operation by predicate here; in case max_reached_for_node(ID) or  time_out_for_node(ID)
   format('State ID = ~w; line number = ~w~n',[ID,LineNr]),
   (max_reached_for_node(ID)
    -> format_to_codes('Try increasing the MAX_OPERATIONS preference. No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
     ; time_out_for_node(ID) -> format_to_codes('TIME-OUT occurred. No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
     ; format_to_codes('No successor found satisfying predicate at step ~w of replay: ',[LineNr],Codes)
   ),
   atom_codes(Msg,Codes),
   translate:translate_bexpression(Pred,TP),
   add_error(tcltk_check_state_sequence_from_file,Msg,TP),
   fail.

:- use_module(bmachine,[b_machine_has_constants_or_properties/0]).
get_op_name_for_state(root,OP) :- !,
 (b_machine_has_constants_or_properties -> OP='$setup_constants' ; OP='$initialise_machine').
get_op_name_for_state(ID,'$initialise_machine') :-
   visited_expression(ID,concrete_constants(_)),!.
%get_op_name_for_state(ID,OpName) :- bmachine:b_top_level_operation(OpName). % not supported yet

% --------------------------------

check_default_trace_for_specfile(File) :- get_default_trace_file(File,TFile),
   tcltk_check_sequence_from_file(TFile).

get_default_trace_file(File,TraceFile) :-
   split_filename(File,FN,_Ext),
   string_concatenate(FN,'.trace',TraceFile).

:- use_module(tools_printing,[print_error/1]).
tcltk_check_sequence_from_file(File) :-
   read_trace_file(File,MachineName,Trace),
   length(Trace,Len),
   (b_mode -> b_machine_name(OurName) ; true),
   ((OurName=MachineName
    ; MachineName = 'dummy(uses)') % old trace files have these
     -> print_message('Checking file for machine: '), print_message(OurName)
      ; print_error('### Trace File has been generated for different machine: '),
        print_error(MachineName),
        print_error('### Our machine: '), print_error(OurName)
   ),
   print_message('Starting trace check'),
   print_message('  -> length of trace: '), print_message(Len),
   statistics(walltime,[Time1,_]),
   (perform_sequence_of_operations(Trace,0)
     -> nl,print_message(''), print_message('Trace checking successful !')
     ;  nl,print_message(''), print_message('### Trace was not possible !!!'), %read(XX),
         add_error(trace_checking_fails,'Trace Checking was not successful: ',(File,MachineName))
        %% ,throw(trace_checking_failed(MachineName))  % comment in if you want to stop immediately upon such an error
   ),
   statistics(walltime,[Time2,_]), Time is Time2-Time1,
   print('Walltime: '), print(Time), print(' ms'),nl,
   (specfile:csp_mode -> print_trace_as_fdr_check ; true),

   (junit_mode(_) -> (get_error(trace_checking,Errors) -> V=error([Errors])
                                           ; otherwise -> V=pass),
                     create_and_print_junit_result(File,'Trace_checking', MachineName, Time, V)
   ; otherwise -> true).

read_trace_file(File,MachineName,Trace) :-
   print_message('opening Trace file: '), print_message(File),
   my_see(File),
   parse_trace_file(MachineName,Trace),!,
   seen,
   print(done),nl.

my_see(File) :- on_exception(error(existence_error(_,_),_),
                   see(File), add_error_fail(my_see,'File does not exist: ',File)).

parse_trace_file(MName,Trace) :-
   read(Term),!,
   ((Term = end_of_file) -> (Trace = [])
     ; (Term = machine(MName)
        -> (Trace = T)
        ;  (Trace = [Term|T])
        ),
        parse_trace_file(MName,T)
    ).


is_a_reset_operation(start_cspm_MAIN).
is_a_reset_operation(start_csp_MAIN).
is_a_reset_operation(Op) :- functor(Op,'$setup_constants',_).
is_a_reset_operation(Op) :- functor(Op,'$initialise_machine',_),
                            b_mode, \+ b_machine_has_constants_or_properties.

:- use_module(bmachine, [b_machine_has_constants_or_properties/0]).
perform_initialisation :- b_mode, b_machine_has_constants_or_properties,!,
   perform_sequence_of_operations(['$setup_constants','$initialise_machine'],0).
perform_initialisation :- perform_sequence_of_operations(['$initialise_machine'],0).


perform_sequence_of_operations([],Ident) :- ident(Ident),print('SUCCESS'),nl.
perform_sequence_of_operations([Op|T],Ident) :-
   perform_one_operation_in_sequence(Op,Ident),
   Ident1 is Ident+1,
   perform_sequence_of_operations(T,Ident1).

:- use_module(store,[lookup_value_for_existing_id/3]).

perform_one_operation_in_sequence('$check_value'(ID,VAL),Ident) :- !,
   ident(Ident), format('checking value of ~w ',[ID]),
   state_space:current_expression(_,State),
   expand_const_and_vars_to_full_store(State,EState),
   lookup_value_for_existing_id(ID,EState,StoredVal),
   translate:print_bvalue(StoredVal),
   (unify(StoredVal,VAL) -> print(' ok'),nl
     ; print(' failed'),nl, %print(StoredVal),nl, print(VAL),nl,
       fail).
perform_one_operation_in_sequence(Op,Ident) :-
   ident(Ident),print('<- '),print(Ident), print(': '),
   my_print_event(Op,OpStr),
   print(' :: '),
   state_space:current_state_id(CurID), print(CurID),nl,
   ((functor(Op,'$initialise_machine',_),\+(state_space:current_expression(_,concrete_constants(_)))
      ; is_a_reset_operation(Op))
     ->  user:tcltk_goto_state(reset,root),
         user:tcltk_get_options(_)
     ;   true),
   convert_action(Op,COp),
   if(perform_single_operation(COp,Ident),true,
      (ident(Ident), print(' **** FAIL ****'),nl,
       %% trace, perform_single_operation(COp,Ident),
       fail)),
   ident(Ident), print(' | '), print(Ident), print(': '), print(OpStr),
   %functor(Op,OpName,_), print(OpName),
   %((OpName = '-->', arg(1,Op,A1), functor(A1,RealOpName,_)) -> print(RealOpName) ; true),
   print(' success -->'),
   state_space:current_state_id(NewCurID),
   statistics(walltime,[Tot,Delta]),
   format('~w  [~w (Delta ~w) ms]~n',[NewCurID,Tot,Delta]).

my_print_event(Op,OpStrTA) :-
        preferences:temporary_set_preference(expand_avl_upto,4,CHNG),
        translate_event(Op,OpStr),
        truncate_atom(OpStr,100,OpStrTA),
        preferences:reset_temporary_preference(expand_avl_upto,CHNG),
        write(OpStrTA).

/* convert global constants such as b with S={a,b,c} into fd(2,'S') */
convert_action(V,R) :- var(V),!,R=V.
convert_action(io(V,Ch),R) :- !, strip_dots(V,SV),R=io(SV,Ch,_SRCSPAN).
convert_action(io(V,Ch,_),R) :- !, strip_dots(V,SV),R=io(SV,Ch,_SRCSPAN).
convert_action(tick(_SRCSPAN),R) :- !, R = tick(_).
convert_action(X,Res) :- nonvar(X),X=..[F|Args],!,
    l_convert_term(Args,CArgs),Res=..[F|CArgs].
convert_action(X,X).

convert_term(V,R) :- var(V),!,R=V.
convert_term(closure(A,B,C),R) :- !, R=closure(A,B,C). % we may mess with identifiers; see test 1575
convert_term(term(bool(0)),R) :- !, R=pred_false /* bool_false */.
convert_term(term(bool(1)),R) :- !, R=pred_true /* bool_true */.
convert_term(bool_true,pred_true /* bool_true */) :- !.
convert_term(bool_false,pred_false /* bool_false */) :- !.
convert_term(string(S),R) :- !, R=string(S).
convert_term(X,Representation) :- atomic(X), lookup_global_constant(X,Representation),!.
convert_term(X,Res) :- nonvar(X),X=..[F|Args],!,
    l_convert_term(Args,CArgs),Res=..[F|CArgs].
convert_term(X,X).

l_convert_term([],[]).
l_convert_term([H|T],[CH|CT]) :- convert_term(H,CH), l_convert_term(T,CT).

/* newer version of ProB no longer has dot() constructor */
strip_dots([],[]).
strip_dots([H|T],[X|ST]) :- (H=dot(X) -> true ; X=H), strip_dots(T,ST).
strip_dots(tail_in(X),[in(XR)]) :-
   (X=dotTuple(R) -> XR=tuple(R) ; XR = X). /* we also no longer use tail_in */

/* argument is a structured operation and performs the action and
  computes the possible options for the next state */


perform_single_operation(ActionAsTerm,Ident) :-
   % tools_printing:print_term_summary(perform_single_op(ActionAsTerm)),
   state_space:current_expression(_CurID,State),
   current_options(Options),
   ( if( (member( (_Id,Action,ActionOpAsTerm,NewID), Options),
          unify_action_as_term(ActionAsTerm,ActionOpAsTerm,State)),
          perform_action_tclk(Action,ActionOpAsTerm,NewID,Ident),
         ( perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,State),
           perform_action_tclk(Action,_AlternateActionOpAsTerm,NewID,Ident)
         ))
    ;
        single_operation_skip(ActionAsTerm,Ident)
   ).
perform_single_operation(ActionAsTerm,Ident) :-
   current_options(Options),
   member( (_Id,PActionS,'*permute*',NewID), Options),  /* permute action from symmetry reduction */
   perform_action_tclk(PActionS,'*permute*',NewID,Ident), !,
   ident(Ident), print(' | ** PERMUTING **'),nl,
   perform_single_operation(ActionAsTerm,Ident).
perform_single_operation(ActionAsTerm,Ident) :-
   %%user:current_expression(CurID,State),
   \+ functor(ActionAsTerm,tau,1),
   current_options(Options), %print(options(Options)),nl,
   member( (_Id,AX,tau(X),NewID), Options),
   ident(Ident), print(' | ** SKIPPING ADDITIONAL TAU in SPEC. **'),nl,
   perform_action_tclk(AX,tau(X),NewID,Ident),
   perform_single_operation(ActionAsTerm,Ident).
%perform_single_operation(ActionAsTerm,Ident) :-
%   ident(Ident),print(unsuccessful(ActionAsTerm)),nl,fail.


perform_action_tclk(Action,ActionOpAsTerm,NewID,_Ident) :-
    my_perform_action_string(Action,ActionOpAsTerm,NewID),
    user:tcltk_get_options(_).
perform_action_tclk(_Action,_ActionOpAsTerm,_NewID,Ident) :- % undo action
    state_space:current_state_id(FromID),
    user:tcltk_backtrack, user:tcltk_get_options(_),
    state_space:current_state_id(ToID),
    ident(Ident),
    print(backtrack_from_to(FromID,ToID)),nl,
    fail.

% a more flexible version: allow string representations of ProB to evolve compared to when test suites where stored
my_perform_action_string(Action,ActionOpAsTerm,NewID) :- (ground(ActionOpAsTerm) ; ground(NewID)),!,
    user:tcltk_perform_action_string(Action,ActionOpAsTerm,NewID),
   true.% (ActionStr2=Action -> true ; print(string_mismatch(ActionStr2,Action)),nl).
my_perform_action_string(Action,ActionOpAsTerm,NewID) :-
    user:tcltk_perform_action_string(Action,ActionOpAsTerm,NewID).

:- use_module(library(avl)).
unify_action_as_term(A,A,_) :- !.
unify_action_as_term('-->'(Op,Results),'-->'(Op2,Results2),State) :- !,
    unify_action_as_term(Op,Op2,State),
    l_unify(Results,Results2).
unify_action_as_term(A,B,State) :- functor(A,F,N), functor(B,F2,N2),
    unify_functor(F,F2,State),
    (N=N2 -> true
     ; format('*** SAME EVENT WITH VARYING ARITY: ~w : ~w vs ~w arguments ***~n',[F,N,N2]),
       % Probably due to show_eventb_any_arguments differently set ?
       fail),
    A=..[F|AA], B=..[F2|BB],
    l_unify(AA,BB).
l_unify([],[]).
l_unify([H|T],[I|S]) :- unify(H,I), l_unify(T,S).

:- use_module(specfile,[animation_minor_mode/1, rename_event_to_internal_name/2]).
unify_functor('initialise_machine','$initialise_machine',_) :- !. % old-style format
unify_functor('setup_constants','$setup_constants',root) :- !. % old-style format
unify_functor(InternalName,OpName,_) :- animation_minor_mode(tla), !,
   (InternalName=OpName -> true
     ; rename_event_to_internal_name(OpName,InternalName)).
unify_functor(X,X,_).

%:- use_module(custom_explicit_sets,[equal_explicit_sets/4]).
:- use_module(debug).
unify(A,A) :- !.
unify(A,avl_set(B)) :- %print(exp(A,B)),nl,
    expand_explicit_set(A,AA),
    %nl,translate:print_bvalue(AA),nl, translate:print_bvalue(avl_set(B)),nl, % print(chck(A,AA,B)),nl,
   AA = avl_set(B),!.
%unify(A,B) :- print(unify(A,B)),nl,fail.
unify(C1,C2) :- is_a_set(C1),(C2=closure(_P,_T,_B) ; C2=global_set(_)),
   %print(unify_closure(C1,C2)),nl,
   (C1=[_|_] -> custom_explicit_sets:convert_to_avl(C1,C11) ; C11=C1), % TO DO: replace by proper expansion predicate
   (unify_explicit_sets(C11,C2)
    -> true
     ; print(ko_unify_closure),nl, translate:print_bvalue(C11),nl, translate:print_bvalue(C2),nl,
       %terms:term_hash((C1,C2),H), print(hash(H)),nl, (H=140674321 -> trace, unify_explicit_sets(C11,C2)),
       fail),!.
 %avl_domain(B,R), print(chck(A,B,R)),nl,l_unify(A,R).
unify(C1,C2) :- %tools_printing:print_term_summary(fail(C1,C2)),nl,
  debug_println(9,unify_failed(C1,C2)),fail.

is_a_set(closure(_,_,_)).
is_a_set([]).
is_a_set([_|_]).
is_a_set(global_set(_)).
is_a_set(avl_set(_)).

expand_explicit_set([H|T],AA) :- !, custom_explicit_sets:convert_to_avl([H|T],AA).
expand_explicit_set(A,AA) :-
   on_enumeration_warning(custom_explicit_sets:try_expand_and_convert_to_avl_if_smaller_than(A,AA,200), fail).
unify_explicit_sets(C1,C2) :-
   ALLOW=no_expansion, %(C1 = avl_set(_) -> ALLOW=allow_expansion ; ALLOW=no_expansion),
   on_enumeration_warning(custom_explicit_sets:equal_explicit_sets4(C1,C2,ALLOW,no_wf_available), fail).

:- use_module(preferences,[preference/2]).


perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,root) :-
   functor(ActionAsTerm,'setup_constants',_), !, % rename to new format
   nl,print('DEPRECATED TRACE FILE'),nl,nl,nl,
   ActionAsTerm =.. [_|Args], ActionAsTerm2 =..['$setup_constants'|Args],
   perform_single_operation_retry(Options,ActionAsTerm2,Action,NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,State) :-
   (State=root ; State=concrete_constants(_)),
   functor(ActionAsTerm,'initialise_machine',_), !, % rename to new format
   nl,print('DEPRECATED TRACE FILE'),nl,nl,nl,
   ActionAsTerm =.. [_|Args], ActionAsTerm2 =..['$initialise_machine'|Args],
   perform_single_operation_retry(Options,ActionAsTerm2,Action,NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID,_State) :-
   perform_single_operation_retry(Options,ActionAsTerm,Action,NewID).


% check that parameters provided for setup_constants in trace file can be found in State
%check_initialised_args([],_).
%check_initialised_args([V|T],State) :-
%   select(bind(_,V2),State,Rest),
%   unify(V,V2),
%   %print('SELECTED: '),translate:print_bvalue(V2),nl,
%   !,
%   check_initialised_args(T,Rest).
%check_initialised_args([V|T],State) :-
%    print('MISMATCH'),nl, %print(mismatch(V,C)),nl,
%    print('COULD NOT FIND: '), translate:print_bvalue(V),nl,
%    tools_printing:print_term_summary(V),nl,
%    %print(V),nl, print(State),nl,
%    V=closure(_,_,_), memberchk(bind(_,closure(_,_,_)),State),
%    check_initialised_args(T,State).

perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,'$setup_constants',_), !,
   perform_alternative_op_with_same_functor(Options, ['$setup_constants' /* ,'$partial_setup_constants' */],
                                            'set up constants', 'initialisation of constants',
                                            ActionAsTerm, Action, NewID),
   %visited_expression(NewID,concrete_constants(C)), ActionAsTerm =.. [_|Args],
   % check that state NewID corresponds to parameters provided for setup_constants in trace file
   %print(check_args),nl, check_initialised_args(Args,C), print('argsok'),nl.  % TO DO: try and enable this for all tests
   true.
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,'$initialise_machine',_),!,
   perform_alternative_op_with_same_functor(Options, ['$initialise_machine'],
                                            'initialise', 'initialisation',
                                            ActionAsTerm, Action, NewID),
     true. %visited_expression(NewID,S), (S=const_and_vars(_,VS) -> true ; VS=S), ActionAsTerm =.. [_|Args],
     %check_initialised_args(Args,VS), print('initargsok'),nl.
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,tau,_), arg(1,ActionAsTerm,Arg), nonvar(Arg),!,
   perform_alternative_op_with_same_functor(Options, [tau],
                                            tau, tau,
                                            ActionAsTerm, Action, NewID).
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,F,0), preference(show_eventb_any_arguments,true),
    /* then allow also the same operation name but with more arguments */
   member( (_Id,Action,ActionAsTerm2,NewID), Options),
   functor(ActionAsTerm2,F,FArity), FArity>0,
   print_message('Allowing extra Event-B style ANY arguments.').
perform_single_operation_retry(Options,ActionAsTerm,Action,NewID) :-
   functor(ActionAsTerm,F,N),N>0, preference(show_eventb_any_arguments,false),
    /* then allow also the same operation name but with more arguments */
   member( (_Id,Action,F,NewID), Options),
   print_message('Allowing operation without Event-B style ANY arguments.').
%perform_single_operation_retry(Options,_ActionsAsTerm,_Action,_NewID) :-
%   print_message('No more options: '), print(Options),nl,fail.

perform_alternative_op_with_same_functor(Options,Functors,ActionMsg,AltMsg,ActionAsTerm,Action,NewID) :-
    \+ member( (_,_,ActionAsTerm,_), Options),
    ajoin(['Could not ', ActionMsg, ' with parameters from trace file.'], Msg1),
    ajoin(['Will attempt any possible ', AltMsg, '.'], Msg2),
    print_message(Msg1), print_message(Msg2),
    %print(ActionAsTerm),nl, print(Options),nl,nl,
    member( (_,ActionStr,MachineActionAsTerm,NewID), Options),
    functor(MachineActionAsTerm,Functor,_),
    member(Functor,Functors),
    (Action=ActionStr -> true ; print(cannot_match(Action,ActionStr)),nl,fail).

single_operation_skip(tau(_),Ident) :-
   ident(Ident), print(' | *** TRYING TO SKIP TAU IN TRACE FILE'),nl.


ident(0).
ident(N) :- N>0, print(' - '), N1 is N-1, ident(N1).


/* ------------------------------------------------------------------ */

:- use_module(state_space,[get_action_trace/1]).
print_trace_as_fdr_check :-
   print('-- Trace Check Generated by ProB:'),nl,
   get_action_trace(T),
   reverse(T,RT),
   print('PROB_TEST_TRACE = '),print_fdr(RT),
   nl,
   print('assert MAIN [T= PROB_TEST_TRACE'),nl.

print_fdr([]) :- print('STOP'),nl.
print_fdr([jump|T]) :- print_fdr(T).
print_fdr([action(Str,Term)|T]) :-
   (Term = i(_)
      -> (nl,print('   ( STOP /\\ '))
    ; (Term = tick(_)
       -> print('SKIP ; ')
       ;  ((Term = tau(_) ; Term = 'start_cspm_MAIN')
           -> true
           ;  (print(Str), print(' -> '))
          )
   )),
   print_fdr(T),
   (Term = i(_)
      -> print('   )')
      ;  true
   ).
