% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(clpfd_interface,[post_constraint/1,  post_constraint/2, post_constraint2/2,
                            force_post_constraint/1,
                            clpfd_neq_expr/2, clpfd_eq_expr/2, clpfd_eq_expr_optimized/2,
                            clpfd_eq_div/3, clpfd_eq_fdiv/3,
                            clpfd_eq_guarded_div/3, clpfd_eq_guarded_fdiv/3,
                            clpfd_eq/2, clpfd_neq/2,
                            clpfd_geq2/3,
                            clpfd_geq/3, clpfd_gt/3, clpfd_leq/3, clpfd_lt/3,
                            clpfd_leq_expr/2, clpfd_lt_expr/2,
                            clpfd_inrange/3, clpfd_inrange/4,
                            clpfd_not_inrange/3, clpfd_not_in_non_empty_range/3,
                            clpfd_sum/2,
                            clpfd_inlist/2, clpfd_not_inlist/2, clpfd_reify_inlist/3, force_clpfd_inlist/2,
                            clpfd_minimum/2, clpfd_maximum/2,
                            clpfd_if_then_else/4,
                            clpfd_domain/3, clpfd_size/2,
                            clpfd_max_bounded/1,
                            clpfd_can_match/2,
                            clpfd_check_geq_nr/2,
                            clpfd_alldifferent/1,
                            %clpfd_labeling/1, % not used at the moment
                            clpfd_degree/2,
                            clpfd_in_domain/1, clpfd_labeling/2,
                            %clpfd_reverse_labeling/1, clpfd_reverse_labeling/2,
                            clpfd_get_next_variable_to_label/2, clpfd_get_next_variable_to_label/3,
                            clpfd_get_next_variable_to_label_ffc/3,
                            is_64_bit_system/0, clpfd_overflow_error_message/0, clpfd_overflow_warning_message/0,
                            catch_clpfd_overflow_call1/1, catch_clpfd_overflow_call2/2,
                            computable_arith_expression/1,
                            clpfd_randomised_labeling/2, clpfd_randomised_in_domain/1]).

:- use_module(library(clpfd)).
 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(\+ environ(disable_chr, true)).
:- use_module(chrsrc(chr_integer_inequality)).
:- endif.


:- if((current_prolog_flag(version_data,sicstus(4,X,Y,_,_)),X<3,Y<3)).
% -------- patch provided by SICS -----------
% solves lost solution in CLP(FD) when calling:
%     when(nonvar(X),member(Y,[1,2])), X in 1..2, X in 2..3.
% patch provided 23rd June 2012 for SICS 4.2.1
%  bug cause ProB to miss solutions in e.g. following set comprehension:
% {xx | xx:{1|->TRUE,2|->TRUE,2|->FALSE} & (xx  :  {(2|->TRUE),(2|->FALSE),(3|->TRUE)} or xx  :  {(1|->TRUE),(1|->FALSE),(2|->TRUE),(2|->FALSE),(3|->TRUE),(3|->FALSE)})}
:- set_prolog_flag(redefine_warnings,proceed).
clpfd:propagate_interval_chk(Pruned, Min, Max) :-
  clpfd:'$fd_in_interval'(Pruned, Min, Max, 1), !, % 20120623
  clpfd:'$fd_evaluate_indexical'(RC, Global),
  clpfd:evaluate(RC, Global).
clpfd:propagate_interval_chk(Pruned, Min, Max) :-
  clpfd:arg_attribute(Pruned, _, in(Pruned,Min..Max), 1), !,
  fail.
clpfd:propagate_interval_chk(_, _, _).

clpfd:prune_and_propagate_chk(Pruned, Set) :-
  clpfd:'$fd_in_set'(Pruned, Set, 1), !, % 20120623
  clpfd:'$fd_evaluate_indexical'(RC, Global),
  clpfd:evaluate(RC, Global).
clpfd:prune_and_propagate_chk(Pruned, Set) :-
  clpfd:arg_attribute(Pruned, _, in_set(Pruned,Set), 1), !,
  fail.
clpfd:prune_and_propagate_chk(_, _).
%:- set_prolog_flag(redefine_warnings,on).
% ---------- end patch ---------------
:- endif.


%:- use_module(typechecker).
:- use_module(tools).
:- use_module(preferences).
:- use_module(library(timeout)).
:- use_module(error_manager).
:- use_module(self_check).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'Provide interface to CLP(FD), with Prolog fallback in case of errors or when CLP(FD) turned off.').

:- meta_predicate post_constraint(0).
:- meta_predicate post_constraint(0,0).
:- meta_predicate post_constraint2(0,*).
:- meta_predicate time_out_constraint(0,0).


% Note: clpfd_eq is sometimes called with arithmetic expressions (treatment of plus/minus/...)
% does not do the computable_arith_expression check
clpfd_eq(E1,E2) :- simple(E1), simple(E2), % was (before 29.8.2014): number(E1),number(E2),
   !,
   E1=E2.
clpfd_eq(E1,E2) :-
   (preferences:preference(use_chr_solver,true) -> chr_eq(E1,E2) ; true),
   post_constraint( E1 #= E2 ).

clpfd_neq(E1,E2) :- number(E1),!,
   (number(E2) -> E1\=E2 ; post_constraint( E1 #\= E2),dif(E1,E2)).  % no need to post dif constraint if one arg is number in clpfd mode
clpfd_neq(E1,E2) :- number(E2),!, % no need to post dif constraint in clpfd mode
   post_constraint( E1 #\= E2, dif(E1,E2)).
clpfd_neq(E1,E2) :-
   % Note: | ?- X #\= Y, Y #=X.  --> does not fail; hence we also execute dif(E1,E2) in CLPFD mode !
   % We might be able to skip calling dif, if we enable the CHR solver by default
   % Let's try it: only set up dif if CHR is disabled!
   (preferences:preference(use_chr_solver,true) -> chr_neq(E1,E2) ; dif(E1,E2)),
   post_constraint( E1 #\= E2).
  % clpfd_domain(E1,D1,D2), clpfd_domain(E2,B1,B2), print(dom(D1:D2,B1:B2)),nl.

% set up a dif co-routine only if necessary; assumption: E1,E2 are either number or variable
dif_when_necesssary(E1,E2) :-  %print(dif_when_necesssary(E1,E2)),nl,
   var(E1),var(E2),!, dif(E1,E2).
dif_when_necesssary(_,_).

% this will force Y to be nonzero
clpfd_eq_div(E1,X,Y) :- post_constraint( E1 #= X/Y, prolog_eq(E1,X//Y)), % truncated division  -1 // 4 = 0
   propagate_div_eq_eq(E1,X,Y).
clpfd_eq_fdiv(E1,X,Y) :- post_constraint( E1 #= X div Y, prolog_eq(E1,X div Y)). % floored division -1 div 4 = -1

% certain == propagation rules not done by CLPFD
propagate_div_eq_eq(E1,X,Y) :-
   ((var(E1),X==Y) -> print(setting1(E1,X,Y)),nl,E1=1 ; true).  %  X/X can only be 1 or undefined when X=0
   % we could check if E1==X --> Y=1 if X /= 0

% this waits until we are sure Y is not zero
clpfd_eq_guarded_div(E1,X,Y) :-
    post_constraint3((E1 #= X/YY,MY in 0..1, Y #= MY*YY), prolog_eq(E1,X//Y),Posted), % Y can either be 0 or equal to YY
    % this is better than the previous solution:
   % post_constraint(Y#\=0 #=> YY#=Y, prolog_eq(Y,YY)).
   % however, x:44..77 & E : 1..5555 & E = x/y & y:0..4 only restricts E to 1..77, not 11..77
   % what is still missing is propagating upper bounds of Y to YY
   % Y in 0..4, M in 0..1, YY in 1..77, Y #=M*YY
   % with this propagation below, we at least obtain propagation for x:44..77 & E : 1..5555 & y:0..4 & E = x/y
   % what we need is continously propagating the bounds from Y to YY (but not vice versa for 0)
   % maybe we can achieve this using indexicals ??
   (Posted==true
     -> fd_max(Y,YUp), (number(YUp) -> YY #=< YUp ; true),
        fd_min(Y,YLow), (number(YLow) -> YY #>= YLow ; true),
        propagate_div_eq_eq(E1,X,Y)
     ;  true).

% now a similar solution for floored division
clpfd_eq_guarded_fdiv(E1,X,Y) :-
    post_constraint3((E1 #= X div YY,MY in 0..1, Y #= MY*YY), prolog_eq(E1,X div Y),_Posted).
    % TO DO: add propagation like above or using indexicals

% do we need to instantiate YY if Y becomes 0 ??
%:- block copy_divisor(-,?).
%copy_divisor(Y,Y).



% we check whether we have simple expressions of the form X = Y OP Z
% in this case it is better to call ProB's implementations, as they have stronger propagation
%  (e.g., 0 = X-Y  =>  X=Y   or   1 = X-Y  =>  X/=Y)
clpfd_eq_expr_optimized(CLPFD_Expr1,CLPFD_Expr2) :-
   (simple(CLPFD_Expr1) -> call_eq_expr1(CLPFD_Expr2,CLPFD_Expr1)
    ; simple(CLPFD_Expr2) -> call_eq_expr2(CLPFD_Expr1,CLPFD_Expr2)
    ; clpfd_eq_expr_nonsimple(CLPFD_Expr1,CLPFD_Expr2)).

:- use_module(kernel_dif).
call_eq_expr1(CLPFD_Expr2,CLPFD_Expr1) :-
    (simple(CLPFD_Expr2) ->  %print(unify(CLPFD_Expr2,CLPFD_Expr1)),nl, translate:print_clpfd_variable(CLPFD_Expr2),nl, translate:print_clpfd_variable(CLPFD_Expr1),nl,
       % first check if we have a pending dif; unification could trigger CLPFD enumeration
            ((var(CLPFD_Expr1),kernel_dif:frozen_dif(CLPFD_Expr2,CLPFD_Expr1)) -> /* print(dif),nl,*/ fail ; true),
            CLPFD_Expr2=CLPFD_Expr1
          ; call_eq_expr2(CLPFD_Expr2,CLPFD_Expr1)).

% first arg not simple, second arg simple
call_eq_expr2(CLPFD_Expr2,CLPFD_Expr1) :- CLPFD_Expr2 = -(A,B),simple(A),simple(B),!,
    %kernel_objects:int_minus(int(A),int(B),int(CLPFD_Expr1)). % has special propagation rules
    (number(A),number(B) -> CLPFD_Expr1 is CLPFD_Expr2
     ; clpfd_eq_expr_nonsimple(CLPFD_Expr2,CLPFD_Expr1),
       propagate_zero(A,B,CLPFD_Expr1)).
call_eq_expr2(CLPFD_Expr2,CLPFD_Expr1) :- CLPFD_Expr2 = +(A,B),simple(A),simple(B),!,
    (number(A),number(B) -> CLPFD_Expr1 is CLPFD_Expr2
      ; clpfd_eq_expr_nonsimple(CLPFD_Expr1,CLPFD_Expr2),
        propagate_zero(CLPFD_Expr1,A,B)).
    %kernel_objects:int_plus(int(A),int(B),int(CLPFD_Expr1)). % has special propagation rules
call_eq_expr2(CLPFD_Expr2,CLPFD_Expr1) :-
    (computable_arith_expression(CLPFD_Expr2) -> CLPFD_Expr1 is CLPFD_Expr2
      ; clpfd_eq_expr_nonsimple(CLPFD_Expr1,CLPFD_Expr2)).

% a propagation rule not done by CLP(FD)
:- block propagate_zero(-,-,-).
propagate_zero(Sum,A,B) :- % print(prop_zero(Sum,A,B)),nl,
   (A==0 -> Sum=B ; B==0 -> Sum=A ; true).

% eq_expr takes expressions as arguments
clpfd_eq_expr(E1,E2) :- simple(E1),!,
   (simple(E2) -> E1=E2
    ; computable_arith_expression(E2) -> E1 is E2
    ; clpfd_eq_expr_nonsimple(E1,E2)).
clpfd_eq_expr(E1,E2) :- simple(E2),computable_arith_expression(E1),!, E2 is E1.
clpfd_eq_expr(E1,E2) :- clpfd_eq_expr_nonsimple(E1,E2).

clpfd_eq_expr_nonsimple(E1,E2) :-
    (preferences:preference(use_chr_solver,true) -> chr_eq(E1,E2) ; true),
    post_constraint( E1 #= E2, prolog_eq(E1,E2)).
%% post_constraint( E1 #= E2). % Used instead of line below because of serious bug in SICStus 4.1.3
                              % BUG: X*X #= 100, X #= 9.  succeeds !

:- assert_must_succeed((clpfd_interface:prolog_eq(X,2+1),X==3)).
:- assert_must_succeed((clpfd_interface:prolog_eq(4-1,X),X==3)).
:- assert_must_succeed((clpfd_interface:prolog_eq(5-2,(2*2)-Y),Y=1)).
:- assert_must_fail((clpfd_interface:prolog_eq(5-2,(2*2)-Y),Y=0)).
prolog_eq(A,B) :- simple(A),!, when(ground(B), A is B).
prolog_eq(A,B) :- simple(B),!, when(ground(A), B is A).
prolog_eq(A,B) :- when((ground(A),ground(B)),=:=(A,B)).

computable_arith_expression(X) :- number(X),!.
computable_arith_expression(X) :- var(X),!,fail.
computable_arith_expression(-(X)) :- computable_arith_expression(X).
computable_arith_expression(+(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).
computable_arith_expression(-(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).
computable_arith_expression(*(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).

clpfd_neq_expr(E1,E2) :-  number(E1),number(E2),!,E1\=E2.
clpfd_neq_expr(E1,E2) :- %print(neq_expr(E1,E2)),nl,
   var(E1), var(E2), !,
   (preferences:preference(use_chr_solver,true) -> chr_neq(E1,E2) ; dif(E1,E2)),
   post_constraint( E1 #\= E2 ). % does not matter if posted
   % Note: | ?- X #\= Y, Y #=X.  --> does not fail; hence we also execute dif(E1,E2) in CLPFD mode !
clpfd_neq_expr(E1,E2) :-    % print(neq_expr(E1,E2)),nl,
   post_constraint3( E1 #\= E2, prolog_neq(E1,E2), Posted),
   (Posted=true -> my_dif(E1,E2) ; true).

:- assert_must_succeed((clpfd_interface:prolog_neq(X,2+1),X=2)).
:- assert_must_fail((clpfd_interface:prolog_neq(X,2+1),X=3)).
:- assert_must_succeed((clpfd_interface:prolog_neq(3,X),X=2)).
:- assert_must_succeed((clpfd_interface:prolog_neq(5-2,(2*2)-Y),Y=2)).
prolog_neq(A,B) :- simple(A),simple(B),!, dif(A,B).
prolog_neq(A,B) :- A \== B, when((ground(A),ground(B)),=\=(A,B)).

my_dif(A,B) :- (simple(A);simple(B)),!. % do we need to setup things like dif(X,0) ?
my_dif(A,B) :- comm_assoc_dif(A,B).

% note: here we do not block, we use \== now rather than waiting on variables to be bound
comm_assoc_dif(A,B) :- var(A),!, A \== B.
comm_assoc_dif(A,B) :- var(B),!, A \== B.
comm_assoc_dif(*(A1,A2),*(B1,B2)) :- !,
    extract_factors2(A1,A2,FA,[]), sort(FA,SFA),
    extract_factors2(B1,B2,FB,[]), sort(FB,SFB),
    %print(comm_assoc_dif_mul(SFA,SFB)),nl,
    SFA \== SFB.
comm_assoc_dif(+(A1,A2),+(B1,B2)) :- !,
    extract_terms2(A1,A2,FA,[]), sort(FA,SFA),
    extract_terms2(B1,B2,FB,[]), sort(FB,SFB),
    %print(comm_assoc_dif_sum(SFA,SFB)),nl,
    SFA \== SFB.
comm_assoc_dif(-(A1),-(B1)) :- !, comm_assoc_dif(A1,B1).
comm_assoc_dif(-(A1,A2),-(B1,B2)) :- !,
    (comm_assoc_dif(A1,B1) -> true ; comm_assoc_dif(A2,B2)).
% note : exponentiation and division are not handed directly to CLPFD at the moment, so no need to check here:
%comm_assoc_dif(**(A1,A2),**(B1,B2)) :- !,  (comm_assoc_dif(A1,B1) -> true ; comm_assoc_dif(A2,B2)).
comm_assoc_dif(A,B) :- A \== B. %  ,((nonvar(A),nonvar(B)) -> nl,print(cadif(A,B)),nl ; true).

extract_factors2(A,B) -->  extract_factors(A), extract_factors(B).
extract_factors(A) --> {var(A)},!, [A].
extract_factors(*(A,B)) --> !, extract_factors(A), extract_factors(B).
extract_factors(+(A,B)) --> !, [sum(ST)], {extract_terms2(A,B,T,[]),sort(T,ST)}.
extract_factors(A) --> [A].

extract_terms2(A,B) -->  extract_terms(A), extract_terms(B).
extract_terms(A) --> {var(A)},!, [A].
extract_terms(+(A,B)) --> !, extract_terms(A), extract_terms(B).
extract_terms(*(A,B)) --> !, [mul(ST)], {extract_factors2(A,B,T,[]),sort(T,ST)}.
extract_terms(A) --> [A].


:- assert_must_succeed((clpfd_interface:clpfd_leq_expr(3,X+1),X=3)).
:- assert_must_fail((clpfd_interface:clpfd_leq_expr(2,X-1),X=2)).
clpfd_leq_expr(E1,E2) :- number(E1),number(E2), !, E1 =< E2.
clpfd_leq_expr(E1,E2) :-
    (preferences:preference(use_chr_solver,true) -> chr_leq(E1,E2) ; true),
    post_constraint( E1 #=< E2, prolog_leq(E1,E2)).
prolog_leq(A,B) :- when((ground(A),ground(B)),=<(A,B)).

:- assert_must_succeed((clpfd_interface:clpfd_lt_expr(3,X+1),X=3)).
:- assert_must_fail((clpfd_interface:clpfd_lt_expr(2,X-1),X=3)).
clpfd_lt_expr(E1,E2) :- number(E1),number(E2), !, E1 < E2.
clpfd_lt_expr(E1,E2) :-
    (preferences:preference(use_chr_solver,true) -> chr_lt(E1,E2) ; dif_when_necesssary(E1,E2)), % dif useful to detect that x<y & x=y is inconsisent; or x+1 < y+1 & x=y; note however x+1 < y & x=y is not detected as inconsistent without CHR
    post_constraint( E1 #< E2, prolog_lt(E1,E2)).
prolog_lt(A,B) :- when((ground(A),ground(B)),<(A,B)).

/* TO DO: do custom equalities div, plus, ...
mydiv(X, Y, Z) :-
   (Y#=0 #\/ (Y#\=0 #/\ X/Y #= Z)). */

% a version of geq which avoids posting constraints if possible
clpfd_geq2(E1,E2,Posted) :- number(E1),number(E2),!, E1 >= E2, Posted=true.
clpfd_geq2(E1,E2,Posted) :-
    (preferences:preference(use_chr_solver,true) -> chr_leq(E2,E1) ; true),
    post_constraint2( E1 #>= E2 , Posted).

clpfd_geq(E1,E2,Posted) :-
    (preferences:preference(use_chr_solver,true) -> chr_leq(E2,E1) ; true),
    post_constraint2( E1 #>= E2 , Posted).
clpfd_gt(E1,E2,Posted) :-
    (preferences:preference(use_chr_solver,true) -> chr_lt(E2,E1) ; dif_when_necesssary(E1,E2)),
    post_constraint2( E1 #> E2  , Posted).
clpfd_leq(E1,E2,Posted) :-
    (preferences:preference(use_chr_solver,true) -> chr_leq(E1,E2) ; true),
    post_constraint2( E1 #=< E2 , Posted).
clpfd_lt(E1,E2,Posted) :-
    (preferences:preference(use_chr_solver,true) -> chr_lt(E1,E2) ; dif_when_necesssary(E1,E2)),
    post_constraint2( E1 #< E2  , Posted).

clpfd_inrange(X,Low,Up) :- clpfd_inrange(X,Low,Up,_Posted).
clpfd_inrange(X,Low,Up,Posted) :-
  (Low==Up -> X=Low, Posted=true
   ;  (preferences:preference(use_chr_solver,true)
           -> chr_leq(Low,X), chr_leq(X,Up)
            ; true),
      ((number(Low),number(Up))
        -> Low =< Up, % otherwise interval empty
         (number(X) -> X >= Low, X =< Up, Posted=true %, Low =<Up
                     ; post_constraint2(X in Low..Up, Posted))  %  Note: X in 1..Y, Y in 0..3. fails ! Range needs to be constant
        ;  post_constraint2( (X #>=Low #/\ X #=< Up),Posted ) %  #/\ Low #=< Up should be implied
           %,translate:print_clpfd_variables([X,Low,Up]),nl
       )
    ).

% assumes Low =< Up !
clpfd_not_in_non_empty_range(X,Low,Up) :-
   post_constraint((X #<Low #\/ X #> Up),outside_range_blocking(X,Low,Up)).

:- block outside_range_blocking(-,?,?).
outside_range_blocking(X,Low,Up) :-  (X<Low ; X>Up).

% assumes Low =< Up !
clpfd_not_inrange(X,Low,Up) :- preferences:preference(use_clpfd_solver,false),
  !, % avoid building up constraint below
  not_in_range_blocking(X,Low,Up).
clpfd_not_inrange(X,Low,Up) :-
   % special treatment due to limitation of CLP(FD) disjunct: | ?- X #>10 #\/ X #>9.  ---> X in inf..sup
   post_constraint2(XLow #<=> (X #<Low), Posted1), Posted1==true,
   post_constraint2(UpX #<=> (Up #< X), Posted2), Posted2==true,
   post_constraint2(UpLow #<=> (Up #< Low), Posted3), Posted3==true,
   !,
   clpfd_not_inrange_prop(XLow,UpX,UpLow,X,Low,Up).
clpfd_not_inrange(X,Low,Up) :-
   force_post_constraint((X #<Low #\/ X #> Up #\/ Up #< Low),not_in_range_blocking(X,Low,Up)).

:- block clpfd_not_inrange_prop(-,-,-,?,?,?).
clpfd_not_inrange_prop(XLow,UpX,UpLow,_X,_Low,_Up) :- %print(unblock(XLow,UpX,UpLow,X,Low,Up)),nl,
   (XLow==1 ; UpX==1 ; UpLow==1),!.
clpfd_not_inrange_prop(XLow,UpX,UpLow,X,Low,Up) :- XLow==0,!,
   post_constraint2(Up #< max(X,Low),_), disjunct_reify(UpX,UpLow).
clpfd_not_inrange_prop(XLow,UpX,UpLow,X,Low,Up) :- UpX==0,!,
   post_constraint2(Low #> min(X,Up),_), disjunct_reify(XLow,UpLow).
clpfd_not_inrange_prop(XLow,UpX,_UpLow,_X,_Low,_Up) :-  % UpLow==0,
   % Anything more we can infer here ??
   disjunct_reify(XLow,UpX).

:- block disjunct_reify(-,-).
disjunct_reify(UpX,UpLow) :- (UpX==1 ; UpLow==1),!.
disjunct_reify(UpX,UpLow) :- (UpX==0 -> UpLow=1 ; UpX=1).


:- block not_in_range_blocking(-,?,?),not_in_range_blocking(?,-,?).
not_in_range_blocking(X,Y,Z) :-
     (X<Y -> true ; not_in_range_blocking2(X,Y,Z)).
:- block not_in_range_blocking2(?,?,-).
not_in_range_blocking2(X,Y,Z) :- (Z<Y -> true ; Z<X).

clpfd_sum(List,Sum) :-
     clpfd_sum(List,Sum,Posted),
     (Posted==true -> true
      ; sum_list(List,int(0),int(Sum))).
clpfd_sum(List,Sum,Posted) :- post_constraint2( sum(List,'#=',Sum),Posted).

% non-CLPFD backup, when posting fails or CLPFD turned off
:- block sum_list(-,?,?).
sum_list([],Acc,Acc).
sum_list([H|T],Acc,Res) :- kernel_objects:int_plus(int(H),Acc,NewAcc), sum_list(T,NewAcc,Res).


% assert that a variable is one of the numbers in the FDList; FDList must be numbers, cannot be unbound FD variable
clpfd_inlist(El,[X]) :- !, El=X.
clpfd_inlist(El,FDList) :-
   post_constraint( (list_to_fdset(FDList,FDSET), El in_set FDSET)).

% a version that always posts constraints independent of CLPFD preference
force_clpfd_inlist(El,[X]) :- !, El=X.
force_clpfd_inlist(El,FDList) :-
    list_to_fdset(FDList,FDSET), El in_set FDSET.

clpfd_not_inlist(El,FDList) :-
   post_constraint( (list_to_fdset(FDList,FDSET), fdset_complement(FDSET,C),El in_set C)).

clpfd_maximum(El,List) :- %print(maximum(El,List)),nl,
                          post_constraint( maximum(El,List)). %, print(done(El)).
clpfd_minimum(El,List) :- %print(minimum(El,List)),nl,
                          post_constraint( minimum(El,List)).


clpfd_if_then_else(PredRes,ThenValue,ElseValue,Value) :-
    % TO DO: catch overflows and revert to simpler treatment then
    element(Idx,[ThenValue,ElseValue],Value), % Idx=1 if Value=ThenValue; Idx=2 if Value=ElseValue
    prop_12(PredRes,Idx). % if PredRes=pred_true -> Idx must be 1, 2 otherwise

:- block prop_12(-,-).
prop_12(P,V12) :- var(P),!,(V12=1 -> P=pred_true ; P=pred_false).
prop_12(pred_true,1).
prop_12(pred_false,2).

/*
on_exception(E,list_to_fdset([0,8589934592],R),true)
E= error(domain_error(integer_list,[0,8589934592]),domain_error(list_to_fdset([0,8589934592],_A),1,integer_list,[0,8589934592]))
domain_error
*/


clpfd_reify_inlist(El,FDList,FDRes) :-
   post_constraint( (list_to_fdset(FDList,FDSET), (El in_set FDSET) #<=> FDRes)).



post_constraint(_) :- preferences:preference(use_clpfd_solver,false),!.
post_constraint(C) :-   % print(posting(C)),nl, %
  on_exception(error(Err,_),time_out_constraint(C,true), check_error(Err,C) ).
    % Note: this will only catch exceptions that occur while posting !
    %  CLPFD integer overflow exceptions can still happen after posting :-(


post_constraint2(_,Posted) :-
   preferences:preference(use_clpfd_solver,false),!,Posted=false.
post_constraint2(C,Posted) :-   % print(posting(C)),nl, %
  on_exception(error(Err,_),time_out_constraint(C,(Posted=false)), (check_error(Err,C),Posted=false) ),
  (Posted==false -> true ; Posted=true).


check_error(representation_error(ErrMsg),C) :-
                 is_clpfd_overflow_representation_error(ErrMsg),
                 !,
                 print_message('Ignoring constraint (integer overflow):'),
                 print_message(C).
check_error(domain_error(_Type,_Val),C) :- !,  print_message('Ignoring constraint (domain error):'),
                 % print(Type), print(' : '), print(Val), print(' : '),
                 print_message(C).
check_error(Err,C) :- print_message('Ignoring constraint (unknown error):'),
                 print(Err), print(' : '), print_message(C).


:- assert_must_succeed((print('Is 64-bit: '),(clpfd_interface:is_64_bit_system -> print('Yes') ; print('NO')),nl)).

is_64_bit_system :-
        Err = error(representation_error(_),_),
        on_exception(Err, _X #= 268435456, fail). % 268435456 = 2^28, will generate overflow in 32-bit system


clpfd_overflow_error_message :- clpf_overflow_msg(M),
  add_error(clpfd_overflow,M).
clpfd_overflow_warning_message :- clpf_overflow_msg(M),
  add_warning(clpfd_overflow,M).
%clpfd_overflow_print_message :- clpf_overflow_msg(M),write(user_error,M),nl(user_error).

clpf_overflow_msg(M) :- clpfd_interface:is_64_bit_system, !,
   M = 'A CLPFD integer overflow occurred.\nSet CLPFD preference to FALSE (Animation Preferences).'.
clpf_overflow_msg('A CLPFD integer overflow occurred.\nSet CLPFD preference to FALSE (Animation Preferences) or use a 64 bit version of ProB.').


% post_constraint3(CLPFD,PrologBackup,Posted)
post_constraint3(_,C,Posted) :- preferences:preference(use_clpfd_solver,false),!,Posted=false,call(C).
post_constraint3(C,PrologBackup,Posted) :-   % print(posting(C,PrologBackup)),nl,
  on_exception(error(Err,_),
               time_out_constraint((C,Posted=true),(Posted=false,PrologBackup)),
               (check_error(Err,C),Posted=false,call(PrologBackup)) ).

% post_constraint(CLPFD,PrologBackup)
post_constraint(_,C) :- preferences:preference(use_clpfd_solver,false),!,call(C).
post_constraint(C,PrologBackup) :-   % print(posting(C,PrologBackup)),nl,
  on_exception(error(Err,_),time_out_constraint(C,PrologBackup), (check_error(Err,C),call(PrologBackup)) ).

% no use_clpfd_solver check:
force_post_constraint(C) :- call(C).
force_post_constraint(C,PrologBackup) :-   % print(posting(C,PrologBackup)),nl,
  on_exception(error(Err,_),time_out_constraint(C,PrologBackup), (check_error(Err,C),call(PrologBackup)) ).

time_out_constraint(C,PrologBackup) :-  %% debug:new_pp(C,PP), %%
  %%preferences:preference(time_out,CurTO), TO is 800 + (CurTO//100), time_out(C,TO,T),
  time_out(C,1200,T),  %% debug:new_sol(C,PP), %%
  (T==time_out -> print_message('Timeout when posting constraint:'), print_message(C),
                  (preferences:preference(fail_if_clpfd_timeout,true)
                    -> print_message('Assuming unsatisfiable !'),fail
                    ;  call(PrologBackup))
               ; true).


:- assert_must_succeed((clpfd_interface:clpfd_domain(3,L,U), L==3, U==3)).
clpfd_domain(Var,Low,Up) :- var(Var),!,
    fd_min(Var,Low), fd_max(Var,Up).
%   used to check (preferences:preference(use_clpfd_solver,false) -> Low=inf,Up=sup     ;  fd_min(Var,Low), fd_max(Var,Up)    ).
% but can also be called when CLPFD false for global_set values
%used to call: fd_dom(Var, Low..Up)). fd_dom will sometimes return {1} \/ 2..3 or something like that !
clpfd_domain(X,X,X).

% succeed if we have a number or a CLPFD Var whose upper value is bounded
clpfd_max_bounded(Var) :- var(Var), !, fd_max(Var,Up), number(Up).
clpfd_max_bounded(N) :- number(N).

:- assert_must_succeed(clpfd_interface:clpfd_can_match(1,1)).
:- assert_must_succeed(clpfd_interface:clpfd_can_match(11,_X)).
:- assert_must_succeed(clpfd_interface:clpfd_can_match(_X,22)).
:- assert_must_succeed(clpfd_interface:clpfd_can_match(X,X)).
:- assert_must_fail(clpfd_interface:clpfd_can_match(21,22)).
:- assert_must_fail((dif(A,B),clpfd_interface:clpfd_can_match(A,B))).
% check if two numbers or fd variables can match without unifying (as this could trigger propagations)
clpfd_can_match(Var1,Var2) :- var(Var1),!, (var(Var2) -> clpfd_can_unify(Var1,Var2) ; clpfd_can_match_nr(Var1,Var2)).
clpfd_can_match(Nr1,Var2) :- var(Var2),!,clpfd_can_match_nr(Var2,Nr1).
clpfd_can_match(Nr1,Nr2) :- Nr1=Nr2.
clpfd_can_match_nr(Var,Nr) :- % is it also worthwhile to check for frozen_dif here ? A #>10, B =15, B #\=A, clpfd_interface:clpfd_can_match(A,B) fails without it
    fd_set(Var,FDS), Nr in_set FDS.
clpfd_can_unify(Var1,Var2) :- \+ kernel_dif:frozen_dif(Var1,Var2),
    fd_set(Var1,FDS1), fd_set(Var2,FDS2), fdset_intersect(FDS1,FDS2).

% check if we can currently determine an integer to be greater or equal to a given number
clpfd_check_geq_nr(X,Nr) :-
   number(X),!,
   X>=Nr.
clpfd_check_geq_nr(X,_) :-
   (nonvar(X) % something like X+Y
    ; preferences:preference(use_clpfd_solver,false)),
   !.
clpfd_check_geq_nr(X,Nr) :-
   fd_min(X,Low), number(Low), Low>=Nr.

clpfd_size(Var,Size) :- number(Var),!, Size=1.
clpfd_size(Var,Size) :-
   (preferences:preference(use_clpfd_solver,false)
     -> Size = inf
     ;  fd_size(Var,Size)
    ).

:- assert_must_succeed(clpfd_interface:clpfd_alldifferent([1,2,3])).
:- assert_must_succeed(clpfd_interface:all_dif([1,2,3],[])).
:- assert_must_fail(clpfd_interface:clpfd_alldifferent([1,2,3,2])).
:- assert_must_fail(clpfd_interface:all_dif([1,2,3,2],[])).

clpfd_alldifferent([]) :- !.
clpfd_alldifferent(L) :- post_constraint(all_different(L),all_dif(L,[])).

all_dif([],_).
all_dif([H|T],All) :- all_dif_aux(All,H), all_dif(T,[H|All]).

all_dif_aux([],_).
all_dif_aux([H|T],X) :- dif(H,X), all_dif_aux(T,X).

/*
clpfd_labeling(Variables) :-
  (preferences:preference(use_clpfd_solver,false)  -> true
    ; labeling([ffc,enum],Variables)).  %ff, enum
*/

% a version of labeling that performs randomized enumeration if requested
% Currently, random enumeration does not support variable selection options
% TODO: Maybe we should only randomize enumeration on "large enough" intervals
:- use_module(library(random)).
:- use_module(extension('random_permutations/random_permutations')).
clpfd_randomised_labeling([],ListOfVars) :-
  preferences:get_preference(randomise_enumeration_order,true)
  -> clpfd_randomised_labeling_aux(ListOfVars)
  ;  labeling([],ListOfVars).
clpfd_randomised_labeling([Op|Ops],ListOfVars) :-
  labeling([Op|Ops],ListOfVars).

clpfd_randomised_labeling_aux([]).
clpfd_randomised_labeling_aux([Var|Vars]) :-
  clpfd_randomised_in_domain(Var),
  clpfd_randomised_labeling_aux(Vars).

clpfd_randomised_in_domain(Var) :-
  init_random_permutations,
  clpfd_domain(Var,Low,Up),
  ((\+ number(Low) ; \+ number(Up))
    -> add_internal_error('Unbounded FD variable: ',clpfd_randomised_in_domain(Var)),
       %trace, kernel_waitflags:get_fd_priority(Var,_),
       fail
    ; true),
  IntervalLength is Up - Low + 1,
  get_num_bits(IntervalLength,MaxIdx,NumBits),
  get_masks(NumBits,LeftMask,RightMask),
  % the seed relies on the random predicate, not on now/1, thus prob can be made deterministic by setting a central random seed
  random(TempSeed),
  Seed is floor(TempSeed * 10000),
  clpfd_randomised_labeling_aux(Var,0,MaxIdx,Low,Up,Seed,NumBits,LeftMask,RightMask).

clpfd_randomised_labeling_aux(N,CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask) :-
    random_permutation_element(CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask,Drawn,NextIdx),
    %format('enumerate_int2: Setting variable ~w to ~w~n',[N,Drawn]),
    ( N=Drawn
    ; clpfd_randomised_labeling_aux(N,NextIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask)).


%:- use_module(library(lists),[reverse/2]).
%clpfd_reverse_labeling(Vars) :- clpfd_reverse_labeling(Vars,[ffc,enum]).
%% label a list of variables which was reversed:
%clpfd_reverse_labeling(Variables,Opt) :- var(Variables),!,
%   add_internal_error('Must be list of variables: ',clpfd_reverse_labeling(Variables,Opt)).
%clpfd_reverse_labeling(Variables,Options) :- !,
%   reverse(Variables,RevV),
%   clpfd_labeling(RevV,Options).

clpfd_labeling(Variables,Options) :- !,
   on_exception(error(instantiation_error,Err),labeling(Options,Variables),
                (print(Err),nl,
                 % this can happen when CLP(FD) has a time-out for other constraints
                 add_error(clpfd_interface,'CLP(FD) Variables not set up for ',labeling(Options,Variables)))). % Options = [ffc,enum]


clpfd_in_domain(Variable) :- !,
   on_exception(error(instantiation_error,Err),indomain(Variable),
                (print(Err),nl,
                 % this can happen when CLP(FD) has a time-out for other constraints
                 add_error(clpfd_interface,'CLP(FD) Variables not set up for ',indomain(Variable)))).


clpfd_degree(Variable,Degree) :- fd_degree(Variable,Degree). % can be called with plain variables and with numbers

% a new predicate to obtain the next variable that would be enumerated
% (useful if you want to interleave other labeling with CLPFD labeling)

:- assert_must_succeed((clpfd:in(X,1..3), clpfd:in(Y,3..5), clpfd:'#>'(X,Z), clpfd_interface:clpfd_get_next_variable_to_label([Y,2,X,3],V),V==X, X=3, Y=3, Z=2)).
clpfd_get_next_variable_to_label(Variables,Var) :-
    clpfd_get_next_variable_to_label(Variables,Var,ffc).

% possible options : ff, ffc, step, enum, bisect, median, ...
clpfd_get_next_variable_to_label(Variables,Var,Opt) :- Variables \= [], % there is a segmentation fault if we call fd_delete with empty list
% this also causes segmentation fault:
% X in 1..2 , Y in 3..5, clpfd_get_next_variable_to_label([2,3,Y,2,X],Var,ffc). rlwrap: warning: sicstus crashed, killed by SIGSEGV.
    clpfd:'$fd_delete'(Variables,Var,Opt).

% a version of the above just for ffc, re-implemented by hand so as not to be dependent on internal predicates of clpfd
% also: it returns the remaining variables (filters out non-variables)
% but it can be much slower than clpfd:$fd_delete (e.g., NQueens50)
clpfd_get_next_variable_to_label_ffc([H|T],Var,RestV) :-
   (get_fd_info(H,HI) -> get_next_aux(T,H,HI,Var,RestV)
      %,((clpfd_get_next_variable_to_label([H|T],VV,ffc), VV\==Var) -> print(different_choice(Var,VV)),nl , tools_printing:print_vars([H|T]) ; true)
    ; clpfd_get_next_variable_to_label_ffc(T,Var,RestV)
    ).
   % , tools_printing:print_vars([H|T]),print(sel(Var,RestV)),nl.

get_next_aux([],Var,_,Var,[]).
get_next_aux([H|T],VarSoFar,VInfo,Var,Rest) :-
   (get_fd_info(H,HI)
      -> (HI @< VInfo -> Rest = [VarSoFar|TRest], get_next_aux(T,H,HI,Var,TRest)
           ; Rest = [H|TRest], get_next_aux(T,VarSoFar,VInfo,Var,TRest)
        )
      ;  get_next_aux(T,VarSoFar,VInfo,Var,Rest)).

get_fd_info(Var,(Sz,NDg)) :- fd_size(Var,Sz), Sz \= 1, % note: we have 2 @<sup
   %  kernel_waitflags:size_of_attached_goals(Var,Dg),
    fd_degree(Var,Dg),
    NDg is -Dg. % negate to give higher priority to goals with high degree

% would could also use size of attached frozen goals as a tiebreaker :  frozen_degree(Var,FDg), NFDg is -FDg.
% maybe we should compute this only on demand
%frozen_degree(Var,Dg) :- var(Var), !,frozen(Var,Goal), size_of_goal(Goal,Dg).
%frozen_degree(_,0).
%size_of_goal((A,B),S) :- !, size_of_goal(A,SA), size_of_goal(B,SB), S is SA+SB.
%size_of_goal(_,1).


:- meta_predicate catch_clpfd_overflow_call1(0).
:- meta_predicate catch_clpfd_overflow_call2(0,0).
catch_clpfd_overflow_call1(Call) :- catch_clpfd_overflow_call2(Call,fail).
catch_clpfd_overflow_call2(Call,Handler) :-
  on_exception(error(representation_error(ErrMsg),E),
               Call,
               handle_representation_error(ErrMsg,E,Handler)).

handle_representation_error(ErrMsg,_E,Call) :-
    (is_clpfd_overflow_representation_error(ErrMsg)
       -> clpfd_overflow_error_message, Call
       ;  throw(representation_error(ErrMsg))).

is_clpfd_overflow_representation_error('CLPFD integer overflow').
is_clpfd_overflow_representation_error('max_clpfd_integer').
is_clpfd_overflow_representation_error('min_clpfd_integer').

/***************************************************************************/
/*                 CSP                                                     */
/***************************************************************************/

/* currently not used anymore :
csp_clpfd_eq(E1,E2) :-
  number(E1),number(E2),!,E1=E2.
csp_clpfd_eq(E1,E2) :- post_constraint(E1 #= E2).
csp_clpfd_neq(E1,E2) :-
  number(E1),number(E2),!,E1\=E2.
csp_clpfd_neq(E1,E2) :-
   post_constraint( E1 #\= E2 , dif(E1,E2)).

csp_clpfd_domain(Vars,Low,Up) :-
    number(Low),number(Up),!,
    post_constraint(domain(Vars,Low,Up)).

csp_clpfd_labeling(Options,Vars) :-
    post_constraint(labeling(Options,Vars)).

:- use_module(probcspsrc(csp_basic),[safe_less_than/2, safe_less_than_equal/2]).

csp_clpfd_lt(E1,E2) :-
   post_constraint( E1 #< E2, safe_less_than(E1,E2)).
csp_clpfd_leq(E1,E2) :-
   post_constraint( E1 #=< E2, safe_less_than_equal(E1,E2)).
csp_clpfd_gt(E1,E2) :-
   post_constraint( E1 #> E2, safe_less_than(E2,E1)).
csp_clpfd_geq(E1,E2) :-
   post_constraint( E1 #>= E2, safe_less_than_equal(E2,E1)).

csp_in_fdset(X,Dom) :-
   list_to_fdset(Dom,FDSet),
   X in_set FDSet.
 */

/*
Testing for how large the integers can be:
| ?- X is integer(2**30), Y#=X+1.
! Representation error in argument 2 of user:'t=u+c'/3
! CLPFD integer overflow
! goal:  't=u+c'(_102,1073741824,1)
| ?- X is integer(2**30), Y#=X.
X = 1073741824,
Y = 1073741824 ?
*/

/* grounding_wait_flags
 :::
enumerate_int
| ?- on_exception(E,X in 0..2147483646, print(E)).
error(representation_error(CLPFD integer overflow),representation_error(_710 in 0..2147483646,2,CLPFD integer overflow))
E = error(representation_error('CLPFD integer overflow'),representation_error(_A in 0..2147483646,2,'CLPFD integer overflow')) ?
yes

X in 0..9, X #>3, fd_dom(X,R).
R = 4..9,
X in 4..9 ?
yes
| ?- X in 0..9, X #>3, fd_dom(X,R), R=.. L.
L = [..,4,9],
R = 4..9,
X in 4..9 ?
yes
| ?- fd_dom(X,R).
R = inf..sup ?
yes
| ?- X in 0..2147483647, X#<2, X#>0.
! Representation error in argument 2 of user:in/2
! CLPFD integer overflow
! goal:  _97 in 0..2147483647
| ?- X in 0..214748364.
X in 0..214748364 ?
yes
| ?- X in 0..2147483646.
! Representation error in argument 2 of user:in/2
! CLPFD integer overflow
! goal:  _97 in 0..2147483646
| ?- X in 0..10, X #>2147483646.
! Representation error in argument 1 of user:'x=<y IND'/2
! CLPFD integer overflow
! goal:  'x=<y IND'(2147483647,_98)
*/
