% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(error_manager,[add_warning/2, add_warning/3, add_warning/4,
           reset_errors/0, add_error/2, add_error/3, add_error/4,
           add_error_and_fail/3, add_error_and_fail/2,
           add_error_fail/3,
           add_failed_call_error/1, add_internal_error/2, add_internal_error/4,
           register_optional_error/1, reset_optional_errors/0,
           assert_true/1,
           add_message/2, add_message/3, add_message/4,
           add_all_perrors/1, add_all_perrors/2, add_all_perrors/3,
           no_real_perror_occurred/1,
           get_all_errors/1,
           get_all_errors_and_reset/1, % call if you want to start with a fresh top-level error scope
           get_all_errors_and_clear/1,
           get_error/2, get_warning/2, check_error_occured/2, check_warning_occured/2,
           get_error_span_linecol/5, get_error_span_linecol/4, get_error_span_linecol_for_main_file/4,
           register_error_span/2,
           check_error_span_file_linecol/6,
           position_is_not_in_main_file/1,
           real_error_occurred/0, reset_real_error_occurred/0, assert_real_error_occurred/0,
           test_error_occurred/1, count_errors_occurred/2,
           get_total_number_of_errors/1,

           backup_errors/0, backed_up_error/4,

           set_error_context/1, clear_error_context/0, get_error_context/1,

           extract_span_description/2,
           extract_line_col/5, extract_all_line_col/5,
           extract_line_col_for_main_file/5,
           extract_subsidiary_tail_file_name/2, extract_tail_file_name/2,
           extract_file_number_and_name/3,
           print_error_span/1,

           turn_show_error_source_on/0, turn_show_error_source_off/0,

            enter_new_error_scope/1, enter_new_error_scope/2,
            enter_new_clean_error_scope/1,
            exit_error_scope/2, exit_error_scope/3,
            error_occurred_at_level/1, specific_event_occurred_at_level/2,
            throw_errors_in_scope/2,
            add_new_event_in_error_scope/1, add_new_event_in_error_scope/2,
            event_occurred_in_error_scope/1, error_occurred_in_error_scope/0,
            clear_events_in_error_scope/2, clear_events_in_current_error_scope/1,
            clear_all_errors_in_error_scope/1,
            call_in_fresh_error_scope_for_one_solution/1,
            enumeration_warning_occured_in_error_scope/0,
            abort_error_occured_in_error_scope/0,
            on_enumeration_warning/2, on_enumeration_warning_with_continue/3,
            throw_enumeration_warnings_in_current_scope/0,
            throw_enumeration_warnings_in_scope/2,
            catch_enumeration_warning_exceptions/2, catch_enumeration_warning_exceptions/3,
            call_with_enumeration_warning/1,
            observe_enumeration_warnings/2,
            critical_enumeration_warning_occured_in_error_scope/4, critical_enumeration_warning_occured_in_error_scope/0,
            clear_enumeration_warnings/0,
            time_out_with_enum_warning_one_solution/3, % use this for finding one solution
            time_out_with_enum_warning_one_solution_no_new_error_scope/3,
            is_time_out_result/1,
            time_out_with_enum_warning_for_findall/3, % use this if you cycle through all solutions
            time_out_with_enum_warning_for_findall_in_current_error_scope/4,
            logged_error/4,
            display_error_statistics/0
            ]).

:- use_module(library(lists)).

:- use_module(tools,[get_tail_filename/2]).
:- use_module(tools_lists,[count_occurences/2]).
:- use_module(tools_printing,[print_error/1]).
:- use_module(tools_strings,[ajoin/2, string_concatenate/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The error_manager controls error messages. Error messages can be added, retrieved, counted,etc. The module also takes care of extracting source spans from error messages.').

%:- use_module(translate).
:- use_module(debug, [debug_mode/1, debug_println/1]).
:- use_module(preferences,[get_preference/2,reset_temporary_preference/2,
                           temporary_set_preference/3]).


:- dynamic optional_error/1.
register_optional_error(Type) :-
   (optional_error(Type) -> true ; assert(optional_error(Type))).
reset_optional_errors :- retractall(optional_error(_)).

% INTERNAL ERRORS

:- meta_predicate add_failed_call_error(0).
:- meta_predicate add_internal_error(-,0).

add_failed_call_error(Call) :- !,
   add_internal_error('Internal Error; call failed: ',Call).

add_internal_error(Module,Msg,Term,Span) :-
   add_error(internal_error(Module),Msg,Term,Span).
add_internal_error(Msg,Module:Call) :- !,
   add_error(internal_error(Module),Msg,Call).
add_internal_error(Msg,Call) :-
   add_internal_error(Msg,unknown_module:Call).


% ERROR SCOPING

:- dynamic cur_block_level/1.
cur_block_level(0).
:- volatile error_occurred_at_level/1.
:- dynamic error_occurred_at_level/1.
:- volatile throw_error/2.
:- dynamic throw_error/2.



enter_new_error_scope(Level,_PP) :-
   enter_new_error_scope(Level). %, tools_printing:print_term_summary(enter(Level,PP)),nl.
enter_new_error_scope(Level) :-
   enter_new_clean_error_scope(Level),
   L1 is Level - 1,
   copy_throw_errors(L1,Level).
copy_throw_errors(L1,L) :- throw_error(L1,Error), %print(copy(throw(L1,L,Error))),nl,
   assert(throw_error(L,Error)),fail.
copy_throw_errors(_,_).
enter_new_clean_error_scope(Level) :-
  retract(cur_block_level(L)), L1 is L+1, assert(cur_block_level(L1)),
  retractall(throw_error(L1,_)),
  % print(enter(L1)),nl,trace,
  Level=L1.
% mark that everything unifying with Error should be thrown rather than collected
throw_errors_in_scope(Level,Error) :-
   (\+ throw_error(Level,Error)  % TO DO: check that we are an instance of existing error !
    -> assert(throw_error(Level,Error))
     ; true).

clear_all_errors_in_error_scope(Level) :- %print(clearing(Level)),nl,
   clear_events_in_error_scope(Level,_),
   error_occurred_at_level(L1), L1>Level, %print(clearing(L1)),nl,
   clear_events_in_error_scope(Level,_),fail.
clear_all_errors_in_error_scope(_).

exit_error_scope(Level,ErrorOcc) :- exit_error_scope(Level,ErrorOcc,unknown).
exit_error_scope(Level,ErrorOcc,PP) :- retract(cur_block_level(CurL)),
    % print(exit(Level,CurL,PP)),nl,
   (CurL = Level -> NewL is CurL-1
    ; Level < CurL -> NewL is Level-1,  format('*** exiting multiple error blocks (~w -> ~w) for ~w ***~n',[CurL,NewL,PP])
    ; NewL is CurL -> format('*** block ~w already exited (currently at ~w) for ~w ***~n',[Level,CurL,PP])
   ), assert(cur_block_level(NewL)),
   copy_error_scope_events(CurL,NewL),
   retractall(throw_error(CurL,_)),
   (retract(error_occurred_at_level(CurL))
    -> ErrorOcc=true,
       (error_occurred_at_level(NewL) -> true ; assert(error_occurred_at_level(NewL)))
    ; ErrorOcc=false).


% versions of enter_new_error_scope and exit_error_scope which re_enter blocks and clean up upon backtracking
%enter_new_error_scope_bt(Level,PP) :- enter_new_error_scope(Level,PP), exit_upon_fail(Level,PP).
%exit_upon_fail(_,_).
%exit_upon_fail(Level,PP) :-  print(exit_upon_backtrack(Level,PP)),nl,
%   exit_error_scope(Level,_,PP),fail.
%
%exit_error_scope_bt(Level,ErrorOcc,PP) :- exit_error_scope(Level,ErrorOcc,PP).
%exit_error_scope_bt(Level,_ErrorOcc,PP) :- print(re_enter_block(Level,PP)),nl,
%   enter_new_error_scope(Level2,PP),
%   (Level2=Level -> true ; print(mismatch_when_reentering_block(Level,PP)),nl),
%   fail.

% ---------------------------



assert_real_error_in_current_block :- cur_block_level(L),L>0,
  \+ error_occurred_at_level(L), assert(error_occurred_at_level(L)),!.
assert_real_error_in_current_block.


reset_error_scopes :- retractall(cur_block_level(_)), assert(cur_block_level(0)),
   retractall(error_occurred_at_level(_)),
   retractall(specific_event_occurred_at_level(_,_)).

% for debugging:
:- public print_error_scopes/0.
print_error_scopes :- cur_block_level(L),
   format(' Cur Block Level = ~w~n',[L]),
   fail.
print_error_scopes :-
   findall(L,error_occurred_at_level(L),LL),
   format(' Errors occured at levels = ~w~n',[LL]),
   specific_event_occurred_at_level(Level,Err),
   format(' ~w  -->  ~w~n',[Level,Err]),
   fail.
print_error_scopes.

:- meta_predicate call_in_fresh_error_scope_for_one_solution(0).
% call a predicate in fresh error scope (assuming one solution is searched for only)
% and then discard all errors
call_in_fresh_error_scope_for_one_solution(Call) :-
     enter_new_error_scope(Level,call_in_fresh_error_scope_for_one_solution),
     call_cleanup(Call,
                   (clear_all_errors_in_error_scope(_),
                    exit_error_scope(Level,_,call_in_fresh_error_scope_for_one_solution))).

:- volatile specific_event_occurred_at_level/2.
:- dynamic specific_event_occurred_at_level/2.

% can be used to keep track of whether specific events occurred within an error_scope
:- meta_predicate add_new_event_in_error_scope(-,1).
add_new_event_in_error_scope(Event) :-
    add_new_event_in_error_scope(Event,do_nothing).
do_nothing(_).
add_new_event_in_error_scope(Event,PrintMessageCall) :- cur_block_level(L),
   (specific_event_occurred_at_level(L,Event) -> fail
     ; throw_error(L,Event) -> (debug:debug_mode(off) -> true ; print_message(informational,throwing(Event))), % trace, % <---
                               call(PrintMessageCall,throwing),
                               throw(Event)
     ; assert(specific_event_occurred_at_level(L,Event)),
       call(PrintMessageCall,not_throwing)
   ).

% call if you just want to add the error information; without throwing
assert_event_in_error_scope_without_throw(Event) :- cur_block_level(L),
   (specific_event_occurred_at_level(L,Event) -> true
    ; assert(specific_event_occurred_at_level(L,Event))).

error_occurred_in_error_scope :- cur_block_level(L),error_occurred_at_level(L).
event_occurred_in_error_scope(Event) :- cur_block_level(L),specific_event_occurred_at_level(L,Event).

copy_error_scope_events(FromLevel,ToLevel) :- retract(specific_event_occurred_at_level(FromLevel,Event)),
   \+ specific_event_occurred_at_level(ToLevel,Event),
   % print(copying_event(FromLevel,ToLevel,Event)),nl,
   assert(specific_event_occurred_at_level(ToLevel,Event)), % should we throw if this level demands it ?
   (throw_error(ToLevel,Event) -> debug:debug_println(20,throw_copied_event(Event)),
                                  throw(Event)),
   fail.
copy_error_scope_events(_,_).

clear_events_in_current_error_scope(Event) :- cur_block_level(L),clear_events_in_error_scope(L,Event).

%clear_events_in_error_scope(L,Event) :-retract(specific_event_occurred_at_level(L,Event)), print(retract(L,Event)),nl,fail.
clear_events_in_error_scope(L,Event) :-
   retractall(specific_event_occurred_at_level(L,Event)),
   (specific_event_occurred_at_level(L,_) -> true
    ; retractall(error_occurred_at_level(L))).

throw_enumeration_warnings_in_current_scope :-
   cur_block_level(L), throw_enumeration_warnings_in_scope(L,_).

throw_enumeration_warnings_in_scope(Level,Critical) :-
   throw_errors_in_scope(Level,enumeration_warning(_,_,_,_,Critical)).

:- meta_predicate on_enumeration_warning(0,0).
:- meta_predicate on_enumeration_warning_with_continue(0,0,0).
on_enumeration_warning(Call, BackupCall) :-
        on_enumeration_warning_with_continue(Call, BackupCall, true).
% sets up a new error scope and forces enumeration_warnings to be thrown
% if an enumeration_warning happens the BackupCall is executed
on_enumeration_warning_with_continue(Call, BackupCall, TrueContinuation) :-
    enter_new_error_scope(Level,on_enumeration_warning_with_continue),
	throw_enumeration_warnings_in_scope(Level,_),
	on_exception(enumeration_warning(_,_,_,_,_),
		    if(Call,
		         (exit_error_scope(Level,_,on_enumeration_warning_with_continue_true),call(TrueContinuation)),
		         (exit_error_scope(Level,_,on_enumeration_warning_with_continue_false),fail)
		       ),
		    (exit_error_scope(Level,_,on_enumeration_warning_with_continue_exception),BackupCall)).
:- meta_predicate catch_enumeration_warning_exceptions(0,0).
:- meta_predicate catch_enumeration_warning_exceptions(0,0,-).
catch_enumeration_warning_exceptions(Call, BackupCall) :-
    catch_enumeration_warning_exceptions(Call,BackupCall,false).



% AddAsEventError = true means that caught exceptions are added as event_errors in scope
% (this can be important if we later want to decide whether an enumeration warning occured)
catch_enumeration_warning_exceptions(Call, BackupCall,AddAsEventError) :-
    enter_new_error_scope(Level,catch_enumeration_warning_exceptions),
	% difference: we do not force to throw_errors_in_scope
	on_exception(enumeration_warning(A,B,C,D,E),
		         call_cleanup(Call, exit_error_scope(Level,_,catch_enumeration_warning_exceptions)),
		         % call_cleanup can be tricky in case of success of Call: the error scope may not be exited immediately; TO DO: maybe rewrite using same style as above (using if + on_exception)
		         ((AddAsEventError=false -> true
		            ; assert_event_in_error_scope_without_throw(enumeration_warning(A,B,C,D,E))),
		          BackupCall)).

:- meta_predicate observe_enumeration_warnings(0,0).
% execute some code when enumeration warning occurs; and then throw exception again
observe_enumeration_warnings(Call,ObserverCallToExecute) :-
	on_exception(enumeration_warning(A,B,C,D,E),Call,
	   call_cleanup(ObserverCallToExecute, throw(enumeration_warning(A,B,C,D,E)))).


:- meta_predicate call_with_enumeration_warning(0).
% sets up a new error scope and forces enumeration_warnings to be thrown
% the exceptions are not caught !
call_with_enumeration_warning(Call) :-
    enter_new_error_scope(Level,call_with_enumeration_warning),
	throw_enumeration_warnings_in_scope(Level,_),
	on_exception(E,
		    if(Call,
		         exit_error_scope(Level,_,call_with_enumeration_warning_true),
		         (exit_error_scope(Level,_,call_with_enumeration_warning_false),fail)
		       ),
		    (exit_error_scope(Level,_,call_with_enumeration_warning_exception),throw(E))).


:- meta_predicate time_out_with_enum_warning_one_solution(0,-,-).
% a version of time_out which treats enumeration_warnings like time-outs
% and also checks enumeration_warnings in case of failure
% assumes we only look for one solution (otherwise error scope cleanup has to be changed !! )
time_out_with_enum_warning_one_solution(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level,time_out_with_enum_warning_one_solution),
    %% print(time_out_with_enum_warning_one_solution(TO)),nl, %%
    call_cleanup(
        on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,perform_cut), % no need to clear: we leave error_scope below
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))),
        exit_error_scope(Level,_,time_out_with_enum_warning_one_solution)).

enumeration_warning_occured_in_error_scope :-
   event_occurred_in_error_scope(enumeration_warning(_,_,_,_,_)).
abort_error_occured_in_error_scope :-
   event_occurred_in_error_scope(abort_error(_)).

% just like above; but does not set up new error_scope (for performance) and clears enumeration warnings in case of failure and success
:- meta_predicate time_out_with_enum_warning_one_solution_no_new_error_scope(0,-,-).
time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes) :-
    (enumeration_warning_occured_in_error_scope
     -> %trace,
        add_internal_error('Error block already has enumeration warnings',time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes))
      ; true),
    on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,clear_errors,perform_cut), % clear errors so we can continue using same error scope
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))).


% check if a result is a time_out:
is_time_out_result(time_out).
is_time_out_result(virtual_time_out(_)).

% a version which is to be used not for a single call but for backtracking
% generate virtual time_out at the end if enumeration warning occured
:- meta_predicate time_out_with_enum_warning_for_findall(0,-,-).
:- meta_predicate time_out_with_enum_warning_for_findall2(-,0,-,-).

time_out_with_enum_warning_for_findall(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level,time_out_with_enum_warning_for_findall),
    time_out_with_enum_warning_for_findall2(Level,Call,TO,TimeOutRes).

time_out_with_enum_warning_for_findall2(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(Cause,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,no_cut),
                 get_virtual_time_out_res(Cause,TimeOutRes)). % virtual time_out
time_out_with_enum_warning_for_findall2(Level,_Call,_TO,_TimeOutRes) :- % print(exit(Level)),nl,
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),
    exit_error_scope(Level,_,time_out_with_enum_warning_for_findall2),fail.

get_virtual_time_out_res(kodkod_timeout,Res) :- !, Res= virtual_time_out(kodkod_timeout).
get_virtual_time_out_res(_,virtual_time_out(exception)).

% a version of time_out_with_enum_warning_for_findall where you manually have to enter_new_error_scope before
% and exit_error_scope afterwards; useful if you repeatedly call this predicate and want to set up the block only once
:- meta_predicate time_out_with_enum_warning_for_findall_in_current_error_scope(-,0,-,-).
time_out_with_enum_warning_for_findall_in_current_error_scope(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(Cause,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear,no_cut),
                 get_virtual_time_out_res(Cause,TimeOutRes)). % virtual time_out
time_out_with_enum_warning_for_findall_in_current_error_scope(Level,_Call,_TO,_TimeOutRes) :-
    % does not exit level
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),fail.

:- use_module(tools_meta,[safe_time_out/3]).
% a version of time_out which upon failure checks if enumeration warnings were generated
time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,ClearErrors,CutAfterFirstSol) :-
    safe_time_out(Call,TO,TimeOutRes),  % a safe version which checks that time-out value TO is not too large
    (TimeOutRes = time_out -> !  %, print(real_timeout),nl
       ; (ClearErrors=clear_errors -> clear_enumeration_warnings ; true),
         (CutAfterFirstSol=perform_cut -> ! ; true)
         %,print(sol),nl
    ).
time_out_check_enum_warning_for_failure(_Call,_TO,TimeOutRes,ClearErrors,_) :- % print(failure),nl,
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)), % print(virtual),nl,
    TimeOutRes = virtual_time_out(failure_enumeration_warning(A,B,C,D,critical)), % we have not investigated all possibilities; generate virtual time_out
    (ClearErrors=clear_errors -> clear_enumeration_warnings ; true).

critical_enumeration_warning_occured_in_error_scope :-
   critical_enumeration_warning_occured_in_error_scope(_,_,_,_).
critical_enumeration_warning_occured_in_error_scope(A,B,C,D) :-
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)).

clear_enumeration_warnings :-
    clear_events_in_current_error_scope(enumeration_warning(_,_,_,_,_)).

% -----------------------

:- volatile logged_error_hash/5, real_error_occurred/0, logged_error_spans/2.
:- dynamic logged_error_hash/5.
:- dynamic real_error_occurred/0.
:- dynamic logged_error_spans/2.

reset_real_error_occurred :- retractall(real_error_occurred).
reset_errors :- %print(reset_errors),nl, trace,
                reset_errors_but_not_scopes,
                reset_error_scopes. % call if you want to start again at a fresh top-level, without error scopes
reset_errors_but_not_scopes :-
                retractall(real_error_occurred),
                retractall(logged_error_hash(_,_,_,_,_)),retractall(logged_error_spans(_,_)),
                clear_error_context.

reset_error_span_if_required :-
	(logged_error_hash(_,_,_,_,_) -> true
		; retractall(logged_error_spans(_,_))).




print_error_span(unknown) :- !.
print_error_span(Span) :-
     extract_span_description(Span,PosMsg), !,
     print_error(PosMsg).
print_error_span(b(E,T,I)) :- translate:translate_subst_or_bexpr(b(E,T,I),S),!,
     print_error(S).
print_error_span(Span) :- print_error(Span). % span not recognized !

% try and extract a string (aka atom) describing the position
extract_span_description(Span,PosMsg) :-
   extract_line_col(Span,StartLine,StartCol,EndLine,EndCol),!,
   POS = ['Line: ',StartLine,' Column: ',StartCol|POS1],
   (get_preference(show_full_error_span,true)
     -> POS1 = [' until ',EndLine,':',EndCol|POS2]
     ;  POS1 = POS2),
   ((extract_file_name(Span,Filename),Filename \=null) -> % Filename = null occurs in eval_strings !?
      POS2 = [' in file: ',Filename] ; POS2 = []),
     ajoin(POS,PosMsg).
extract_span_description(Span,PosMsg) :-
   extract_symbolic_pos(Span,Symbolic),!,
   % TO DO: we could try and find section for Rodin
   ajoin(['@label: ',Symbolic],PosMsg).

:- meta_predicate assert_true(0).
assert_true(Code) :-
  (call(Code) -> true ; add_internal_error('Assertion failed: ',Code)).

add_error_fail(Source,ErrMsg, ErrTerm) :- add_error_and_fail(Source,ErrMsg, ErrTerm).
add_error_and_fail(Source,ErrMsg, ErrTerm) :- add_error(Source,ErrMsg, ErrTerm),fail.
add_error_and_fail(Source,ErrMsg) :- add_error(Source,ErrMsg),fail.

% add_error/2, add_error/3 and add_error/4
add_error(Source,ErrMsg) :- add_error4(Source,ErrMsg, '',unknown).
add_error(Source,ErrMsg, ErrTerm) :- add_error4(Source,ErrMsg, ErrTerm,unknown).
add_error(Source,ErrMsg,ErrTerm,Span) :- add_error4(Source,ErrMsg,ErrTerm,Span).

add_error4(Source,ErrMsg,ErrTerm,Span) :-
    assert_real_error_occurred,
    combine_msg_and_error_term(ErrMsg,ErrTerm,S),
    add_error_or_warning(Source,S,Span).

%:- use_module(translate, [translate_error_term/2, translate_context/2]).
combine_msg_and_error_term(Msg,'',Res) :- !, Res=Msg.
combine_msg_and_error_term(ErrMsg,ErrTerm,S) :-
    string_concatenate(ErrMsg,' ',ST),
	safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ST,RString,S).


:- dynamic show_error_source/0.
show_error_source. % True if error source and span to be printed
turn_show_error_source_on :- (show_error_source -> true ; assert(show_error_source)).
turn_show_error_source_off :-  retractall(show_error_source).

trace_if_user_wants_it :- (preferences:get_preference(trace_upon_error,true) -> safe_trace ; true).

% try Prolog's trace; if this fails (compiled code), then ask user for input.
safe_trace :- on_exception(error(existence_error(_,_),_), trace, type_to_continue).

:- dynamic do_not_interrupt_user_anymore/0.

type_to_continue :- do_not_interrupt_user_anymore,!.
type_to_continue :- format(user_output,'(type <RETURN> to continue, z<RETURN> to zip) ===> ',[]),
   read_line(user_input,X),
   (zip(X) -> assert(do_not_interrupt_user_anymore) ; true).

zip([122|_]). %starts with z


print_error_source(well_definedness_error) :- !,print_error('A well-definedness error occured !').
print_error_source(warning(Source)) :- show_error_source,!,
    print_error('A warning occurred !'),print_error(source(Source)).
print_error_source(Source) :- show_error_source,!,
    print_error('An error occurred !'),print_error(source(Source)).
print_error_source(_).

print_error_context(Context) :- ((Context = '' ; \+show_error_source) -> true ; print_error(Context)).

safe_translate_error_term(ErrTerm,RString) :-
    preferences:temporary_set_preference(expand_avl_upto,5,CHNG),
    translate:translate_error_term(ErrTerm,RString),
    preferences:reset_temporary_preference(expand_avl_upto,CHNG).

safe_translate_context(Result) :-
% avoid infinite loop when error occurs inside translate_context
    get_error_context(Context),
    set_error_context(translate_context),
    preferences:temporary_set_preference(expand_avl_upto,4,CHNG),
    (translate:translate_context(Context,TContext)
      -> true ; TContext=Context),
    preferences:reset_temporary_preference(expand_avl_upto,CHNG),
    set_error_context(Context),
    Result=TContext.


add_error_or_warning(Source,ErrMsg,Span) :-
    (preferences:get_preference(strict_raise_warnings,true)
    -> assert_real_error_occurred ; true),
    safe_translate_context(Context),
    reset_error_span_if_required,
    assertz_logged_error_if_new(Source,ErrMsg,Context,Span,New),
    (New=false -> true
      ; optional_error(Source) ->
        (extract_span_description(Span,PosMsg) -> true ; PosMsg=''),
        format(user_output,'% Optional error ~w occured: ~w ~w~n',[Source,ErrMsg,PosMsg]),
        (Span=unknown -> true ; assert(logged_error_spans(Source,Span)))
      ; print_error_source(Source),
        print_error(ErrMsg),
        print_error_context(Context),
        (Span=unknown -> true
          ; print_error_span(Span),
            assert(logged_error_spans(Source,Span))),
        trace_if_user_wants_it
    ).

:- dynamic total_number_of_errors/1.
total_number_of_errors(0).

% the total number of errors since start; no way to reset this; useful for validation reports
% TO DO: maybe use counter extension; but tricky to get dependency and initialisation right
get_total_number_of_errors(X) :- total_number_of_errors(X).

assert_real_error_occurred :-
    (real_error_occurred -> true ; assert(real_error_occurred)),
    (retract(total_number_of_errors(X)) -> X1 is X+1, assert(total_number_of_errors(X1)) ;
      assert(total_number_of_errors(1))),
    assert_real_error_in_current_block.

:- use_module(library(terms),[term_hash/2]).
assertz_logged_error_if_new(Source,S,TContext,Span,New) :-
    term_hash(logged_error_hash(Source,S,TContext,Span),Hash),
    (logged_error_hash(Hash,Source,S,TContext,Span)
     -> debug:debug_println(repeated_error(Hash)),New=false
      ; %print(new(Hash,Source,S,TContext,Span)),nl,
        assertz(logged_error_hash(Hash,Source,S,TContext,Span)),
        New=true).

logged_error(Source,S,TContext,Span) :- logged_error_hash(_Hash,Source,S,TContext,Span).

add_warning(Source,ErrMsg, ErrTerm) :-
   add_warning(Source,ErrMsg, ErrTerm,unknown).
add_warning(Source,ErrMsg, ErrTerm,Span) :-
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_error_or_warning(warning(Source),S,Span).
add_warning(Source,Message) :- add_error_or_warning(warning(Source),Message,unknown).

add_message(Source,ErrMsg, ErrTerm,Span) :- /* same as add_error but does not add err msg */
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_message(Source,S),
    (Span=unknown -> true
          ; safe_translate_context(Context),
            print_error_context(Context),
            print_error_span(Span)).

add_message(Source,ErrMsg, ErrTerm) :- /* same as add_error but does not add err msg */
    add_message(Source,ErrMsg, ErrTerm,unknown).

add_message(Source,ErrMsg) :-
   (show_error_source -> print_error(source(Source)) ; true),
   print_error(ErrMsg).
   %assertz(logged_error(Source,ErrMsg,_,_)).

% adds a list of errors with position information
% [error(Msg,Pos)] or [warning(Msg,Pos)...]
no_real_perror_occurred([]).
no_real_perror_occurred([warning(_,_)|T]) :- no_real_perror_occurred(T).

add_all_perrors(Errors) :- add_all_perrors(Errors,[],bmachine).
add_all_perrors(Errors,Files) :- add_all_perrors(Errors,Files,bmachine).
add_all_perrors(Errors,Files,Context) :- add_all_perrors(Errors,Files,Context,0,100).
add_all_perrors([],_,_,_,_).
add_all_perrors([H|Rest],_Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    NrOfErrors >= MaxErrors,!,
    length([H|Rest],Len),
    add_error(ContextOfError,'Reached maximium number of errors; number of errors not reported: ',Len).
add_all_perrors([Error|Rest],Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    get_perror(Error,ErrOrWarn,Msg,Pos),
    extract_perror_pos(Pos,Filenames,NewPos),
    (ErrOrWarn == warning -> add_warning(ContextOfError,Msg,'',NewPos)
      ; ErrOrWarn = internal_error -> add_internal_error(ContextOfError,Msg,'',NewPos)
      ; add_error(ContextOfError,Msg,'',NewPos)),
    N1 is NrOfErrors+1,
    add_all_perrors(Rest,Filenames,ContextOfError,N1,MaxErrors).

extract_perror_pos(pos(_,Filenumber,Srow,Scol,Erow,Ecol),Filenames,POS) :-
    get_filename(Filenumber,Filenames,Filename),!,
    Ecol1 is Ecol+1,
    POS = src_span(Filenumber,Filename,Srow,Scol,Erow,Ecol1,[],[]).
extract_perror_pos(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),_,POS) :- !,
    Ecol1 is Ecol+1,
    POS = src_span(Srow,Scol,Erow,Ecol1,[],[]).
extract_perror_pos(pos(Srow,Scol,Filename),_,POS) :- !,
    POS = src_position_with_filename(Srow,Scol,1,Filename).
extract_perror_pos(pos(Srow,Scol),_,POS) :- !, print(deprecated_pos(Srow,Scol)),nl,
    POS = src_position(Srow,Scol,none,1).
extract_perror_pos(Pos,_,Pos).


get_perror(error(Msg,Pos),error,Msg,Pos) :- !.
get_perror(internal_error(Msg,Pos),internal_error,Msg,Pos) :- !.
get_perror(warning(Msg,Pos),warning,Msg,Pos) :- !.
get_perror(M,_,M,unknown) :- format('!!! unknown perror: ~w~n',[M]).

get_filename(Filenumber,Filenames,Filename) :-
	number(Filenumber),!,
	nth1(Filenumber,Filenames,Filename).
get_filename(N,F,R) :-
   add_internal_error('Cannot get filename: ',get_filename(N,F,R)),
   N=1,F=[R|_]. % return first file if possible

% check if error/warning occured and retract matching error/warning
get_error(Source,ErrMsg) :-
   retract(logged_error_hash(_,Source,ErrMsg,_,_)).
get_warning(Source,ErrMsg) :-
   retract(logged_error_hash(_,warning(Source),ErrMsg,_,_)).

% now the same without actually retracting the error/warning:
check_error_occured(Source,ErrMsg) :- logged_error(Source,ErrMsg,_,_).
check_warning_occured(Source,ErrMsg) :- logged_error(warning(Source),ErrMsg,_,_).

get_error(Msg) :-
    retract(logged_error_hash(_Hash,Source,ErrMsg,Context,Span)),
    (extract_line_col(Span,StartLine,StartCol,EndLine,EndCol)
      -> ( POS = ['; ### Line: ',StartLine,', Column: ',StartCol|POS1],
		   (get_preference(show_full_error_span,true)
			 -> POS1 = ['; ### Until Line: ',EndLine,', Column: ',EndCol, ';;']
			 ;  POS1 = [';;']),
          (extract_file_name(Span,Filename)
            -> PosMsg = ['; ### File: ',Filename | POS]
             ; PosMsg = POS
          ),
           ajoin(PosMsg,EndStr)
         )
      ;  EndStr = ';;'
     ),
    ((Context='' ; Context =[])
      ->  string_concatenate(ErrMsg,EndStr,S0)
       ;  safe_translate_error_term(Context,CString),
          string_concatenate(CString,EndStr,CString2),
          string_concatenate(':  Context = ',CString2,CString3),
          string_concatenate(ErrMsg,CString3,S0)
     ),
    (debug:debug_mode(off) -> Msg = S0
      ;  string_concatenate(Source,': ',S1),
         string_concatenate(S1,S0,Msg)
     ).


get_all_errors(All) :-
    findall(Err,get_error(Err),All),
    All \= [], clear_error_context.

get_all_errors_and_clear(All) :- findall(Err,get_error(Err),All),reset_errors_but_not_scopes,All \= [].
get_all_errors_and_reset(All) :- findall(Err,get_error(Err),All),reset_errors,All \= [].

test_error_occurred(Source) :- logged_error(Source,_,_,_),real_error_occurred,!.

%:- use_module(library(lists),[length/2]). length is builtin

count_errors_occurred(Source,NrOfErrors) :-
  findall(1,logged_error(Source,_,_,_),Ls),
  length(Ls,NrOfErrors).

display_error_statistics :- findall(Source,logged_error(Source,_,_,_),L),
  count_occurences(L,Occ),
  format('Error manager summary: ~w~n',[Occ]).



:- volatile current_context/1.
:- dynamic current_context/1.

/* add context information; will be added to any error messages */
/* will be removed upon backtracking */
%push_error_context(_Ctx) :- !.
set_error_context(Ctx) :- % print(ctx(Ctx)),nl,
  retractall(current_context(_)),
  assert(current_context(Ctx)).
%push_error_context(_Ctx) :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_retract_context(_Ctx))).

clear_error_context :- %print(clearing_ctx),nl,
   retractall(current_context(_)).

%pop_error_context :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_pop_context(_Ctx))).

get_error_context(C) :- findall(Ct,current_context(Ct),C).


get_error_span(Source,Span) :-
   retract(logged_error_spans(Source,Span)).

% a way for external modules to register error locations, e.g., for display by Tcl/Tk HighlightSyntaxErrors
register_error_span(Source,Span) :- assert(logged_error_spans(Source,Span)).

get_error_span_linecol(Line,Col,EndLine,EndCol) :- get_error_span_linecol(_Source,Line,Col,EndLine,EndCol).
get_error_span_linecol(Source,Line,Col,EndLine,EndCol) :-
    get_error_span(Source,Span),  % looks at logged_error_spans and retracts
    extract_line_col(Span,Line,Col,EndLine,EndCol).

get_error_span_linecol_for_main_file(Line,Col,EndLine,EndCol) :-
    get_error_span(_Source,Span),  % looks at logged_error_spans and retracts even if not in main file !
    extract_line_col_for_main_file(Span,Line,Col,EndLine,EndCol).

% just like get_error_span_linecol but does not retract and gets Filename
check_error_span_file_linecol(Source, FullFilename,Line,Col,EndLine,EndCol) :-
    logged_error_spans(Source,Span),
    (extract_file_number_and_name(Span,_,FullFilename) -> true ; FullFilename = unknown),
    extract_line_col(Span,Line,Col,EndLine,EndCol).

% try and extract symbolic (Rodin label) information
extract_symbolic_pos([],_) :- !,fail.
extract_symbolic_pos([H|T],SymbolicPos) :- !,
   posinfo_member(N,[H|T]),
   extract_symbolic_pos(N,SymbolicPos),!.
extract_symbolic_pos(rodinpos(Label,_),Label).  % see also get_texpr_labels in bsyntaxtree
extract_symbolic_pos(rodinpos(Model,Name,_),Label) :- ajoin([Model,':',Name],Label).
extract_symbolic_pos(label(Label),Label).

% choose first sub-expression in case top-expression has no position info (it would be better if extract_info/3 in bsyntaxtree merges position info)
first_sub_expr(negation(A,_),A).
first_sub_expr(conjunct(A,_),A).
first_sub_expr(disjunct(A,_),A).
first_sub_expr(implication(A,_),A).
first_sub_expr(equivalence(A,_),A).
first_sub_expr(exists([A|_],_),A).
first_sub_expr(forall([A|_],_),A).

extract_line_col([],_,_,_,_) :- !,fail.
extract_line_col([H|T],Srow,Scol,Erow,Ecol) :- !,
   posinfo_member(N,[H|T]),
   extract_line_col(N,Srow,Scol,Erow,Ecol),!.
extract_line_col(b(E,_,Infos),Srow,Scol,Erow,Ecol) :- !,
   (extract_line_col(Infos,Srow,Scol,Erow,Ecol) -> true
     ; first_sub_expr(E,A) -> extract_line_col(A,Srow,Scol,Erow,Ecol)).
extract_line_col(span_predicate(P,_,_),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(P,Srow,Scol,Erow,Ecol).
extract_line_col(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),Srow,S1,Erow,E1) :-  number(Srow),
  % _Filenumber is an index into list provided by b_get_all_used_filenames
  number(Scol), number(Erow), number(Ecol),!,
  (Scol>0 -> S1 is Scol-1 ; S1 = Scol),  % the BParser starts column numbering at 1
  E1=Ecol. %Somehow the Ecol info is ok for Tcl
extract_line_col(unknown,_,_,_,_) :- !,fail.
extract_line_col(src_span(_Filenumber,_Filename,StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol)
 :- !, extract_line_col(src_span(StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol).
extract_line_col(src_span(StartLine,SC,EndLine,EC,_Off,[]),StartLine,StartCol,EndLine,EndCol)
 :-  !, StartCol is SC-1, EndCol is EC-1.
extract_line_col(src_span(SL,SC,EL,EC,_Off,_LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartLine=SL,StartCol is SC-1,EndLine = EL, EndCol is EC-1.
 %:- !, StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine = StartLine. %% fix because new library generates erroneous EndLine/EndCol info
% EndCol is EC-1.
extract_line_col(span_info(_Info,Span),SL,SC,EL,EC) :- !, extract_line_col(Span,SL,SC,EL,EC).
extract_line_col(src_span_operator(SP1,SP2),SL,SC,EL,EC) :- !,
 (extract_line_col(SP1,SL,SC,EL,EC) ; extract_line_col(SP2,SL,SC,EL,EC)).
extract_line_col(multi_span(SL,SC,EL,EC,_Off,_Len,TAIL),StartLine,StartCol,EndLine,EndCol) :- !,
  ((StartLine = SL, StartCol is SC-1,
     EndLine = EL, EndCol is EC-1
     )
   ; extract_line_col(TAIL,StartLine,StartCol,EndLine,EndCol)).
extract_line_col(src_position(StartLine,SC,_Off,LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(src_position_with_filename(StartLine,SC,LenBytes,_Filename),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(lineCol(StartLine,StartCol),StartLine,StartCol,EndLine,EndCol) :- !,
   EndLine=StartLine, EndCol=StartCol.
extract_line_col(Struct,StartLine,StartCol,EndLine,EndCol) :-
    \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), extract_line_col(Arg,StartLine,StartCol,EndLine,EndCol),!.
%extract_line_col(S,_,_,_,_) :- print('COULD NOT EXTRACT LINE/COL INFO: '), print(S),
%    functor(S,F,N), print(' '),print(F/N),nl,nl,fail.


/* For TCL TK: extract all line_col information in one go and in 4 lists */
extract_all_line_col(Span, StartLineList,StartColList,EndLineList,EndColList) :-
    findall(lc(SL,SC,EL,EC), extract_line_col(Span,SL,SC,EL,EC),All),
    %print(all(All,Span)),nl,
    split_lc(All,StartLineList,StartColList,EndLineList,EndColList).

split_lc([], [],[],[],[]).
split_lc([lc(SL,SC,EL,EC)|T],[SL|T1],[SC|T2],[EL|T3],[EC|T4]) :-
   split_lc(T,T1,T2,T3,T4).

% TO DO: merge with line_col and in Tcl/Tk use info to highlight proper file !

extract_file_name(Pos,Filename) :- extract_valid_file_number_and_name(Pos,_,Filename).
extract_tail_file_name(Pos,Filename) :-
   extract_valid_file_number_and_name(Pos,_,FullFilename),
   tools:get_tail_filename(FullFilename,Filename).
extract_subsidiary_tail_file_name(Pos,Filename) :-
   extract_valid_file_number_and_name(Pos,Nr,FullFilename),
   not_main_file(Nr,FullFilename),
   tools:get_tail_filename(FullFilename,Filename).

extract_valid_file_number_and_name(Pos,Nr,Filename) :-
   extract_file_number_and_name(Pos,Nr,Filename),
   valid_filename(Filename).
valid_filename(F) :- atom(F). % not unknown(-1) which can come from REPL

% a version of extract_line_col which fails if the error is not in the main file
extract_line_col_for_main_file(Span,Line,Col,EndLine,EndCol) :-
   extract_line_col(Span,Line,Col,EndLine,EndCol),
   \+((extract_file_number_and_name(Span,Nr,FullFilename),
       not_main_file(Nr,FullFilename))).


extract_file_number_and_name(V,_,_) :- var(V),!,fail.
extract_file_number_and_name(value(_),_,_) :- !,fail.
extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), Filenumber,Filename) :-
   bmachine:b_get_all_used_filenames(FN), get_filename(Filenumber,FN,Filename),!.
extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), Filenumber,unknown(Filenumber)) :- !. % unknown(-1) which can come from REPL
extract_file_number_and_name(src_span(_,Filename,_,_,_,_,_,_), unknown, Filename) :- !.
extract_file_number_and_name(src_position_with_filename(_,_,_,Filename), unknown, Filename) :- !.
extract_file_number_and_name(section(RodinFilename), unknown, RodinFilename) :- !.
extract_file_number_and_name([H|T],Filenumber,Filename) :- !,
    posinfo_member(Arg,[H|T]),
    extract_file_number_and_name(Arg,Filenumber,Filename),!.
extract_file_number_and_name(Struct,Filenumber,Filename) :-
    \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args),
    extract_file_number_and_name(Arg,Filenumber,Filename),!.

posinfo_member(Arg,Args) :-  % first look for a nodeid(.) member
  (member(nodeid(POS),Args) -> Arg=POS ; member(Arg,Args)).

%:- use_module(bmachine,[b_get_all_used_filenames/1,b_get_main_filenumber/1]).
position_is_not_in_main_file(pos(_,Filenumber,_,_,_,_)) :- !, % B position info
   not_main_filenumber(Filenumber).
position_is_not_in_main_file(src_span(_,Filename,_,_,_,_,_,_)) :- !,
   specfile:currently_opened_file_status(Main,_), Main \= Filename.
position_is_not_in_main_file(src_position_with_filename(_,_,_,Filename)) :- !,
   specfile:currently_opened_file_status(Main,_), Main \= Filename.
position_is_not_in_main_file(Struct) :- \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), position_is_not_in_main_file(Arg),!.


not_main_file(Filenumber,_) :- number(Filenumber),!, % B Position info
    not_main_filenumber(Filenumber).
not_main_file(_,Filename) :- % probably CSP position info; or unknown
    specfile:currently_opened_file_status(Main,_), \+ same_file_name(Main,Filename). %Main \= Filename.

:- use_module(tools,[get_tail_filename/2]).
same_file_name(F1,F2) :-
   (F1=F2 -> true
  ; get_tail_filename(F1,T1), get_tail_filename(F2,T2), T1=T2).

not_main_filenumber(Filenumber) :-
  bmachine:b_get_main_filenumber(MainFN),
  MainFN \= Filenumber.


% ---------------------------------
:- dynamic backed_up_error/4.

% copy current error facts into backed_up_error facts
backup_errors :-
   retractall(backed_up_error(_,_,_,_)),
   logged_error(Source,S,TContext,Span),
   assert(backed_up_error(Source,S,TContext,Span)),
   fail.
backup_errors.


