% (c) 2015-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(prob_state_predicates, [create_state_pred/4]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module translates ProB\'s (local) states to predicates understandable for CVC4/Z3').

:- use_module(library(lists), [maplist/2]).

:- use_module(probsrc(bsyntaxtree), [get_texpr_id/2,
                                     safe_create_texpr/3,
                                     conjunct_predicates/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(debug), [debug_format/3]).
:- use_module(probsrc(b_global_sets), [is_b_global_constant/3]).
:- use_module(probsrc(custom_explicit_sets), [expand_custom_set_to_list_now/2]).
:- use_module(probsrc(clpfd_interface), [clpfd_domain/3]).

create_state_pred(IDs,LS,S,P) :-
    create_state_preds(IDs,LS,S,Preds),
    conjunct_predicates(Preds,P),
    translate_bexpression(P,PP),
    debug_format(9,'smt_solvers_interface state_pred: ~w~n', [PP]).
create_state_preds([],_,_,[b(truth,pred,[])]).
create_state_preds([ID|IDs],LS,S,[SubPred|Preds]) :-
    get_texpr_id(ID,Identifier),
    member(bind(Identifier,Value),LS), !,
    create_state_preds_aux(ID,Value,SubPred),
    create_state_preds(IDs,LS,S,Preds).
create_state_preds([ID|IDs],LS,S,[SubPred|Preds]) :-
    get_texpr_id(ID,Identifier),
    member(bind(Identifier,Value),S), !,
    create_state_preds_aux(ID,Value,SubPred),
    create_state_preds(IDs,LS,S,Preds).
create_state_preds([_|IDs],LS,S,Preds) :-
    create_state_preds(IDs,LS,S,Preds).

% for integers: create fd interval
create_state_preds_aux(b(identifier(Name),integer,Infos),int(Value),Pred) :- !,
    clpfd_domain(Value,Min,Max),
    (number(Min)
     -> safe_create_texpr(integer(Min),integer,MinValue),
        safe_create_texpr(greater_equal(b(identifier(Name),integer,Infos),MinValue),pred,Greater)
     ;  Greater = b(truth,pred,[])),
    (number(Max)
     -> safe_create_texpr(integer(Max),integer,MaxValue),
        safe_create_texpr(less_equal(b(identifier(Name),integer,Infos),MaxValue),pred,Less)
     ;  Less = b(truth,pred,[])),
     conjunct_predicates([Less,Greater],Pred).
% non-integers: create equality if ground and simple enough
create_state_preds_aux(b(identifier(Name),Type,Infos),Value,Pred) :-
    ground(Value), useful_value_for_state_pred(Value), !,
    Pred = b(equal(b(identifier(Name),Type,Infos),b(value(Value),Type,[])),pred,[]).
% else: create truth (i.e. no futher state information)
create_state_preds_aux(_,_,b(truth,pred,[])).

useful_value_for_state_pred([]). % empty sets
useful_value_for_state_pred([H|T]) :- % sets as lists
    maplist(useful_value_for_state_pred,[H|T]).
useful_value_for_state_pred(avl_set(Val)) :-
    expand_custom_set_to_list_now(avl_set(Val),Values),
    maplist(useful_value_for_state_pred,Values).
useful_value_for_state_pred(pred_true).
useful_value_for_state_pred(pred_false).
useful_value_for_state_pred(global_set(_)).
useful_value_for_state_pred(fd(Num,Name)) :-
    is_b_global_constant(Name,Num,_).
useful_value_for_state_pred((A,B)) :-
    useful_value_for_state_pred(A),
    useful_value_for_state_pred(B).
