% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(state_space,
    [get_state_space_stats/3, gen_new_state_id/1,
     history/1, forward_history/1,
     get_action_trace/1, % old trace/1
     get_action_term_trace/1,
     op_trace_ids/1, add_to_op_trace_ids/1, remove_from_op_trace_ids/1, reset_op_trace_ids/0,
     current_state_id/1, current_expression/2,
     current_options/1, set_current_options/1,

     add_id_at_front/1, add_id_at_end/1,
     add_id_random/1, /* from state_space_open_nodes_c */
     add_id_with_weight/2, add_id_to_process/3,
     pop_id_from_front/1, pop_id_from_end/1, pop_id_oldest/1,
     retract_open_node/1, open_ids_empty/0,

     visited_expression/3, visited_expression/2, visited_expression_id/1,
     find_hashed_packed_visited_expression/3,
     retract_visited_expression/2,
     not_all_transitions_added/1,
     not_invariant_checked/1, set_invariant_checked/1,
     invariant_not_yet_checked/1,
     not_interesting/1,max_reached_for_node/1,
     max_reached_or_timeout_for_node/1,
     use_no_timeout/1, time_out_for_node/1, time_out_for_node/3,
     hash_to_id/2,id_to_marker/2, hash_to_nauty_id/2,
     invariant_violated/1, time_out_for_invariant/1, time_out_for_assertions/1,
     set_invariant_violated/1,

     state_error_exists/0, state_error/3, store_state_error/3,
     set_context_state/1, clear_context_state/0, get_current_context_state/1,
     store_error_for_context_state/2,
     store_abort_error_for_context_state_if_possible/4,

     transition/3,transition/4, any_transition/3,
     store_transition/4,
     deadlocked_state/1, % no outgoing edge
     is_initial_state_id/1, is_concrete_constants_state_id/1,
     multiple_concrete_constants_exist/0,
     out_degree/2,
     operation_not_yet_covered/1, operation_name_not_yet_covered/1,

     transition_info/2, store_transition_infos/2,
     compute_transitions_if_necessary/1,

     state_space_initialise/0, state_space_initialise_with_stats/0,
     state_space_reset/0,
     state_space_add/2, state_space_packed_add/2,
     delete_node/1,

     current_state_corresponds_to_initialised_b_machine/0,
     current_state_corresponds_to_fully_setup_b_machine/0,
     current_state_corresponds_to_setup_constants_b_machine/0,

     specialized_inv/2, reuse_operation/4,
     assert_max_reached_for_node/1, assert_time_out_for_node/3,
     assert_time_out_for_invariant/1, assert_time_out_for_assertions/1,

     set_max_nr_of_new_impl_trans_nodes/1,
     get_max_nr_of_new_impl_trans_nodes/1,
     impl_trans_term/3, impl_trans_term_all/2,
     compute_transitions_if_necessary_saved/1,
     max_nr_of_new_nodes_limit_not_reached/0,

     tcltk_save_state/1, tcltk_load_state/1,
     compute_full_state_space_hash/1,

     execute_id_trace_from_current/3,
     set_trace_by_transition_ids/1,
     extract_term_trace_from_transition_ids/2]).

:- use_module(library(lists)).

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(gensym).
:- use_module(preferences).
:- use_module(tools).
:- use_module(translate,[translate_event/2]).
%:- use_module(state_space_exploration_modes,[compute_hash/3]).

:- use_module(extension('counter/counter')).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module keeps track of the visited states by the animator/model checker.').

% ----------------------------------

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_myheap,false)).
:- use_module(state_space_open_nodes). %% comment in to use only Prolog datastructures
:- else.
:- use_module(state_space_open_nodes_c).  %% comment in to use C++ multimap queue; can make use of HEURISTIC_FUNCTION
:- endif.

% ----------------------------------

get_state_space_stats(NrNodes,NrTransitions,ProcessedNodes) :-
  get_counter(states,N), NrNodes is N+1,
  get_counter(transitions,NrTransitions),
  get_counter(processed_nodes,ProcessedNodes).

gen_new_state_id(Nr) :-
   inc_counter(states,N1), Nr is N1-1.  % only one C call
   %get_counter(states,Nr), inc_counter(states).
reset_state_counter :- reset_counter(states).
reset_state_counter(Nr) :- set_counter(states,Nr).


:- dynamic history/1.
history([]).


:- dynamic forward_history/1.

:- dynamic current_state_id/1.
current_state_id(v_0).
 /* INITIAL STATE, third arg: clp(fd) constraints as generated
   by fd_copy_term */

current_expression(ID,State) :- current_state_id(ID),
   visited_expression(ID,State).

:- use_module(specfile,[state_corresponds_to_initialised_b_machine/1,
                        state_corresponds_to_fully_setup_b_machine/1,
                        state_corresponds_to_set_up_constants/1]).
current_state_corresponds_to_initialised_b_machine:-
   current_expression(_,State),
   specfile:state_corresponds_to_initialised_b_machine(State).

current_state_corresponds_to_fully_setup_b_machine :-
   current_expression(_,State),
   state_corresponds_to_fully_setup_b_machine(State).

current_state_corresponds_to_setup_constants_b_machine :-
   current_expression(_,State),
   specfile:state_corresponds_to_set_up_constants(State).


:- dynamic current_options/1.
current_options([]).

set_current_options(Options) :-
    retractall( current_options(_) ),
    assert( current_options(Options) ).

:- dynamic packed_visited_expression/2.
%packed_visited_expression(v_0,true).

:- use_module(state_packing).

retract_visited_expression(ID,State) :- retract(packed_visited_expression(ID,PState)),
 unpack_state(PState,State).

retractall_visited_expression(ID) :- retractall(packed_visited_expression(ID,_)).

state_space_packed_add(Id,PackedTerm) :- assert(packed_visited_expression(Id,PackedTerm)).

state_space_add(Id,Term) :-
  (pack_state(Term,PackedTerm) -> assert(packed_visited_expression(Id,PackedTerm))
   ; add_internal_error('State packing failed: ',pack_state(Term,_)),
     assert(packed_visited_expression(Id,Term))).

% deprecated:
visited_expression(ID,State,true) :- visited_expression(ID,State).
  % not call(CurBody); for the moment we always have true as last argument

visited_expression(A,B) :- packed_visited_expression(A,PB),
  (unpack_state(PB,R) -> B=R ; add_internal_error('Unpacking state failed: ',unpack_state(PB,R)),R=A).

visited_expression_id(A) :- packed_visited_expression(A,_). % avoid unpacking state



% given a hash and a packed state: find ID (fail if does not exist)
find_hashed_packed_visited_expression(Hash,PackedState,ID) :-
   hash_to_id(Hash,ID),
   (packed_visited_expression(ID,PackedState)
             -> true  /* warning: may instantiate State if not ground */
              ; print(hash_collision(Hash,ID)),nl,fail
    ).


:- dynamic not_invariant_checked/1.
set_invariant_checked(ID) :- %print(inv_checked(ID)),nl,
    retract(not_invariant_checked(ID)).

invariant_not_yet_checked(ID) :-
   not_all_transitions_added(ID) ; /* assumption: if not all transitions added then we haven't checked invariant yet */
   not_invariant_checked(ID).

:- dynamic not_interesting/1.
%not_interesting(v_0).

:- dynamic max_reached_for_node/1.
/* true if not all outgoing transistions were computed due to the limit
 on the number of operations/initialisations computed */
:- dynamic time_out_for_node/3, use_no_timeout/1, time_out_for_invariant/1, time_out_for_assertions/1.

time_out_for_node(ID) :- (var(ID) -> packed_visited_expression(ID,_) ; true),
    (time_out_for_node(ID,_,_) -> true ; fail).

:- dynamic transition/4.
%transition(v_0,a,1,v_1).

store_transition(Org,Action,Dest,Id) :-
        %get_counter(transitions,Id), inc_counter(transitions),
        inc_counter(transitions,Id1), Id is Id1-1, % only one C call
	assert(transition(Org,Action,Id,Dest)).

deadlocked_state(Origin) :- \+ any_transition(Origin,_,_).

is_concrete_constants_state_id(ID) :-
  transition(root,_,ID),
  packed_visited_expression(ID,concrete_constants(_)).

% check if we have multiple constant setups
multiple_concrete_constants_exist :-
  is_concrete_constants_state_id(ID),
  is_concrete_constants_state_id(ID2), ID2 \= ID,!.


is_initial_state_id(InitialStateID) :-
  transition(root,_,State),
  packed_visited_expression(State,P),
  (P = concrete_constants(_)
    -> state_space:transition(State,_,InitialStateID)
    ;  InitialStateID=State).

any_transition(Origin,TransID,Destination) :- transition(Origin,_,TransID,Destination).

transition(Origin,Action,Destination) :- transition(Origin,Action,_TransID,Destination).

:- dynamic transition_info/2.
store_transition_infos([],_TransId).
store_transition_infos([Info|Irest],TransId) :-
	store_transition_info(Info,TransId),
	store_transition_infos(Irest,TransId).
store_transition_info(Info,TransId) :- %print(info(Info,TransId)),nl,
	(keep_transition_info(Info)
	  -> assert(transition_info(TransId,Info))
	  ;  true).

% Do not store path info by default
keep_transition_info(path(_)) :- !,fail.
keep_transition_info(eventtrace(_)) :- !,preference(eventtrace,true).
keep_transition_info(event(_)) :- !,preference(store_event_transinfo,true).
keep_transition_info(_). % store everything else

reset_transition_store :-
	retractall(transition(_,_,_,_)),
	retractall(transition_info(_,_)),
	reset_counter(transitions).

/*
 Version with packing of transitions:
store_transition(Org,Action,Dest,Id) :-
	retract(transition_counter(Id)),
	NewId is Id+1,
	assert(transition_counter(NewId)),
	Action =.. [ActionName|Parameters],
	pack_values(Parameters,PackedParameters),
	assert(packed_transition(Org,ActionName,PackedParameters,Id,Dest)).

transition(Origin,Action,TransID,Destination) :- nonvar(Action),!,
   Action =.. [ActionName|Parameters],
   packed_transition(Origin,ActionName,PackedParameters,TransID,Destination),
   unpack_values(PackedParameters,Parameters).
transition(Origin,Action,TransID,Destination) :-
   packed_transition(Origin,ActionName,PackedParameters,TransID,Destination),
   unpack_values(PackedParameters,Parameters),
   Action =.. [ActionName|Parameters].
any_transition(Origin,TransID,Destination) :- packed_transition(Origin,_,_,TransID,Destination).
*/

% compute out-degree of a node
out_degree(ID,OutDegree) :- findall(0, transition(ID,_,_,_), L), length(L,OutDegree).

operation_name_not_yet_covered(OpName) :- operation_not_yet_covered(OpName).

:- dynamic operation_not_yet_covered/1.
%operation_not_yet_covered(b).

state_error_exists :- state_error(_,_,_),!.
:- dynamic state_error/3, context_state/2.

%state_error([],invariant_violated).

reset_next_state_error_id_counter :- reset_counter(next_state_error_id).
:- use_module(tools_printing, [print_error/1]).
:- use_module(error_manager,[print_error_span/1]).
store_state_error(State,Error,Id) :- state_error(State,Id,Error),!. % do not store identical error twice
store_state_error(State,Error,Id) :-
    %retract( next_state_error_id(Id) ),
    inc_counter(next_state_error_id,Id),
	assert( state_error(State,Id,Error) ).
store_error_for_context_state(Error,Id) :-
	( retract(context_state(State,Errs)) ->
	    (Errs<25 -> store_state_error(State,Error,Id), E1 is Errs+1,assert(context_state(State,E1))
	    ; store_state_error(State,max_state_errors_reached(25),Id),assert(context_state(State,25))
	    )
	; otherwise ->
	    add_internal_error('No known context when calling store_error_for_context_state: ',store_error_for_context_state(Error,Id)),
	    fail).
store_abort_error_for_context_state_if_possible(ErrType,Msg,Term,Span) :-
    %print(store(Msg,Term,Span)),nl,
	( context_state(State,_Errs) ->
	    (error_manager:get_error_context(Context)
	     -> true ; Context = '???'),
	    ((abort_error_for_same_location_exists(State,Id1,ErrType,Msg,Span),
	      abort_error_for_same_location_exists(State,Id2,ErrType,Msg,Span), Id2>Id1  )
	      -> /* two errors of same type, for same state and same source location exists */
	          /* TO DO: maybe merge state errors */
	          simplify_span(Span,Span2),
	          store_state_error(State,abort_error(ErrType,'Further identical errors occurred (not stored !)',Term,span_context(Span2,Context)),_)
	       ;
	         print_error('An abort error occurred !'), print_error(ErrType),
			 print_error(Msg),
			 print_error_term(Term),
			 print_error(context_state_id(State)),
			 print_error_span(Span),
			 (add_new_event_in_error_scope(abort_error(ErrType)) -> true ; true),
	         store_state_error(State,abort_error(ErrType,Msg,Term,span_context(Span,Context)),_)
	    )
	; otherwise ->
	    add_error(ErrType,Msg,Term,Span)).

:- use_module(translate, [translate_error_term/2]).
print_error_term(Term) :- translate_error_term(Term,S), print_error(S).

abort_error_for_same_location_exists(State,Id,ErrType,Msg,Span) :-
    state_error(State,Id,abort_error(ErrType,Msg,_Term2,span_context(Span2,_Ctxt2))),
	same_span_location(Span2,Span).
% should be moved to error_manager ?
same_span_location(span_predicate(Pred,_,_),span_predicate(Pred,_,_)) :- !.
same_span_location(X,X).

% remove state information; quite often the same WD error occurs with slightly different stores
simplify_span(span_predicate(Pred,_,_),R) :- !, R=Pred.
simplify_span(X,X).

clear_context_state :-
	retractall(context_state(_,_)).
set_context_state(State) :-
    %print(setting_ctxt_state(State,Event)),nl,
	retractall(context_state(_,_)), assert(context_state(State,0)).
get_current_context_state(ID) :- context_state(ID,_).

retractall_invariant_violated(State) :-
   retractall(state_error(State,_,invariant_violated)).
invariant_violated(State) :-
	state_error(State,_,invariant_violated).
set_invariant_violated(State) :-
	( invariant_violated(State) -> true
	; time_out_for_invariant(ID) -> print('Timeout for node: '), print(ID),nl,
	                                print('Not setting invariant violation status'),nl
	  ; store_state_error(State,invariant_violated,_)
	 ).

%:- set_invariant_violated([]). % why is this ??


:- dynamic hash_to_id/2.
:- dynamic id_to_marker/2.

:- dynamic hash_to_nauty_id/2.  % used in nauty mode to map nauty id's to hash values

:- dynamic specialized_inv/2.  /* stores whether for a node a specialized invariant
 version could be computed */

 :- dynamic reuse_operation/4. /* when for a state and given operation name we can reuse the operation computed for another state */

:- use_module(hashing).
state_space_startup :- % call once at startup to ensure all counters exist
      counter_init,
      new_counter(states), new_counter(processed_nodes), new_counter(transitions),
      new_counter(next_state_error_id).
state_space_initialise :- counter_init, reset_gennum, reset_gensym,
     new_counter(states), new_counter(processed_nodes), new_counter(transitions),
      new_counter(next_state_error_id),
     reset_state_counter, reset_processed_nodes_counter, reset_next_state_error_id_counter,
     retractall_visited_expression(_),
     reset_open_ids, reset_stored_values,
     retractall(not_invariant_checked(_)),
     retractall(not_interesting(_)),
     retractall(max_reached_for_node(_)),
     retractall(time_out_for_node(_,_,_)),
     retractall(time_out_for_invariant(_)),
     retractall(time_out_for_assertions(_)),
     retractall(use_no_timeout(_)),
     retractall(state_error(_,_,_)),
     clear_context_state,
     reset_transition_store,
     retractall(operation_not_yet_covered(_)),
     retractall(hash_to_id(_,_)),
     retractall(hash_to_nauty_id(_,_)),
     retractall(id_to_marker(_,_)),
     retractall(specialized_inv(_,_)),
     retractall(reuse_operation(_,_,_,_)),
     state_space_add(root,root),
     add_id_at_front(root),
     hashing:my_term_hash(root,RootHash),
     assert(hash_to_id(RootHash,root)),
     %assert(not_invariant_checked(root)),
     state_space_reset.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,state_space_startup,
                           'Initialise Statespace Counters.').
:- register_event_listener(reset_specification,state_space_initialise,
                           'Reset Statespace.').
:- register_event_listener(change_of_animation_mode,state_space_initialise,
                           'Reset Statespace.').

/* A version of reset which checks how much memory is used by each fact */
/* state_space:init_with_stats */
state_space_initialise_with_stats :-
     reset_gennum, reset_gensym, reset_state_counter, reset_processed_nodes_counter,
     reset_next_state_error_id_counter,
     retract_open_ids_with_statistics,
     retract_with_statistics(state_space,[packed_visited_expression(_,_),
                             not_invariant_checked(_),
                             not_interesting(_),
                             max_reached_for_node(_),
                             time_out_for_node(_,_,_),
                             time_out_for_invariant(_),
                             time_out_for_assertions(_),
                             use_no_timeout(_),
                             state_error(_,_,_),
                             transition(_,_,_,_),
                             transition_info(_,_),
                             operation_not_yet_covered(_),
                             hash_to_id(_,_),
                             hash_to_nauty_id(_,_),
                             id_to_marker(_,_),
                             specialized_inv(_,_),
                             reuse_operation(_,_,_,_),
                             history(_), forward_history(_), op_trace_ids(_)]),
     retract_stored_values_with_statistics,
     clear_context_state,
     reset_transition_store,
     state_space_add(root,root),
     add_id_at_front(root),
     %assert(not_invariant_checked(root)),
     state_space_reset.



:- dynamic op_trace_ids/1.
reset_trace :- retractall(op_trace_ids(_)), assert(op_trace_ids([])).
get_action_trace(T) :- trace(T).
get_action_term_trace(PT) :- trace_with_limit(0,T), project_on_action_term(T,PT).
trace(Trace) :- trace_with_limit(1000,Trace).
trace_with_limit(Limit,Trace) :- op_trace_ids(IDT), reverse(IDT,RIDT),
    extract_trace_from_transition_ids(RIDT,root,Limit,[],Trace).

reset_op_trace_ids :- retractall(op_trace_ids(_)), assert(op_trace_ids([])).
add_to_op_trace_ids(OpID) :- retract(op_trace_ids(OpIDS)), assert(op_trace_ids([OpID|OpIDS])).
remove_from_op_trace_ids(OpID) :- retract(op_trace_ids(OpIDS)),
  (OpIDS = [R|Rest] -> assert(op_trace_ids(Rest)), OpID = R
    ; assert(op_trace_ids(OpIDS)),fail).

% translate a list of transition ids (from root) into a list of operation terms
extract_term_trace_from_transition_ids(TransIDListFromRoot,Trace) :-
  extract_trace_from_transition_ids(TransIDListFromRoot,root,0,[],ActionTrace),
  reverse_and_project_on_action_term(ActionTrace,[],Trace).

reverse_and_project_on_action_term([],A,A).
reverse_and_project_on_action_term([action(_,Term)|T],Acc,Res) :- !,
  reverse_and_project_on_action_term(T,[Term|Acc],Res).
reverse_and_project_on_action_term([H|T],Acc,Res) :-
  add_error(reverse_and_project_on_action_term,'Illegal action: ',H),
  reverse_and_project_on_action_term(T,[H|Acc],Res).

project_on_action_term([],[]).
project_on_action_term([action(_,Term)|T],Res) :- !, Res=[Term|TR],
  project_on_action_term(T,TR).
project_on_action_term([H|T],Res) :-
  add_error(project_on_action_term,'Illegal action: ',H),
  project_on_action_term(T,Res).

extract_trace_from_transition_ids([],_CurrentState,_,Trace,Trace).
extract_trace_from_transition_ids([TransId|Rest],CurrentState,Limit,AccTrace,Trace) :-
    compute_op_string(TransId,CurrentState,Limit,OpTerm,OpString,DestState),!,
    extract_trace_from_transition_ids(Rest,DestState,Limit,
                                        [action(OpString,OpTerm)|AccTrace],Trace).
extract_trace_from_transition_ids([TransId|_],CurrentState,_,_,_Trace) :-
    add_error(state_space,'Could not execute transition id: ', TransId:from(CurrentState)),fail.

:- use_module(translate,[translate_event_with_target_id/4]).
compute_op_string(jump(TO),_CurID,_,Term,String,DestID) :- !, Term=jump,String=jump,DestID=TO.
compute_op_string(TransId,CurID,Limit,Term,String,DestID) :- transition(CurID,Term,TransId,DestID),
    translate_event_with_target_id(Term,DestID,Limit,String).

state_space_reset :-
                  reset_trace,
                  retractall(history(_)),
                  retractall(forward_history(_)),
                  retractall(current_state_id(_)),
                  retractall(current_options(_)),
                  assert(history([])),
                  assert(current_state_id(root)).


state_space_clean_all :-
     retractall(state_space_version_in_file(_)),
     retractall_visited_expression(_),
	 reset_open_ids,
	 retractall(not_invariant_checked(_)),
	 retractall(not_interesting(_)),
	 retractall(max_reached_for_node(_)),
	 retractall(time_out_for_node(_,_,_)),
	 retractall(time_out_for_invariant(_)),
	 retractall(time_out_for_assertions(_)),
	 retractall(use_no_timeout(_)),
	 retractall(state_error(_,_,_)),
	 clear_context_state,
     reset_transition_store,
     retractall(operation_not_yet_covered(_)),
     retractall(hash_to_id(_,_)),
     retractall(hash_to_nauty_id(_,_)),
     retractall(id_to_marker(_,_)),
     retractall(specialized_inv(_,_)),
     retractall(reuse_operation(_,_,_,_)),
     retractall(history(_)),
     retractall(forward_history(_)),
     retractall(op_trace_ids(_)),
     retractall(current_state_id(_)),
     retractall(current_options(_)).

% this is only used from within the Tcl/Tk animator at the moment:
delete_node(ID) :-        print(deleting(ID)),nl,
     retractall_visited_expression(ID),
     retractall_invariant_violated(ID),
	 retract_open_node_and_update_processed_nodes(ID),
	 retractall(not_invariant_checked(ID)),
	 retractall(not_interesting(ID)),
	 retractall(max_reached_for_node(ID)),
	 retractall(time_out_for_node(ID,_,_)),
	 retractall(time_out_for_invariant(ID)),
	 retractall(time_out_for_assertions(ID)),
	 retractall(use_no_timeout(ID)),
	 retractall(state_error(ID,_,_)),
     retractall(transition(ID,_,_,_)),
     % to do: check if  operation_not_yet_covered(_) changes
     retract_hash(ID),
     retractall(id_to_marker(ID,_)).

retract_hash(ID) :- retract(hash_to_id(Hash,ID)), retractall(hash_to_nauty_id(_TermHash,Hash)),fail.
retract_hash(_).

assert_max_reached_for_node(Id) :- %print_message(max_reached_for_node(Id)),
    (max_reached_for_node(Id) -> true ; assert(max_reached_for_node(Id))).

assert_time_out_for_node(Id,OpName,TypeOfTimeOut) :- print_message(time_out_for_node(Id,OpName,TypeOfTimeOut)),
    (time_out_for_node(Id,OpName,_) -> true ; assert(time_out_for_node(Id,OpName,TypeOfTimeOut))).
assert_time_out_for_invariant(Id) :- print_message(time_out_for_invariant(Id)),
    (time_out_for_invariant(Id) -> true ; assert(time_out_for_invariant(Id))).
assert_time_out_for_assertions(Id) :- print_message(time_out_for_assertions(Id)),
    (time_out_for_assertions(Id) -> true ; assert(time_out_for_assertions(Id))).

max_reached_or_timeout_for_node(Id) :-
  (max_reached_for_node(Id) ; time_out_for_node(Id,_,_)).
/* ---------------------- */
/* state space saving     */
/* ---------------------- */

:- dynamic state_space_version_in_file/1. %
state_space_version(1).

check_state_space_version :- state_space_version(V),
   (state_space_version_in_file(F) -> true ; F=0),
   (V>F -> add_message(state_space,'Warning: saved state_space may be incompatible with current version: ',F:V) ; true).

tcltk_save_state(File) :-
   print('% saving state to: '), print(File),nl,
   tell(File),
   print_state_space,
   told,
   print_message(done).


:- use_module(tools_printing, [print_dynamic_fact/1,print_dynamic_pred/3]).
print_state_space :-
  state_space_version(V),
  print_dynamic_fact(state_space_version_in_file(V)),
  % TO DO: maybe also save some important preferences, and warn user and/or propose to adapt preferences ?
  print_dynamic_pred(state_space,history,1),
  print_dynamic_pred(state_space,forward_history,1),
  print_dynamic_pred(state_space,op_trace_ids,1),
  print_dynamic_pred(state_space,current_state_id,1),
  print_dynamic_pred(state_space,current_options,1),
  print_dynamic_pred(state_space,packed_visited_expression,2),
  print_dynamic_pred(state_space,not_invariant_checked,1),
  print_dynamic_pred(state_space,not_interesting,1),
  print_dynamic_pred(state_space,max_reached_for_node,1),
  print_dynamic_pred(state_space,time_out_for_node,3),
  print_dynamic_pred(state_space,use_no_timeout,1),
  print_dynamic_pred(state_space,transition,4),
  print_dynamic_pred(state_space,transition_info,2),
  print_dynamic_pred(state_space,operation_not_yet_covered,1),
  print_dynamic_pred(state_space,state_error,3),
  print_state_space_open_nodes,
  print_stored_values,
  get_counter(states,X),
  write_term(saved_gennum_count(X),[quoted(true)]),print('.'),nl.

saved_gennum_count(99999).

/* ---------------------- */
/* state space loading    */
/* ---------------------- */

tcltk_load_state(File) :- state_space_clean_all,
   print('Loading: '), print(File),nl,
   user_consult_without_redefine_warning(File), % this will read in bind_skeleton/2, ..., next_value_id/1
   check_state_space_version,
   print('Generating open node info'),nl,
   transfer_open_node_info,
   print('Transfer state packing info'),nl,
   transfer_state_packing_info,
   print('Recomputing hash index'),nl,
   recompute_all_hash,
   (saved_gennum_count(X) -> reset_state_counter(X) ; true),
   reset_processed_nodes_counter, % TO DO: restore or save it
   reset_next_state_error_id_counter, % DITTO
   print('Done'),nl,!.
tcltk_load_state(File) :-
   add_error(tcltk_load_state,'Could not load state from file: ',File),
   state_space_initialise.

:- dynamic not_all_z_saved/1, not_all_transitions_added_saved/1.

% transfer facts read into state_space into other modules:
transfer_open_node_info :- retract(not_all_z_saved(X)), %print(not_all_z(X)),nl,
   assert_not_all_z(X),fail.
transfer_open_node_info :- retract(not_all_transitions_added_saved(X)),
   assert_not_all_transitions_added(X),fail.
transfer_open_node_info.
% now for transferring to state_packing module info generated by print_stored_values
transfer_state_packing_info :- retract(bind_skeleton(X,Y)), %print(skel(X)),nl,
   assert(state_packing:bind_skeleton(X,Y)),fail.
transfer_state_packing_info :- retract(stored_value(X,Y)),
   assert(state_packing:stored_value(X,Y)),fail.
transfer_state_packing_info :- retract(stored_value_hash_to_id(X,Y)),
   assert(state_packing:stored_value_hash_to_id(X,Y)),fail.
transfer_state_packing_info :- retract(next_value_id(X)),
   assert(state_packing:next_value_id(X)),fail.
transfer_state_packing_info.

recompute_all_hash :-
  retractall(hash_to_id(_,_)),retractall(id_to_marker(_,_)),
  retractall(hash_to_nauty_id(_,_)),
  visited_expression(ID,StateTemplate),
  state_space_exploration_modes:compute_hash(StateTemplate,Hash,Marker),
  assert(hash_to_id(Hash,ID)),
  assert(id_to_marker(ID,Marker)),
  fail.
recompute_all_hash.

:- use_module(hashing,[my_term_hash/2]).
% generates a hash for the entire state space not depending on the order in which states where added
compute_full_state_space_hash(Hash) :-
   findall(Hash,hash_to_id(Hash,_),ListOfHashCodes),
   sort(ListOfHashCodes,SortedList),
   my_term_hash(SortedList,Hash).
% TO DO: also provide transition hashes

:- use_module(tools_meta,[safe_on_exception/3]).
user_consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off),
    (safe_on_exception(Exc,consult(File),(nl,print('Exception occurred:'),print(Exc),nl,fail))
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    OK=true.



execute_id_trace_from_current(ID,OpIDL,StateIDList) :-
   current_state_id(CurID),
   reverse([CurID|StateIDList],Rev),
   Rev = [Dest|TRev],  (Dest==ID -> true ; (print(not_eq(Dest,ID)),nl)),
   retract(history(H)), retractall(forward_history(_)),
   append(TRev,H,NewH),
   assert(history(NewH)),
   retract(op_trace_ids(OldTrace)),
   reverse(OpIDL,NewTrace),
   append(NewTrace,OldTrace,Trace),
   assert(op_trace_ids(Trace)),
   retractall(current_state_id(_)),
   assert(current_state_id(ID)).
   %execute_trace_to_node(OpL,StateIDList).   /* <----- BOTTLENECK FOR LONG SEQUENCES */
%generate_trace([],Acc,Acc).
%generate_trace([OpTerm|T],Acc,Res) :-
%   translate_event(OpTerm,OpString),
%   generate_trace(T,[action(OpString,OpTerm)|Acc],Res).

set_trace_by_transition_ids(TransitionIds) :-
    extract_history_from_transition_ids(TransitionIds,root,[],[],Last,History,OpTrace),
    %visited_expression(Last,LastState,LastCond),
    retractall(history(_)), retractall(forward_history(_)),
    retractall(current_state_id(_)),
    retractall(op_trace_ids(_)),
    assert(history(History)), assert(op_trace_ids(OpTrace)),
    assert(current_state_id(Last)).

extract_history_from_transition_ids([],CurrentState,History,Trace,CurrentState,History,Trace).
extract_history_from_transition_ids([TransId|Rest],CurrentState,AccHist,AccTrace,Last,History,Trace) :-
    transition(CurrentState,_,TransId,DestState),!,
    extract_history_from_transition_ids(Rest,DestState,[CurrentState|AccHist],
                                        [TransId|AccTrace],Last,History,Trace).
extract_history_from_transition_ids([TransId|_],CurrentState,_,_,_,_,_Trace) :-
    add_error(state_space,'Could not execute transition id: ', TransId:from(CurrentState)),fail.



/* --------------------------------- */
:- dynamic max_nr_of_new_nodes/1.

% negative number or non-number signifies no limit
set_max_nr_of_new_impl_trans_nodes(MaxNrOfNewNodes) :-
     retractall(max_nr_of_new_nodes(_)),
     ((number(MaxNrOfNewNodes), MaxNrOfNewNodes>=0)
      -> assert(max_nr_of_new_nodes(MaxNrOfNewNodes))
      ;  true). % no need to store limit

get_max_nr_of_new_impl_trans_nodes(MaxNrOfNewNodes) :-
     (max_nr_of_new_nodes(Max) -> MaxNrOfNewNodes=Max; MaxNrOfNewNodes = 0).

impl_trans_term(From,ActionAsTerm,To) :-
   compute_transitions_if_necessary_saved(From),
   transition(From,ActionAsTerm,_TID,To).

impl_trans_term_all(From,Ops) :-
   compute_transitions_if_necessary_saved(From),
   findall(op(Id,ActionAsTerm,To),
	   transition(From,ActionAsTerm,Id,To),
	   Ops).

compute_transitions_if_necessary_saved(From) :-
   on_exception(error(forced_interrupt_error('User has interrupted the current execution'),_),
                compute_transitions_if_necessary(From),
                user:process_interrupted_error_message).

compute_transitions_if_necessary(From) :-
    not_all_transitions_added(From),!,
    decrease_max_nr_of_new_nodes(From),
    %user:do_one_gui_event,
    user:compute_all_transitions_if_necessary(From,false).
compute_transitions_if_necessary(_From).

decrease_max_nr_of_new_nodes(ID) :-
    retract(max_nr_of_new_nodes(Max)),!,
    ( Max>0 ->
        NewMax is Max-1,
        assert(max_nr_of_new_nodes(NewMax))
     ; Max=0 -> NM is -1,
        assert(max_nr_of_new_nodes(NM)),
        add_warning(state_space,'Maximum number of new nodes reached for CTL/LTL/refinement check, node id = ',ID),
        %trace,
        fail
    ; otherwise ->  % negative number: re-assert and fail
        assert(max_nr_of_new_nodes(Max)),
        fail).
decrease_max_nr_of_new_nodes(_). % no limit stored; just proceed

% will be called from TCL/TK side
max_nr_of_new_nodes_limit_not_reached :-
    max_nr_of_new_nodes(N),N>0.

:- use_module(specfile,[b_or_z_mode/0]).
retract_open_node(NodeID) :- retract_open_node_and_update_processed_nodes(NodeID),
  (b_or_z_mode -> assert(not_invariant_checked(NodeID)) ; true).

reset_processed_nodes_counter :- reset_counter(processed_nodes).
%reset_processed_nodes_counter(Nr) :- set_counter(processed_nodes,Nr).

retract_open_node_and_update_processed_nodes(NodeID) :-
   retract_open_node_direct(NodeID),
   inc_processed.

inc_processed :-
   inc_counter(processed_nodes).

pop_id_from_front(ID) :- pop_id_from_front_direct(ID), inc_processed.
pop_id_from_end(ID) :- pop_id_from_end_direct(ID), inc_processed.
pop_id_oldest(ID) :- pop_id_oldest_direct(ID), inc_processed.

/* --------------------------------- */

%
% Code to compute equivalence classes
%  using the standard DFA minimization algorithm

:- dynamic equivalent/2.
% state_space:compute_equivalence_classes
:- public compute_equivalence_classes/0.

compute_equivalence_classes :- init_equi,
   split_equivalence_classes,nl,
   print_equi.

print_equi :- state_space:equivalent(A,B), visited_expression(A,State),
  visited_expression(B,StateB),
  nl,
  print(A), print(' : '),  print(State),nl,
  print(B), print(' : '),  print(StateB),nl,fail.
print_equi.

init_equi :- retractall(equivalent(_,_)),
  packed_visited_expression(ID,_State),
  \+ not_all_transitions_added(ID),
  findall(Action,transition(ID,Action,_,_),List),
  packed_visited_expression(ID2,_S2), ID2 @> ID,
  \+ not_all_transitions_added(ID2),
  findall(Action,transition(ID2,Action,_,_),List),
  assert(equivalent(ID,ID2)), % they have the same signature
  %print(equivalent(ID,ID2)),nl,
  fail.
init_equi :- print(finished_initialising),nl.

split_equivalence_classes :- retractall(echange),
   equivalent(ID1,ID2),
   transition(ID1,A,_,Dest1),
   transition(ID2,A,_,Dest2),
   \+ check_equi(Dest1,Dest2),
   retract(equivalent(ID1,ID2)), % splitting class
  % print(diff(ID1,ID2, A, Dest1, Dest2)),nl,
   assert_echange,
   fail.
split_equivalence_classes :- echange -> split_equivalence_classes ; true.

:- dynamic echange/0.
assert_echange :- echange -> true ; assert(echange),print('.'),flush_output.

check_equi(A,B) :- A=B -> true ; A @<B -> equivalent(A,B) ; equivalent(B,A).



