:- module(b_synthesis,[start_synthesis_from_ui/7]).

:- use_module(synthesis(logging)).
:- use_module(synthesis(constraints)).
:- use_module(synthesis(library_setup)).
:- use_module(synthesis(location_vars_to_program)).
:- use_module(synthesis(synthesis_util)).

:- use_module(probsrc(bmachine)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(translate),[translate_bexpression/2,translate_machine/3]).
:- use_module(probsrc(solver_interface),[solve_predicate/3]).
:- use_module(probsrc(btypechecker),[btype/7]).

:- use_module(library(lists)).
:- use_module(library(sets),[subtract/3,intersection/3]).
:- use_module(library(terms),[term_size/2]).

:- use_module(extension('plspec/plspec/plspec')).
:- use_module(extension('plspec/plspec/plspec_checker')).

:- defspec(solution_bindings,one_of([list(compound(binding(any,any,any))),list(compound(bind(any,any)))])).
:- defspec(solution_compound,compound(solution(solution_bindings))).
:- defspec(ast_example,list(ast)).
:- defspec(ast_examples,list(ast_example)).
:- defspec(ast,compound(b(any,any,list(any)))).
:- defspec(pair(X,Y),compound(','(X,Y))).
%:- enable_all_spec_checks.

:- dynamic synthesis_context/8.
:- volatile synthesis_context/8.

% Start synthesis from the UI based on the ProB2 Java API. Returns the adapted machine code with
% the generated changes if synthesis succeeded. Otherwise the unchanged machine code is returned.
% TODO: spec this predicate
start_synthesis_from_ui(ConsiderIf,Operation,SynthesisType,Positive,Negative,NewMachineAtom,Distinguishing) :-
    preferences:set_preference(randomise_enumeration_order,true) ,
    % split examples that are tuples of input and output state
    split_input_output_tuples(Positive,PosInput,PosOutput) ,
    split_input_output_tuples(Negative,NegInput,NegOutput) ,
    start_synthesis_from_ui_aux(ConsiderIf,Operation,SynthesisType,(PosInput,PosOutput),(NegInput,NegOutput),
                                NewMachineAtom,Distinguishing).
start_synthesis_from_ui_aux(ConsiderIf,Operation,action,(PosInput,PosOutput),(NegInput,NegOutput),
                            NewMachineAtom,Distinguishing) :-
    % synthesis type is action but the simultaneous generation of a guard has been suspended
    on_exception(error(existence_error(_,_),_),synthesis_context(_,PreDistStateCache,_,PreCondVars,ValidInvariantsOrGuards,
                                                                 guard,OldPosInput,OldNegInput),fail) ,
    retractall(synthesis_context(_,_,_,_,_,guard,_,_)) ,
    keep_asserted_initial_examples(PosInput,NegInput,OldPosInput,OldNegInput,NewPosInput,NewNegInput) ,
    maplist(get_varname_from_id_node,PreCondVars,PreVarNames) ,
    % therefore we start/restart synthesis of a guard
    prepare_synthesis_of_guard_by_examples(PreVarNames,NewPosInput,NewNegInput,PrePositive,PreNegative,PreLibrary,
                                           PreParameterAmount,PreM1,PreCurrentVars,PreBehavioralConstraint,
                                           PreLocationVars,PreSolution) ,
    synthesis_from_examples(b(truth,pred,[]),PreDistStateCache,PreLibrary,1,PreCurrentVars,PreCurrentVars,
                            ValidInvariantsOrGuards,guard,PrePositive,PreNegative,PreBehavioralConstraint,PreLocationVars,
                            PreM1,PreParameterAmount,PreSolution,PreCondition,NewDistStateCache) ,
    distinguishing_state_or_synthesize_action(ConsiderIf,Operation,PreCondition,NewDistStateCache,(PosInput,PosOutput),
                                              (NegInput,NegOutput),NewMachineAtom,Distinguishing).
start_synthesis_from_ui_aux(ConsiderIf,Operation,action,(PosInput,PosOutput),(NegInput,NegOutput),
                            NewMachineAtom,Distinguishing) :-
    % check if the machine already satisfies the behavior and relax a guard or synthesize a new operation
    \+on_exception(error(existence_error(_,_),_),synthesis_context(_,_,_,_,_,guard,_,_),fail) ,
    behavior_satisfied_or_extend_machine(ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing).
start_synthesis_from_ui_aux(_,Operation,Type,(PosInput,_),(NegInput,_),NewMachineAtom,Distinguishing) :-
    Type \= action ,
    % go on synthesizing the guard/invariant using the validated distinguishing state
    on_exception(error(existence_error(_,_),_),synthesis_context(_,PreDistStateCache,_,PreCondVars,ValidInvariantsOrGuards,
                                                                 Type,OldPosInput,OldNegInput),fail) ,
    retractall(synthesis_context(_,_,_,_,_,Type,_,_)) ,
    keep_asserted_initial_examples(PosInput,NegInput,OldPosInput,OldNegInput,NewPosInput,NewNegInput) ,
    maplist(get_varname_from_id_node,PreCondVars,PreVarNames) ,
    prepare_synthesis_of_guard_by_examples(PreVarNames,NewPosInput,NewNegInput,PrePositive,PreNegative,PreLibrary,
                                           PreParameterAmount,PreM1,PreCurrentVars,PreBehavioralConstraint,
                                           PreLocationVars,PreSolution) ,
    synthesis_from_examples(b(truth,pred,[]),PreDistStateCache,PreLibrary,1,PreCurrentVars,PreCurrentVars,
                            ValidInvariantsOrGuards,guard,PrePositive,PreNegative,PreBehavioralConstraint,PreLocationVars,
                            PreM1,PreParameterAmount,PreSolution,Synthesized,_) ,
    Synthesized \= none ,
    prepare_solution(Operation,action,Synthesized,NewMachineAtom,Distinguishing).
start_synthesis_from_ui_aux(ConsiderIf,Operation,SynthesisType,(PosInput,PosOutput),(NegInput,NegOutput),
                            NewMachineAtom,Distinguishing) :-
    % no context asserted, we start a new synthesis instance
    \+on_exception(error(existence_error(_,_),_),synthesis_context(_,_,_,_,_,_,_,_),fail) ,
    start_specific_synthesis(ConsiderIf,Operation,SynthesisType,PosInput,NegInput,PosOutput,NegOutput,Synthesized) , ! ,
    Synthesized \= none ,
    prepare_solution(Operation,SynthesisType,Synthesized,NewMachineAtom,Distinguishing).

% Check if an action already satisfies the provided behavior or if a guard just has to be relaxed.
% Otherwise, synthesize a new action.
behavior_satisfied_or_extend_machine(ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing) :-
    PosInput = [Example|_] ,
    get_current_vars_from_example(Example,CurrentVars) ,
    check_transitions_against_machine_operations(CurrentVars,PosInput,PosOutput,Result) ,
    behavior_satisfied_or_extend_machine_aux(Result,ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing).
% An operation satisfies the provided behavior. We pass the term operaiton_satisfied/1 to the UI and no distingushing example.
behavior_satisfied_or_extend_machine_aux((Operation,done),_,_,_,_,_,_,operation_satisfied(Operation),none).
behavior_satisfied_or_extend_machine_aux((Operation,guard),ConsiderIf,_,PosInput,NegInput,PosOutput,NegOutput,
                                         NewMachineAtom,Distinguishing) :-
     behavior_satisfied_or_extend_machine_aux2(guard,ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing).
behavior_satisfied_or_extend_machine_aux(no_match,ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,
                                         NewMachineAtom,Distinguishing) :-
    behavior_satisfied_or_extend_machine_aux2(action,ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing).
% Synthesize guard or action.
behavior_satisfied_or_extend_machine_aux2(SynthesisType,ConsiderIf,Operation,PosInput,NegInput,PosOutput,NegOutput,NewMachineAtom,Distinguishing) :-
      start_specific_synthesis(ConsiderIf,Operation,SynthesisType,PosInput,NegInput,PosOutput,NegOutput,Synthesized) , ! ,
      Synthesized \= none ,
      prepare_solution(Operation,SynthesisType,Synthesized,NewMachineAtom,Distinguishing).

% Synthesizing the action has been suspended and we synthesized a guard.
% We derived a distinguishing state.
distinguishing_state_or_synthesize_action(_,Operation,PreCondition,_,_,_,NewMachineAtom,Distinguishing) :-
    is_distinguishing(PreCondition) ,
    prepare_solution(Operation,guard,PreCondition,NewMachineAtom,Distinguishing).
% Synthesizing the guard either succeded or failed. The latter means, we can't pass a guard. Thus, we use the old one.
% We cached the distinguishing transitions so we try to go on synthesizing the action and maybe
% find a guard later depending on the distinguishing interaction with the user.
distinguishing_state_or_synthesize_action(ConsiderIf,Operation,PreCondition,NewDistStateCache,(PosInput,PosOutput),
                                          (NegInput,NegOutput),NewMachineAtom,Distinguishing) :-
    % assert the synthesized precondition to the currently suspended synthesis context of the action and restart it
    synthesis_context(DistTransitionCache,_,CurrentVars,PreCondVars,OldPreCond,action,Input,Output) ,
    new_or_old_precondition(OldPreCond,PreCondition,NewPreCond) ,
    retractall(synthesis_context(_,_,_,_,_,action,_,_)) ,
    assert(synthesis_context(DistTransitionCache,NewDistStateCache,CurrentVars,PreCondVars,NewPreCond,action,
                             Input,Output)) ,
    start_synthesis_from_ui_aux(ConsiderIf,Operation,action,(PosInput,PosOutput),(NegInput,NegOutput),
                                NewMachineAtom,Distinguishing).
new_or_old_precondition(OldPreCond,none,OldPreCond).
new_or_old_precondition(_,PreCond,PreCond) :- PreCond \= none.

% Synthesis returned a distinguishing state or transition. The machine code is not changed.
prepare_solution(_,_,state(State),none,state(TranslatedState)) :-
    maplist(translate_example_ast,State,TranslatedState).
prepare_solution(_,_,transition(Input,Output),none,transition(TranslatedInput,TranslatedOutput)) :-
    maplist(translate_example_ast,Input,TranslatedInput) ,
    maplist(translate_example_ast,Output,TranslatedOutput).
% Succeeded synthesizing a program fitting the desired behavior. Apply the solution to the machine code.
prepare_solution(Operation,SynthesisType,Synthesized,NewMachineAtom,distinguishing(none)) :-
    full_b_machine(Machine) ,
    apply_solution_to_machine(Machine,Operation,SynthesisType,Synthesized,NewMachineAst) ,
    translate_machine(NewMachineAst,Codes,_) , atom_codes(NewMachineAtom,Codes) ,
    retractall(synthesis_context(_,_,_,_,_,_,_,_)).

% Return a tuple of machine variable name and value for a given state's value ast.
translate_example_ast(ExampleAst,(MachineVarName,TranslatedValue)) :-
    translate_bexpression(ExampleAst,TranslatedValue) ,
    get_varname_from_node_info(ExampleAst,MachineVarName).

apply_solution_to_machine(Machine,_,_,none,Machine).
% TODO: replace existing operation's substitution? by now just appending
apply_solution_to_machine(machine(MachineName,Content),_,action,Synthesized,NewMachine) :-
    member(operation_bodies/OperationBodies,Content) ,
    subtract(Content,[operation_bodies/_],TempContent) ,
    apply_solution_to_machine_aux(action,OperationBodies,Synthesized,NewOperationBodies) ,
    append(TempContent,[operation_bodies/NewOperationBodies],NewContent) ,
    NewMachine = machine(MachineName,NewContent).
apply_solution_to_machine(machine(MachineName,Content),Operation,guard,Synthesized,NewMachine) :-
    member(operation_bodies/OperationBodies,Content) ,
    subtract(Content,[operation_bodies/_],TempContent) ,
    apply_solution_to_machine_aux(guard,Operation,Synthesized,OperationBodies,NewOperationBodies) ,
    append(TempContent,[operation_bodies/NewOperationBodies],NewContent) ,
    NewMachine = machine(MachineName,NewContent).
apply_solution_to_machine(machine(MachineName,Content),_,invariant,Synthesized,NewMachine) :-
    subtract(Content,[invariant/_],TempContent) ,
    NewMachine = machine(MachineName,[invariant/Synthesized|TempContent]). % TODO: fix order of invariant

apply_solution_to_machine_aux(action,OperationBodies,Synthesized,NewOperationBodies) :-
    is_list(Synthesized) ,
    append(Synthesized,OperationBodies,NewOperationBodies).
apply_solution_to_machine_aux(action,OperationBodies,Synthesized,[Synthesized|OperationBodies]) :-
    \+is_list(Synthesized).

apply_solution_to_machine_aux(guard,OperationName,Synthesized,[Operation|T],[NewOperation|T]) :-
    replace_precondition_if_operation_fits(OperationName,Operation,Synthesized,NewOperation) , !.
apply_solution_to_machine_aux(guard,OperationName,Synthesized,[Operation|T],[Operation|NT]) :-
    apply_solution_to_machine_aux(guard,OperationName,Synthesized,T,NT).

replace_precondition_if_operation_fits(OperationName,Operation,Synthesized,NewOperation) :-
    Operation = b(Node,Type,Info) ,
    Node = operation(OperationNameId,X,Y,Body) ,
    get_varname_from_id_node(OperationNameId,IdName) ,
    IdName = op(OperationName) , % operation fits
    replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) ,
    NewOperation = b(operation(OperationNameId,X,Y,NewBody),Type,Info).
% Replace or create a new precondition if the given operation name is equal to the one of the operation.
replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) :-
    Body = b(precondition(_,Substitutions),subst,Info) , ! ,
    NewBody = b(precondition(Synthesized,Substitutions),subst,Info).
replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) :-
    Body = b(Substitutions,subst,Info) ,
    NewBody = b(precondition(Synthesized,Substitutions),subst,Info).

% The Java UI provides tuples of input and output examples.
split_input_output_tuples(InputOutputTupleList,InputList,OutputList) :-
    split_input_output_tuples(InputOutputTupleList,[],[],InputList,OutputList).
split_input_output_tuples([],InputListAcc,OutputListAcc,InputListAcc,OutputListAcc).
split_input_output_tuples([(Input,Output)|T],InputListAcc,OutputListAcc,InputList,OutputList) :-
    maplist(variable_parsedvalue_tuple_to_ast,Input,InputAsts) ,
    maplist(variable_parsedvalue_tuple_to_ast,Output,OutputAsts) ,
    accumulate_input_output_if_non_empty(InputAsts,InputListAcc,NewInputListAcc) ,
    accumulate_input_output_if_non_empty(OutputAsts,OutputListAcc,NewOutputListAcc) ,
    split_input_output_tuples(T,NewInputListAcc,NewOutputListAcc,InputList,OutputList).

accumulate_input_output_if_non_empty([],Acc,Acc).
accumulate_input_output_if_non_empty(Asts,Acc,[Asts|Acc]) :- Asts \= [].

% Given a tuple of variable name and parsed term. Create the ast using the typechecker and
% add synthesis/2 term to the node's info describing the corresponding machine variable.
variable_parsedvalue_tuple_to_ast((VarName,ParsedValue),Ast) :-
    btype(bsynthesis,ParsedValue,openenv(_,env_empty),_,TypedAst,[],_) ,
    add_texpr_infos(TypedAst,[synthesis(machinevar,VarName)],Ast).

% Synthesis returned a distinguishing transition.
post_synthesize_action(_,_,_,_,Synthesized,Synthesized) :-
    is_distinguishing(Synthesized) ,
    synthesis_log(solution(_,Synthesized)).
% Synthesis succeded. If no guard has been synthesized simultaneously but at least one negative input
% example is given we generate an appropriate guard afterwards.
post_synthesize_action(Operation,Vars,InitialInput,NegDistInput,Synthesized,FinalAction) :-
    \+is_distinguishing(Synthesized) ,
    maplist(get_varname_from_id_node,Vars,VarNames) ,
    synthesize_guard_if_necessary(VarNames,InitialInput,NegDistInput,Synthesized,Action) ,
    finalize_synthesized_action(Operation,Action,FinalAction) ,
    synthesis_log(solution(action,FinalAction)).

keep_asserted_initial_examples(Input,Output,OldInput,OldOutput,NewInput,NewOutput) :-
    keep_asserted_initial_states(Input,OldInput,NewInput) ,
    keep_asserted_initial_states(Output,OldOutput,NewOutput).
keep_asserted_initial_states([],_,[]).
keep_asserted_initial_states([State|T],OldStates,[NewState|NT]) :-
    contains_equivalent_state(State,OldStates,EquivalentState) , ! ,
    NewState = EquivalentState ,
    keep_asserted_initial_states(T,OldStates,NT).
keep_asserted_initial_states([State|T],OldStates,[State|NT]) :-
    keep_asserted_initial_states(T,OldStates,NT).

%% synthesize actions, invariants or guards
% We already checked the machine operations and synthesize a new operation.
% The first argument defines if we want to check for if-statements, i.e. using synthesize_action_loop/4,
% or the user knows that we don't need an if-statement so that we consider all example transitions at once.
start_specific_synthesis(yes,_Operation,action,_Input,_,_Output,_,none) :-
    true.
    %synthesize_action_loop(Input,Output,[],Tuples) ,
% TODO: fix this to work with the java ui
    %findall(Action,member((Action,_),Tuples),Actions) ,
    %finalize_synthesized_actions(Operation,Actions,FinalActions) ,
    %synthesis_log(solution(action,FinalActions)).
start_specific_synthesis(no,Operation,action,PosInput,NegInput,PosOutput,_,Action) :-
    % no context asserted
    \+on_exception(error(existence_error(_,_),_),synthesis_context(_,_,_,_,_,_,_,_),fail) ,
    mark_examples_as_initial(PosInput,PosOutput,InitialInput,InitialOutput) ,
    prepare_synthesis_of_action_by_examples(InitialInput,InitialOutput,Library,ParameterAmount,M1,CurrentVars,
                                            BehavioralConstraint,LocationVars,Solution) ,
    % use the negative input states for the distinguishing state cache
    synthesis_from_examples(b(truth,pred,[]),([],NegInput),Library,1,CurrentVars,CurrentVars,any,action,InitialInput,
                            InitialOutput,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,Synthesized,(_,NegDistInput)) ,
    post_synthesize_action(Operation,CurrentVars,InitialInput,NegDistInput,Synthesized,Action).
start_specific_synthesis(no,Operation,action,PosInput,NegInput,PosOutput,_,Action) :-
    % context asserted, synthesis has been suspended
    on_exception(error(existence_error(_,_),_),synthesis_context(DistTransitionCache,DistStateCache,CurrentVars,PreCondVars,
                                                                 ValidInvariantsOrGuards,action,OldInput,OldOutput),fail) ,
    keep_asserted_initial_examples(PosInput,PosOutput,OldInput,OldOutput,NewPosInput,NewPosOutput) ,
    start_specific_synthesis_aux(Operation,action,NewPosInput,NegInput,NewPosOutput,_,DistTransitionCache,DistStateCache,
                                 CurrentVars,PreCondVars,ValidInvariantsOrGuards,Action).

% if we synthesize an invariant we reduce the examples to violating vars only
start_specific_synthesis(_,Operation,invariant,PosInput,NegInput,PosOutput,NegOutput,Synthesized) :-
    % use PosInput because at least one invariant violating state has been graduated as valid by the user
    get_violated_and_valid_machine_invariants(PosOutput,ViolatedInvariants,ViolatingVars,ValidInvariants) ,
    findall(Name,(member(IDNode,ViolatingVars) , IDNode = b(identifier(Name),_,_)),VarNames) ,
    synthesis_log(involved_variables(VarNames)) ,
   % for synthesizing invariants we also need output states
    append(PosInput,PosOutput,TempPositive) ,
    append(NegInput,NegOutput,TempNegative) ,
    get_distinct_states(invariant,TempPositive,TempNegative,Positive,Negative) ,
    % don't use an example more than one time, only possible for invariant synthesis
    % because we split every transition in two examples (input,output)
    remove_dups(Positive,Pos) , remove_dups(Negative,Neg) ,
    maplist(reduce_input_to_violating_vars(VarNames),Pos,TRPositive) ,
    maplist(reduce_input_to_violating_vars(VarNames),Neg,RNegative) ,

    TRPositive = [Example|_] ,
    library_setup(invariant,1,Example,Library) ,

    get_valid_states(5,TValidStates) ,
    maplist(reduce_input_to_violating_vars(VarNames),TValidStates,ValidStates) ,
    append(TRPositive,ValidStates,RPositive) ,
    remove_dups(RPositive,NRPositive) ,
% TODO: what about several invariant violating examples that are graduated to be valid?
    % get the example that violates an invariant but has been graduated to be valid by the user
    get_invariant_violating_example(ViolatedInvariants,PosInput,PosOutput,(In,Out)) ,
    % get counter example
    get_invariant_violating_transition(Operation,ViolatingVars,ViolatedInvariants,In,Out,Counter) ,
    start_specific_synthesis_aux(invariant,ValidInvariants,VarNames,NRPositive,RNegative,ViolatingVars,Library,Counter,Synthesized).

start_specific_synthesis(_,Operation,guard,PosInput,NegInput,_,NegOutput,SynthesizedGuard) :-
    % try to reduce input examples by checking the operations guard
    get_guard_list_from_operation(Operation,Guards) ,
    get_involved_variables(Guards,PosInput,NegOutput,InvolvedVarNames) ,
    % filter partial guards that are valid for the examples, and thus, do not need to be relaxed
    findall(ValidGuard,(member(ValidGuard,Guards) , find_identifier_uses(ValidGuard,[],UsedIDs) ,
                        intersection(UsedIDs,InvolvedVarNames,[])),ValidGuards) ,
    mark_examples_as_initial(PosInput,NegInput,InitialPosInput,InitialNegInput) ,
    get_valid_transitions_for_operation(10,Operation,[],[],ValidInput,_) ,
    append(InitialPosInput,ValidInput,ExtendedPosInput) ,
    prepare_synthesis_of_guard_by_examples(InvolvedVarNames,ExtendedPosInput,InitialNegInput,Positive,Negative,Library,
                                           ParameterAmount,M1,CurrentVars,BehavioralConstraint,LocationVars,Solution) ,
    synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,CurrentVars,CurrentVars,ValidGuards,guard,Positive,Negative,
                            BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,SynthesizedGuard,_) ,
    synthesis_log(solution(guard,SynthesizedGuard)).

start_specific_synthesis_aux(Operation,action,NewPosInput,NegInput,NewPosOutput,_,DistTransitionCache,DistStateCache,
                             CurrentVars,PreCondVars,_,Action) :-
    % synthesize guard if negative input states are given and synthesis has been suspended
    NegInput \= [] ,
    maplist(get_varname_from_id_node,PreCondVars,PreCondVarNames) ,
    DistStateCache = (PosDist,NegDist) ,
    append(NewPosInput,PosDist,GuardPositive) ,
    append(NegInput,NegDist,GuardNegative) ,
    % if neginput empty
    prepare_synthesis_of_guard_by_examples(PreCondVarNames,GuardPositive,GuardNegative,PrePositive,PreNegative,
                                           PreLibrary,PreParameterAmount,PreM1,_,PreBehavioralConstraint,
                                           PreLocationVars,Solution) ,
    synthesis_from_examples(b(truth,pred,[]),(PrePositive,PreNegative),PreLibrary,1,PreCondVars,PreCondVars,
                            any,guard,PrePositive,PreNegative,PreBehavioralConstraint,PreLocationVars,PreM1,
                            PreParameterAmount,Solution,Guards,PreDistStateCache) ,
    ((is_distinguishing(Guards) ; Guards = none)
    ->  Action = Guards
    ;   prepare_synthesis_of_action_by_examples(NewPosInput,NewPosOutput,Library,ParameterAmount,M1,CurrentVars,BehavioralConstraint,
                                                LocationVars,Solution) ,
        synthesis_from_examples(DistTransitionCache,PreDistStateCache,Library,1,CurrentVars,PreCondVars,
                                Guards,action,NewPosInput,NewPosOutput,BehavioralConstraint,LocationVars,
                                M1,ParameterAmount,Solution,Synthesized,(_,NegDistInput)) ,
        post_synthesize_action(Operation,CurrentVars,NewPosInput,NegDistInput,Synthesized,Action)).
start_specific_synthesis_aux(Operation,action,NewPosInput,[],NewPosOutput,_,DistTransitionCache,DistStateCache,
                             CurrentVars,PreCondVars,ValidInvariantsOrGuards,Action) :-
        prepare_synthesis_of_action_by_examples(NewPosInput,NewPosOutput,Library,ParameterAmount,M1,CurrentVars,BehavioralConstraint,
                                                LocationVars,Solution) ,
        synthesis_from_examples(DistTransitionCache,DistStateCache,Library,1,CurrentVars,PreCondVars,
                                ValidInvariantsOrGuards,action,NewPosInput,NewPosOutput,BehavioralConstraint,LocationVars,
                                M1,ParameterAmount,Solution,Synthesized,(_,NegDistInput)) ,
        post_synthesize_action(Operation,CurrentVars,NewPosInput,NegDistInput,Synthesized,Action).

% Auxiliary predicate for synthesis of invariants to check if we can find a counter example.
start_specific_synthesis_aux(invariant,ValidInvariants,VarNames,Positive,TNegative,ViolatingVars,
                             Library,(TCounterIn,TCounterOut),SynthesizedInvariant) :-
    maplist(reduce_input_to_violating_vars(VarNames),[TCounterIn,TCounterOut],[CounterIn,CounterOut]) ,
    % don't create contradiction by using the same example as positive and negative
    findall(Counter,(member(Counter,[CounterIn,CounterOut]) , \+contains_equivalent_state(Counter,Positive,_)),CounterStates) ,
    append(CounterStates,TNegative,Negative) ,
    mark_examples_as_initial(Positive,Negative,InitialPositive,InitialNegative) ,
    get_behavioral_constraint_from_input(Library,InitialPositive,InitialNegative,BehavioralConstraint,LocationVars) ,
    %smt_solvers_interface:smt_solve_predicate(z3,BehavioralConstraint,_,Solution) ,
    solve_predicate(BehavioralConstraint,_,Solution) ,
    get_line_amount(Library,CounterIn,M) , M1 is  M - 1 ,
    length(CounterIn,ParameterAmount) ,
    synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,ViolatingVars,ViolatingVars,ValidInvariants,invariant,InitialPositive,
                            InitialNegative,BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,SynthesizedInvariant,_) ,
    synthesis_log(solution(invariant,SynthesizedInvariant)).
% We can't find any invariant violating transition (counter example), therefore we just remove the violating invariants.
% no invariant given
start_specific_synthesis_aux(invariant,[],_,_,_,_,_,_,b(truth,pred,[])) :-
    synthesis_log(solution(invariant,b(truth,pred,[]))).
% print valid invariants
start_specific_synthesis_aux(invariant,ValidInvariants,_,_,_,_,_,_,Invariant) :-
    conjunct_predicates(ValidInvariants,Invariant) ,
    synthesis_log(solution(invariant,Invariant)).

prepare_synthesis_of_action_by_examples(Input,Output,Library,ParameterAmount,M1,CurrentVars,BehavioralConstraint,
                                        LocationVars,Solution) :-
    Input = [Example|_] ,
    length(Example,ParameterAmount) ,
    library_setup(action,1,Example,Library) ,
    get_current_vars_from_example(Example,CurrentVars) ,
    get_line_amount(Library,Example,M) , M1 is  M - 1 ,
    get_behavioral_constraint(Library,Input,Output,BehavioralConstraint,LocationVars) ,
    %smt_solvers_interface:smt_solve_predicate(z3,BehavioralConstraint,_,Solution).
    solve_predicate(BehavioralConstraint,_,Solution).

prepare_synthesis_of_guard_by_examples(VarNames,TPositive,TNegative,Positive,Negative,Library,ParameterAmount,M1,
                                       CurrentVars,BehavioralConstraint,LocationVars,Solution) :-
    remove_dups(TPositive,TTPositive) , remove_dups(TNegative,TTNegative) ,
    maplist(reduce_input_to_violating_vars(VarNames),TTPositive,TempPositive) ,
    maplist(reduce_input_to_violating_vars(VarNames),TTNegative,TempNegative) ,
    get_distinct_states(guard,TempPositive,TempNegative,Positive,Negative) ,
    Positive = [Example|_] ,
    length(Example,ParameterAmount) ,
    library_setup(guard,1,Example,Library) ,
    get_current_vars_from_example(Example,CurrentVars) ,
    get_behavioral_constraint_from_input(Library,Positive,Negative,BehavioralConstraint,LocationVars) ,
    %smt_solvers_interface:smt_solve_predicate(z3,BehavioralConstraint,_,Solution) ,
    solve_predicate(BehavioralConstraint,_,Solution) ,
    get_line_amount(Library,Example,M) , M1 is  M - 1.

synthesize_guard_if_necessary(_,_,NegativeInputStates,Synthesized,Synthesized) :-
    Synthesized = b(precondition(_,_),subst,_) ; NegativeInputStates = [].
synthesize_guard_if_necessary(VarNames,PositiveInputStates,NegativeInputStates,SynthesizedBody,Synthesized) :-
    prepare_synthesis_of_guard_by_examples(VarNames,PositiveInputStates,NegativeInputStates,
                                           Positive,Negative,Library,ParameterAmount,M1,
                                           CurrentVars,BehavioralConstraint,LocationVars,Solution) ,
    synthesis_from_examples(b(truth,pred,[]),([],[]),Library,1,CurrentVars,CurrentVars,any,guard,Positive,Negative,
                            BehavioralConstraint,LocationVars,M1,ParameterAmount,Solution,SynthesizedGuard,_) ,
    SynthesizedGuard \= none ,
    Synthesized = b(precondition(SynthesizedGuard,SynthesizedBody),subst,[]).

%% SYNTHESIS LOOP
%% Cache a predicate excluding already visited distinguishing transitions for each new run (see
%% get_distinguishing_state_or_trans/8).
%% Only for synthesis of an action (ignored otherwise):
%%    When simultaneously synthesizing a guard to an action we synthesize a new guard each time so that
%%    we also store each session's distinguishing states.
%%    The sixth argument PreCondVars defines the variables that should be considered for the simultaneous
%%    generation of a guard if known. Otherwise, 'PreCondVars = CurrentVars' holds.
% search for further solution
synthesis_from_examples(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                        ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,
                        M1,ParameterAmount,Solution,Synthesized,NewDistStateCache) :-
    synthesis_log(start_synthesis_loop(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                             ValidInvariantsOrGuards,Type,PosInput,NegInput,LocationVars,M1,ParameterAmount,Solution)) ,
    Solution = solution(SolutionBindings) ,
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,UsedLVars1,ProgramAST) ,
    % extend the BehavioralConstraint by a constraint excluding specific location mappings that
    % lead to just swapping equivalent components without changing the semantics
    get_prevent_component_swap_constraint(UsedLVars1,SolutionBindings,LocationVars,ComponentSwapExclusion) ,
    ExtendedBehavioralConstraint = b(conjunct(BehavioralConstraint,ComponentSwapExclusion),pred,[]) ,
    % In UsedLVars we have the used location variables from the synthesized solution, so we are able
    % to exclude those specific bindings when searching for a further solution (otherwise it would
    % be enough to just change some deadcode).
    find_further_solution(ProgramAST,LocationVars,UsedLVars1,Solution,ExtendedBehavioralConstraint,FurtherSolution,NewBehavioralConstraint) ,
    synthesis_log(defective_further_solution_and_fail(Type,NewBehavioralConstraint,FurtherSolution,Library,CurrentVars,
                                            PosInput,NegInput,LocationVars,M1,ParameterAmount)) , ! ,
    add_distinguishing_states_from_cache(Type,DistStateCache,PosInput,NegInput,NewPosInput,NewPosOutput) ,
    % Use the constraint to prevent component swapping only when searching for a further solution and pass the
    % BehavioralConstraint here instead of ExtendedBehavioralConstraint.
    synthesis_from_examples_aux1(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                                 ValidInvariantsOrGuards,Type,NewPosInput,NewPosOutput,BehavioralConstraint,LocationVars,
                                 M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Synthesized,NewDistStateCache).
% expand the library if no solution can be found
synthesis_from_examples(DistTransitionCache,DistStateCache,OldLibrary,LibExpansion,CurrentVars,PreCondVars,
                        ValidInvariantsOrGuards,Type,PosInput,NegInput,_,_,M1,ParameterAmount,Solution,
                        Synthesized,NewDistStateCache) :-
    Solution \= solution(_) ,
    NewLibExpansion is LibExpansion + 1 ,
    PosInput = [Example|_] ,
    library_setup(Type,NewLibExpansion,Example,NewLibrary) , ! ,
    % library has changed
    synthesis_callback(DistTransitionCache,DistStateCache,OldLibrary,NewLibrary,NewLibExpansion,
                       CurrentVars,PreCondVars,ValidInvariantsOrGuards,M1,ParameterAmount,Type,NewLibrary,PosInput,
                       NegInput,Synthesized,NewDistStateCache).

% Synthesis failed if library has not changed, i.e. there is no library expansion step left.
synthesis_callback(_,DistStateCache,OldLibrary,OldLibrary,_,_,_,_,_,_,_,_,_,_,Synthesized,NewDistStateCache) :- ! ,
    Synthesized = none ,
    NewDistStateCache = DistStateCache.
% Otherwise, restart synthesis with the expanded library.
synthesis_callback(DistTransitionCache,DistStateCache,_,NewLibrary,NewLibExpansion,CurrentVars,PreCondVars,
                   ValidInvariantsOrGuards,_,ParameterAmount,Type,NewLibrary,PosInput,NegInput,Synthesized,NewDistStateCache) :-
    synthesis_log(expand_library) ,
    get_behavioral_constraint_aux(Type,NewLibrary,PosInput,NegInput,NewBehavioralConstraint,NewLocationVars) ,
    %smt_solvers_interface:smt_solve_predicate(z3,NewBehavioralConstraint,_,NewSolution) , ! ,
    solve_predicate(NewBehavioralConstraint,_,NewSolution) , ! ,
    % compute the new amount of lines of the synthesized program using the expanded library
    PosInput = [Input|_] ,
    get_line_amount(NewLibrary,Input,M) , NewM1 is  M - 1 ,
    synthesis_from_examples(DistTransitionCache,DistStateCache,NewLibrary,NewLibExpansion,CurrentVars,PreCondVars,
                            ValidInvariantsOrGuards,Type,PosInput,NegInput,NewBehavioralConstraint,NewLocationVars,
                            NewM1,ParameterAmount,NewSolution,Synthesized,NewDistStateCache).

% done
% copy distinguishing state cache
synthesis_from_examples_aux1(_,DistStateCache,Library,_,CurrentVars,_,ValidInvariantsOrGuards,Type,PosInput,NegInput,
                             _,LocationVars,M1,ParameterAmount,Solution,none,_,Synthesized,NewDistStateCache) :-
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_,ProgramAST) ,
    replace_preds_by_subst_if_neccessary(Type,ProgramAST,TSolutionAST) ,
    %prime_or_unprime_variables_in_ast(unprime,TSolutionAST,SolutionAST) ,
    synthesis_util:expand_state_cache_if_guard_and_solution(Type,Solution,DistStateCache,PosInput,NegInput,NewDistStateCache) ,
    conjunct_previous_invariant_or_guard_if_necessary(Type,ValidInvariantsOrGuards,TSolutionAST,Synthesized).
% try to find a distinguishing input
synthesis_from_examples_aux1(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                             ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,
                             M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Synthesized,NewDistStateCache) :-
    FurtherSolution \= none ,
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_,ProgramAST) ,
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,FurtherSolution,_,ProgramAST2) ,
    synthesis_log(solution_tuple(ProgramAST,ProgramAST2)) ,
    get_distinguishing_state_or_trans(DistTransitionCache,CurrentVars,Type,ValidInvariantsOrGuards,
                                      ProgramAST,ProgramAST2,NewDistTransitionCache,Distinguishing) ,
    synthesis_from_examples_aux2(NewDistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                                 ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,
                                 M1,ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Distinguishing,
                                 Synthesized,NewDistStateCache).

% two non equivalent programs
synthesis_from_examples_aux2(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                             ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,ParameterAmount,
                             Solution,FurtherSolution,NewBehavioralConstraint,solution(Dist),Synthesized,NewDistStateCache) :-
    synthesis_log(two_non_equivalent_solutions(Type,Solution,FurtherSolution)) ,
    synthesis_from_examples_aux3(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                                 ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,
                                 ParameterAmount,Solution,FurtherSolution,NewBehavioralConstraint,Dist,
                                 Synthesized,NewDistStateCache).
% or we choose the one with a lower term size since we know that both ASTs are equivalent
synthesis_from_examples_aux2(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                             ValidInvariantsOrGuards,Type,PosInput,NegInput,BehavioralConstraint,LocationVars,M1,
                             ParameterAmount,Solution,FurtherSolution,FurtherBehavioralConstraint,Distinguishing,
                             Synthesized,NewDistStateCache) :-
    Distinguishing \= solution(_) ,
    % No distinguishing input, no possible conclusion about examples.
    % Therefore we pass the "new" behavioral constraint and start synthesis again, we keep the
    % solution with the smaller term size.
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,Solution,_,ProgramAST) ,
    location_vars_to_program(Library,Type,CurrentVars,LocationVars,M1,ParameterAmount,FurtherSolution,_,ProgramAST2) ,
    synthesis_log(distinguishing_error(Distinguishing,ProgramAST,ProgramAST2)) ,
    term_size(ProgramAST,TS1) , term_size(ProgramAST2,TS2) ,
    (TS1 < TS2
    ->  get_new_behavioral_constraint(ProgramAST2,LocationVars,FurtherSolution,BehavioralConstraint,NewBehavioralConstraint) ,
        Sol = Solution
    ;   NewBehavioralConstraint = FurtherBehavioralConstraint ,
        Sol = FurtherSolution) ,
    synthesis_from_examples(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                            ValidInvariantsOrGuards,Type,PosInput,NegInput,NewBehavioralConstraint,LocationVars,M1,
                            ParameterAmount,Sol,Synthesized,NewDistStateCache).

% distinguishing transition (actions)
synthesis_from_examples_aux3(DistTransitionCache,DistStateCache,Library,LibExpansion,CurrentVars,PreCondVars,
                             ValidInvariantsOrGuards,action,Input,Output,_,_,M1,ParameterAmount,_,_,_,
                             Distinguishing,Synthesized,NewDistStateCache) :-
    get_input_output_nodes_from_bindings(Distinguishing,DistIn,DistOut) ,
    DistStateCache = (_,NegativeInputStates) ,
    (member(DistIn,NegativeInputStates)
    ->  % synthesize guard for the action, the synthesis of the action is suspended for this time
        maplist(get_varname_from_id_node,PreCondVars,PreCondVarNames) ,
        DistStateCache = (PosDist,NegDist) ,
        append(PosDist,Input,PreInput) , append(NegDist,[DistIn],PreOutput) ,
        prepare_synthesis_of_guard_by_examples(PreCondVarNames,PreInput,PreOutput,PrePositive,PreNegative,
                                               PreLibrary,PreParameterAmount,PreM1,_,PreBehavioralConstraint,
                                               PreLocationVars,Solution) ,
        synthesis_from_examples(b(truth,pred,[]),(PrePositive,PreNegative),PreLibrary,1,PreCondVars,PreCondVars,
                                any,guard,PrePositive,PreNegative,PreBehavioralConstraint,PreLocationVars,PreM1,
                                PreParameterAmount,Solution,Guards,PreDistStateCache) ,
        (is_distinguishing(Guards)
        ->  Synthesized = Guards , NewDistStateCache = PreDistStateCache ,
            % the synthesis context of the action is still asserted
            assert_synthesis_context(DistTransitionCache,NewDistStateCache,CurrentVars,PreCondVars,ValidInvariantsOrGuards,guard,
                                     PrePositive,PreNegative)
        ;   NewInput = Input , NewOutput = Output , NewLibrary = Library , NewLibExpansion = LibExpansion) ,
            get_behavioral_constraint(NewLibrary,NewInput,NewOutput,DistBehavioralConstraint,DistLocationVars) ,
            %smt_solvers_interface:smt_solve_predicate(z3,DistBehavioralConstraint,_,DistSolution) ,
            solve_predicate(DistBehavioralConstraint,_,DistSolution) ,
            synthesis_from_examples(DistTransitionCache,PreDistStateCache,Library,NewLibExpansion,CurrentVars,PreCondVars,
                                    Guards,action,NewInput,NewOutput,DistBehavioralConstraint,
                                    DistLocationVars,M1,ParameterAmount,DistSolution,Synthesized,NewDistStateCache)
    ;   Synthesized = transition(DistIn,DistOut) , NewDistStateCache = DistStateCache ,
        assert_synthesis_context(DistTransitionCache,NewDistStateCache,CurrentVars,PreCondVars,ValidInvariantsOrGuards,action,
                                 Input,Output)).
% distinguishing state (guards, invariants)
synthesis_from_examples_aux3(DistTransitionCache,DistStateCache,_,_,CurrentVars,PreCondVars,ValidInvariantsOrGuards,Type,
                             Input,Output,_,_,_,_,_,_,_,Distinguishing,state(DistIn),DistStateCache) :-
    Type \= action ,
    %adjust_distinguishing_state_binding(CurrentVars,Distinguishing,CompleteDist) ,
    get_input_nodes_from_bindings(Distinguishing,DistIn) ,
    assert_synthesis_context(DistTransitionCache,DistStateCache,CurrentVars,PreCondVars,ValidInvariantsOrGuards,Type,
                             Input,Output).
%%

assert_synthesis_context(DistTransitionCache,NewDistStateCache,CurrentVars,PreCondVars,ValidInvariantsOrGuards,Type,
                         Input,Output) :-
    retractall(synthesis_context(_,_,_,_,_,Type,_,_)) ,
    assert(synthesis_context(DistTransitionCache,NewDistStateCache,CurrentVars,PreCondVars,ValidInvariantsOrGuards,Type,
                             Input,Output)).

is_distinguishing(transition(_,_)).
is_distinguishing(state(_)).

find_further_solution(ProgramAST,LocationVars,_,solution(SolutionBindings),BehavioralConstraint,FurtherSolution,NewBehavioralConstraint) :-
    get_new_behavioral_constraint(ProgramAST,LocationVars,solution(SolutionBindings),BehavioralConstraint,NewBehavioralConstraint) ,
    %smt_solvers_interface:smt_solve_predicate(z3,NewBehavioralConstraint,_,TempSolution) ,
    solve_predicate(NewBehavioralConstraint,_,TempSolution) ,
    find_further_solution_aux(TempSolution,FurtherSolution).
find_further_solution_aux(solution(S),solution(S)) :- !.
find_further_solution_aux(_,none).

% Set an enumeration as postfix of the operation name to prevent collisions in case we synthesized several operations at once
% representing if-statements.
%finalize_synthesized_actions(OperationName,Actions,FinalizedActions) :-
%    finalize_synthesized_actions_aux(0,OperationName,Actions,[],FinalizedActions).
%finalize_synthesized_actions_aux(_,_,[],Acc,Acc).
%finalize_synthesized_actions_aux(Index,OperationName,[Action|T],Acc,FinalizedActions) :-
%    atom_codes(OperationName,OperationCodes) ,
%    number_codes(Index,NumberCodes) ,
%    append(OperationCodes,NumberCodes,NewOperationCodes) ,
%    atom_codes(NewOperationName,NewOperationCodes) ,
%    finalize_synthesized_action(NewOperationName,Action,Operation) ,
%    NewIndex is Index + 1 ,
%    finalize_synthesized_actions_aux(NewIndex,OperationName,T,[Operation|Acc],FinalizedActions).

finalize_synthesized_action(_,none,none).
finalize_synthesized_action(OperationName,b(precondition(PreCondition,Substitutions),subst,[]),
                            b(operation(OpNameId,_,_,b(precondition(PreCondition,UnprimedSubstitutions),subst,[])),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    prime_or_unprime_variables_in_ast(unprime,Substitutions,UnprimedSubstitutions).
finalize_synthesized_action(OperationName,Action,b(operation(OpNameId,_,_,UnprimedAction),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    prime_or_unprime_variables_in_ast(unprime,Action,UnprimedAction).
