:- module(constraints,[get_behavioral_constraint_from_input/5,
                       get_behavioral_constraint/5,
                       get_behavioral_constraint_aux/6,
                       get_new_behavioral_constraint/5,
                       get_prevent_component_swap_constraint/4,
                       get_line_amount/3]).

:- use_module(extension('plspec/plspec/plspec')).
:- use_module(extension('plspec/plspec/plspec_checker')).

:- use_module(library(lists)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(tools),[flatten/2]).
:- use_module(synthesis(synthesis_util)).
:- use_module(synthesis(symmetry_reduction)).

% Exclude a solution and add this node to the behavioral constraint. We exclude only the solution given
% by the program ast without involving generated deadcode.
get_new_behavioral_constraint(ProgramAST,LocationVars,solution(SolutionBindings),BehavioralConstraint,NewBehavioralConstraint) :-
    % get the used operators output location variables to exclude from the further solution
    get_used_location_vars_and_op_names(ProgramAST,LocationVars,[],[],_,UsedOperatorNames) ,
    exclude_solution_for_further_search(UsedOperatorNames,LocationVars,SolutionBindings,Exclusion) ,
    conjunct_predicates([BehavioralConstraint,Exclusion],NewBehavioralConstraint).

get_behavioral_constraint_aux(action,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :-
    get_behavioral_constraint(Library,PosInput,PosOutput,BehavioralConstraint,LocationVars).
get_behavioral_constraint_aux(_,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :-
    get_behavioral_constraint_from_input(Library,PosInput,PosOutput,BehavioralConstraint,LocationVars).

% create outputs and get behavioral constraint
get_behavioral_constraint_from_input(Library,PosInput,NegInput,BehavioralConstraint,LocationVars) :-
    % get output values
    length(PosInput,PosLength) , length(NegInput,NegLength) ,
    % create empty lists
    length(TempPosOutput,PosLength) , length(TempNegOutput,NegLength) ,
    % for our task to repair invariant violations output is either "true" or "false"
    findall(In,(member(In,TempPosOutput) , In = ['TRUE']),PosOutput) ,
    findall(In,(member(In,TempNegOutput) , In = ['FALSE']),NegOutput) ,
    append(PosInput,NegInput,Input) ,
    append(PosOutput,NegOutput,Output) ,
    get_behavioral_constraint(Library,Input,Output,BehavioralConstraint,LocationVars).

:- dynamic used_library/2.
:- volatile used_library/2.

% Get the functional constraint for each Input-Output Example
% It is sufficient to create one well-definedness constraint surrounding the behavioral constraint
% but the library constraint needs to be defined for every examples functional constraint.
get_behavioral_constraint(Library,InputExamples,OutputExamples,BehavioralConstraint,LocationVars) :-
    % lists of currently used components, necessary if a library is used several times
    assert(used_library([],[])) ,
    % create library constraint separated from the funtional constraint
    % to make sure to just obtain one set of location variables
    get_library_constraint(Library,LibConstraint,InputVars,OutputVars) ,
    % set the temporary location variables with prefix "l"
    maplist(get_location_var_from_ast_node,InputVars,LocationVarsIn) ,
    maplist(get_location_var_from_ast_node,OutputVars,LocationVarsOut) ,
    % tuples of raw component name and internal component name like (add,add1)
    used_library(PredicateTuples,ExpressionTuples) ,
    get_symmetry_reduction_constraint(ExpressionTuples,LocationVarsIn,LocationVarsOut,SymmetryReductionExpr) ,
    get_symmetry_reduction_constraint(PredicateTuples,LocationVarsIn,LocationVarsOut,SymmetryReductionPred) ,
    conjunct_predicates([LibConstraint,SymmetryReductionPred,SymmetryReductionExpr],ExtendedLibConstraint) ,

    % get some information needed for well-definedness constraint
    InputExamples = [InputValues|_] ,
    get_line_amount(Library,InputValues,M) ,
    length(InputValues,VarAmount) ,
    % create well-definedness constraint separated from function constraints
    get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) ,
    get_range_of_integer_nodes(InputExamples,InMin,InMax) ,
    map_get_function_constraint(Library,InMin,InMax,ExtendedLibConstraint,InputVars,OutputVars,
                                InputExamples,OutputExamples,FunctionConstraints,ProgramInLocationVars) ,
    conjunct_predicates([WellDefinednessConstraint|FunctionConstraints],BehavioralConstraint) ,
    append([LocationVarsIn,LocationVarsOut,ProgramInLocationVars],LocationVars) ,
    retract(used_library(_,_)).

% Get the range of integer nodes from the initial examples to restrict constant domains.
get_range_of_integer_nodes(InputExamples,InMin,InMax) :-
    flatten(InputExamples,InputSet) ,
    findall(Num,(member(b(integer(Num),integer,Info),InputSet) ,
                 member(synthesis(initial_example),Info)),IntegerNodes) ,
    IntegerNodes \= [] , ! ,
    min_member(TInMin,IntegerNodes) , max_member(TInMax,IntegerNodes) ,
    InMin is TInMin - 10 , InMax is TInMax + 10.

map_get_function_constraint(_,_,_,_,_,_,[],[],[],[]).
map_get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars,[Input|TIn],
                            [Output|TOut],[FunctionConstraint|NTFunction],LocationVars) :-
    get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars,Input,Output,
                            FunctionConstraint,NewLocationVars) ,
    map_get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars,
                            TIn,TOut,NTFunction,NTLVars) ,
    append(NewLocationVars,NTLVars,LocationVars).

get_function_constraint(Library,InMin,InMax,LibConstraint,InputVars,OutputVars,InputValues,
                        OutputValues,FunctionConstraint,ProgramInLocationVars) :-
    get_line_amount(Library,InputValues,M) ,
    % enumerate input/output values
    transition_values_to_ast(InputValues,input,InputValueNodes) ,
    transition_values_to_ast(OutputValues,output,OutputValueNodes) ,
    get_connectivity_constraint(InputValueNodes,OutputValueNodes,InputVars,OutputVars,ConnectivityConstraint) ,
    set_input_location_vars(InputValueNodes,InputLocationVars) ,
    set_output_location_vars(OutputValueNodes,M,OutputLocationVars) ,
    find_typed_identifier_uses_with_info(InputLocationVars,ProgramInLocationVars) ,
    conjunct_predicates([OutputLocationVars,InputLocationVars,LibConstraint,ConnectivityConstraint],
                         TempFunctionConstraint) ,
    % get all used identifier as ast nodes
    find_typed_identifier_uses_with_info(TempFunctionConstraint,TListOfIDNodes) ,
    remove_dup_id_nodes(TListOfIDNodes,ListOfIDNodes) ,
    get_temporary_ids(ListOfIDNodes,TemporaryIDs) ,
    % and their corresponding identifier names
    maplist(get_texpr_id,ListOfIDNodes,UsedIDs) ,
    % create upper and lower bound for input values (kodkod warning) and constants
    create_texpr(integer(InMin),integer,[],LowerBound) ,
    create_texpr(integer(InMax),integer,[],UpperBound) ,
    findall(InputRestriction,(member(In,InputVars) , In = b(_,integer,_) , InputRestriction = b(member(
                In,b(interval(LowerBound,UpperBound),set(integer),[])),pred,[])),TempInputRestrictions) ,
    findall(ConstantRestriction,(member(Out,OutputVars) , is_integer_constant(Out) ,
                                 ConstantRestriction = b(member(Out,b(interval(LowerBound,UpperBound),set(integer),[])),
                                    pred,[])),TempConstantRestrictions) ,
    conjunct_predicates(TempInputRestrictions,InputRestrictions) ,
    conjunct_predicates(TempConstantRestrictions,ConstantRestrictions) ,
    exclude_trivial_assignments(InputVars,[],Exclusion) ,
    % TODO: don't add this to every example but once for all variables
    conjunct_predicates([TempFunctionConstraint,InputRestrictions,ConstantRestrictions],ExistsBody) ,
    FunctionConstraint = b(exists(TemporaryIDs,b(conjunct(Exclusion,ExistsBody),pred,[])),pred,[used_ids(UsedIDs)]).

is_integer_constant(b(_,integer,Info)) :-
    member(synthesis(Constant,output),Info) ,
    delete_numbers_from_atom(Constant,Operator) , ! ,
    Operator = constant.

% all temporary library variables like i0,i1,o1 in exists node except 'constant'
get_temporary_ids([],[]).
get_temporary_ids([ID|T],[ID|NT]) :-
    get_texpr_id(ID,Name) ,
    atom_codes(Name,[Char|_]) ,
    Char \= 108 ,
    \+atom_concat(constant,_,Name) , ! ,
    get_temporary_ids(T,NT).
get_temporary_ids([_|T],NT) :-
    get_temporary_ids(T,NT).

% we can have duplicate nodes because of undefined types of sets
remove_dup_id_nodes(Nodes,Distinct) :-
    remove_dup_id_nodes(Nodes,[],Distinct).
remove_dup_id_nodes([],Acc,Acc).
remove_dup_id_nodes([Node|T],Acc,Distinct) :-
    Node = b(identifier(Name),_,_) ,
    \+member(b(identifier(Name),_,_),Acc) , ! ,
    remove_dup_id_nodes(T,[Node|Acc],Distinct).
remove_dup_id_nodes([_|T],Acc,Distinct) :-
    remove_dup_id_nodes(T,Acc,Distinct).

% define connections of temporary inputs and outputs
get_connectivity_constraint(InputExampleNodes,OutputExampleNodes,InputVars,OutputVars,ConnectivityConstraint) :-
    append([InputVars,OutputVars,InputExampleNodes,OutputExampleNodes],SetOfVars) ,
    findall((X,Y),
        (member(X,SetOfVars) , member(Y,SetOfVars) ,
         X \= Y , nth0(I1,SetOfVars,X) , nth0(I2,SetOfVars,Y) , I1 < I2),ListOfVarPairs) ,
    maplist(get_single_connectivity_constraint,ListOfVarPairs,ConstraintList) ,
    conjunct_predicates(ConstraintList,ConnectivityConstraint).

get_single_connectivity_constraint((Var1,Var2),ConnectivityConstraint) :-
    get_location_var_from_ast_node(Var1,LocationVar1) ,
    get_location_var_from_ast_node(Var2,LocationVar2) ,
    % only connect two variables if they have the same type
    (valid_connection(Var1,Var2)
    ->  LHS = b(equal(LocationVar1,LocationVar2),pred,[]) ,
        RHS = b(equal(Var1,Var2),pred,[]) ,
        create_implication(LHS,RHS,ConnectivityConstraint)
    % prevent equality of such incompatible location variables
    ;   ConnectivityConstraint = b(not_equal(LocationVar1,LocationVar2),pred,[])).

valid_connection(Var1,Var2) :-
    get_texpr_info(Var1,Info1) ,
    get_texpr_info(Var2,Info2) ,
    % same type
    valid_type_connection(Var1,Var2) ,
    % no predicates within boolean operators (boolean_true /= pred_true), i.e. avoid (c>0) = d
    valid_connection_aux(Info1,Info2).

% constant_set only right input of member, subset, subset_strict
valid_connection_aux(Info1,Info2) :-
    member(synthesis(ConstantSet,output),Info1) ,
    delete_numbers_from_atom(ConstantSet,constant_set) , ! ,
    member(synthesis(Member,right_input),Info2) ,
    delete_numbers_from_atom(Member,Op) ,
    (Op = member ; Op = subset ; Op = subset_strict).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(ConstantSet,output),Info1) ,
    delete_numbers_from_atom(ConstantSet,constant_set) , ! ,
    member(synthesis(Member,right_input),Info2) ,
    delete_numbers_from_atom(Member,Op) ,
    (Op = member ; Op = subset ; Op = subset_strict).
% Some nodes are marked with 'noPred' like we do not want to map a predicate to an equality node's input
% which is not possible in B/Event-B.
valid_connection_aux(Info1,Info2) :-
    member(synthesis(noPred),Info1) ,
    (member(synthesis(input(_)),Info2) ;
    (member(synthesis(Comp2,_),Info2) , atom_concat(constant,_,Comp2))).
valid_connection_aux(Info1,Info2) :-
    member(synthesis(noPred),Info2) ,
    (member(synthesis(input(_)),Info1) ;
    (member(synthesis(Comp1,_),Info1) , atom_concat(constant,_,Comp1))).
valid_connection_aux(Info1,Info2) :-
    member(synthesis(noInput),Info1) ,
    \+member(synthesis(input(_)),Info2).
valid_connection_aux(Info1,Info2) :-
    member(synthesis(noInput),Info2) ,
    \+member(synthesis(input(_)),Info1).
valid_connection_aux(Info1,Info2) :-
    (\+member(synthesis(noPred),Info1) , \+member(synthesis(noPred),Info2)).

valid_type_connection(b(_,Type,_),b(_,Type,_)).
valid_type_connection(b(_,Type1,_),b(_,Type2,_)) :-
    (Type1 = seq(InnerType) ; Type1 = set(couple(integer,InnerType))) ,
    (Type2 = seq(InnerType) ; Type2 = set(couple(integer,InnerType))).

get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) :-
    get_consistency_constraint(LocationVarsOut,ConsistencyConstraint) ,
    get_acyclic_constraint(LocationVarsIn,LocationVarsOut,AcyclicConstraint) ,
    NewM is M - 1 , % minus one to use an interval
    get_location_constraint(LocationVarsIn,0,NewM,InputRestriction) ,
    get_location_constraint(LocationVarsOut,VarAmount,NewM,OutputRestriction) ,
    conjunct_predicates([InputRestriction,OutputRestriction,ConsistencyConstraint,AcyclicConstraint],
        WellDefinednessConstraint).

% create restriction nodes (membership to an interval) to the domain
% of each location variable
get_location_constraint(LocationVars,A,B,RestrictionNode) :-
    maplist(get_single_location_constraint(A,B),LocationVars,RestrictionList) ,
    conjunct_predicates(RestrictionList,RestrictionNode).
get_single_location_constraint(A,B,LocationVar,RestrictionNode) :-
    RestrictionNode = b(member(LocationVar,
        b(interval(b(integer(A),integer,[]),b(integer(B),integer,[])),set(integer),[])),pred,[]).

% consistency constraint guarantees that there is only one output in each line
get_consistency_constraint(LocationVarsOut,Constraint) :-
    % find all pairs of location variables (not redundant)
    % and create "not_equal"-Node for each pair
    findall(InequalityNode,
        (member(L1,LocationVarsOut) , member(L2,LocationVarsOut) ,
         L1 \= L2 , nth0(I1,LocationVarsOut,L1) , nth0(I2,LocationVarsOut,L2) ,
         I1 < I2 , create_texpr(not_equal(L1,L2),pred,[],InequalityNode)),InequalityList) ,
    conjunct_predicates(InequalityList,Constraint).

% acyclic constraint guarantees that every operators input is defined before
% its corresponding output
get_acyclic_constraint(LocationVarsIn,LocationVarsOut,Constraint) :-
    get_acyclic_constraint_list(LocationVarsIn,LocationVarsOut,ConstraintList) ,
    conjunct_predicates(ConstraintList,Constraint).
get_acyclic_constraint_list(_,[],[]).
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :-
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) ,
    delete_numbers_from_atom(ComponentID,Component) ,
    % constant has no input
    \+member(Component,[constant,negation]) ,
    member(In1,LocationVarsIn) ,
    In1 = b(_,_,Info1) ,
    member(synthesis(ComponentID,left_input),Info1) , ! ,
    member(In2,LocationVarsIn) ,
    In2 = b(_,_,Info2) ,
    member(synthesis(ComponentID,right_input),Info2) ,
    create_texpr(less(In1,Out),pred,[],Node1) ,
    create_texpr(less(In2,Out),pred,[],Node2) ,
    conjunct_predicates([Node1,Node2],Constraint) ,
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :-
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) ,
    delete_numbers_from_atom(ComponentID,Component) ,
    % negation just one input
    Component = negation , ! ,
    member(In,LocationVarsIn) ,
    In = b(_,_,Info) ,
    member(synthesis(ComponentID,input),Info) ,
    create_texpr(less(In,Out),pred,[],Constraint) ,
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
% skip constants
get_acyclic_constraint_list(LocationVarsIn,[_|ROut],ConstraintList) :-
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).

% construct the library constraint by encoding the given library components as logical formulas
get_library_constraint((Operators,Arithmetic),Constraint,InputVars,OutputVars) :-
    maplist(get_single_operator_constraint,Operators,ConstraintOpList) ,
    maplist(get_single_arithmetic_constraint,Arithmetic,ConstraintArithList) ,
    append(ConstraintOpList,ConstraintArithList,ConstraintList) ,
    conjunct_predicates(ConstraintList,TConstraint) ,
    find_typed_identifier_uses_with_info(TConstraint,ListOfIDNodes) ,
    % split used identifier by input and output
    findall(I,(member(I,ListOfIDNodes) , get_texpr_id(I,IName) ,
                atom_concat(i,_Suffix1,IName)),InputVars) ,
    findall(O,(member(O,ListOfIDNodes) , get_texpr_id(O,OName) ,
                atom_concat(o,_Suffix2,OName)),OutputVars) ,
    get_wd_constraints(ListOfIDNodes,[],WDConstraintList) ,
    conjunct_predicates([TConstraint|WDConstraintList],Constraint).

% Create well-definedness constraints to prevent division by zero, max/min of empty sets or
% front/tail/last/first on empty sequences.
get_wd_constraints([],Acc,Acc).
get_wd_constraints([b(Node,Type,Info)|T],Acc,WDConstraintList) :-
    \+member(synthesis(_,output),Info) ,
    member(synthesis(OpName,_),Info) ,
    get_wd_constraints_aux(OpName,b(Node,Type,Info),AccWDConstraint) ,
    get_wd_constraints(T,[AccWDConstraint|Acc],WDConstraintList).
get_wd_constraints([_|T],Acc,WDConstraintList) :-
    get_wd_constraints(T,Acc,WDConstraintList).

% TODO: maybe explicitly exclude same nominator and denominator for div/2 since it is just the constant value 1
get_wd_constraints_aux(OpName,Node,b(not_equal(Node,b(integer(0),integer,[])),pred,[])) :-
    % to only check division by zero we would need to check that the variable is the right input, thus the denominator,
    % but we also want to prevent 0/c since this is just the constant value 0
    delete_numbers_from_atom(OpName,div).
get_wd_constraints_aux(OpName,Node,b(greater(b(card(Node),integer,[]),b(integer(0),integer,[])),pred,[])) :-
    delete_numbers_from_atom(OpName,RawOpName) ,
    member(RawOpName,[max,min,union,concat,set_subtraction,front,tail,first,last]).

% Library components represented as logical formulas.
% (BOOLEAN,BOOLEAN) :: BOOLEAN
get_single_operator_constraint(Operator,Constraint) :-
    member(Operator,[conjunct,disjunct,implication,equivalence]) , ! ,
    unique_typed_id("i",boolean,TempIn1) ,
    unique_typed_id("i",boolean,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(pred,Operator,OperatorName) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn1,[synthesis(OperatorName,left_input)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(OperatorName,right_input)],In2) ,
    add_texpr_infos(TempOut,[synthesis(OperatorName,output)],Out) ,
    get_single_operator_constraint_aux2(Operator,In1,In2,Out,Truth,Falsity) ,
    disjunct_predicates([Truth,Falsity],Constraint).
% (BOOLEAN :: BOOLEAN)
get_single_operator_constraint(negation,Constraint) :- ! ,
    unique_typed_id("i",boolean,TempIn) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(pred,negation,Negation) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn,[synthesis(Negation,input),synthesis(noInput)],In) ,
    add_texpr_infos(TempOut,[synthesis(Negation,output)],Out) ,
    Truth = b(conjunct(b(equal(In,b(boolean_false,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(equal(In,b(boolean_true,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]) ,
    disjunct_predicates([Truth,Falsity],Constraint).
% constant :: (INTEGER,BOOLEAN,SET)
get_single_operator_constraint(constant-Type,Constraint) :-
    get_single_constant_or_set_operator_id(constant-Type,TempOut,Constant) , ! ,
    get_texpr_id(Constant,ConstantName) ,
    % set information of identifier node
    add_texpr_infos(TempOut,[synthesis(ConstantName,output)],Out) ,
    Constraint = b(equal(Constant,Out),pred,[]).
% bool_set, string_set, integer_set
get_single_operator_constraint(Operator,Constraint) :-
    get_single_constant_or_set_operator_id(Operator,TempOut,b(_,Type,_)) , ! ,
    unique_typed_id("constant_set",Type,b(identifier(ConstantName),Type,Info)) ,
    add_texpr_infos(TempOut,[synthesis(ConstantName,output),synthesis(type,Operator)],Out) ,
    % define as subset to avoid infinite sets (virtual timeout)
    Constraint = b(equal(Out,b(identifier(ConstantName),Type,Info)),pred,[]).
% set-expressions
% SET :: INTEGER
get_single_operator_constraint(Operator-SetType,Constraint) :-
    member(Operator,[card,min,max]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(expr,Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).
% SET :: SET(SET)
get_single_operator_constraint(Operator-SetType,Constraint) :-
    member(Operator,[pow_subset,pow_subset1,fin_subset,fin1_subset]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",set(SetType),TempOut) ,
    set_component_id(expr,Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,set(SetType),[]),Out),pred,[]).
% SET :: SET
get_single_operator_constraint(Operator-SetType,Constraint) :-
    member(Operator,[closure,reflexive_closure]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",SetType,TempOut) ,
    set_component_id(expr,Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
% (SET,SET) :: SET
get_single_operator_constraint(Operator-SetType,Constraint) :-
    member(Operator,[union,intersection,concat,insert_front,insert_tail,restrict_front,
                     restrict_tail,overwrite,set_subtraction,cartesian_product,image,
                     domain_restriction,domain_subtraction,range_restriction,range_subtraction]) ,
    get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) ,
    set_component_id(expr,Operator,NOperator) ,
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input)],In2) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Node =.. [Operator,In1,In2] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_operator_constraint(Operator-SetType,Constraint) :-
    member(Operator,[first,last,front,tail,general_union,general_intersection,reverse,range,domain]) ,
    get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) ,
    set_component_id(expr,Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_operator_constraint(size-SetType,Constraint) :-
    get_fn_types(SetType,integer,Type) ,
    unique_typed_id("i",seq(Type),TempIn) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(expr,size,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    Constraint = b(equal(b(size(In),integer,[]),Out),pred,[]).

% We use boolean_true/boolean_false within the synthesis constraint and no pred_true/pred_false
% (INTEGER,INTEGER) :: BOOLEAN || (BOOLEAN,BOOLEAN) :: BOOLEAN || (SET,SET) :: BOOLEAN
get_single_operator_constraint(Operator,Constraint) :-
    get_single_operator_constraint_aux2(Operator,TempIn1,TempIn2,TempOut,Op,FurtherInfo) ,
    set_component_id(pred,Op,NOperator) ,
    % set information of identifier nodes
    % further info 'noPred' to avoid predicates within boolean equality node later on
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input)|FurtherInfo],In1) ,
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input)|FurtherInfo],In2) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    TempTrueNode =.. [Op,In1,In2] ,
    TrueNode = b(TempTrueNode,pred,[]) ,
    BoolNodeTrue = b(equal(Out,b(boolean_true,boolean,[])),pred,[]) ,
    inverse_operator(Op,InverseOp) ,
    TempFalseNode =.. [InverseOp,In1,In2] ,
    FalseNode = b(TempFalseNode,pred,[]) ,
    BoolNodeFalse = b(equal(Out,b(boolean_false,boolean,[])),pred,[]) ,
    create_equivalence(TrueNode,BoolNodeTrue,Truth) ,
    create_equivalence(FalseNode,BoolNodeFalse,Falsity) ,
    conjunct_predicates([Truth,Falsity],Constraint).

% operators
get_single_operator_constraint_aux2(Equal-Type,TempIn1,TempIn2,TempOut,Equal,Info) :-
    member(Equal,[equal,not_equal]) ,
    unique_typed_id("i",Type,TempIn1) ,
    unique_typed_id("i",Type,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    (Type = boolean
    ->  Info = [synthesis(noPred)]
    ;   Info = []).
get_single_operator_constraint_aux2(implication,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_single_operator_constraint_aux2(equivalence,In1,In2,Out,Truth,Falsity) :-
    get_single_operator_constraint_aux2(implication,In1,In2,Out,TruthImpl1,FalsityImpl1) ,
    get_single_operator_constraint_aux2(implication,In2,In1,Out,TruthImpl2,FalsityImpl2) ,
    Truth = b(conjunct(TruthImpl1,TruthImpl2),pred,[]) ,
    Falsity = b(disjunct(FalsityImpl1,FalsityImpl2),pred,[]).
get_single_operator_constraint_aux2(conjunct,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_single_operator_constraint_aux2(disjunct,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_single_operator_constraint_aux2(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[subset,subset_strict,not_subset,not_subset_strict]) , ! ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",boolean,TempOut).
get_single_operator_constraint_aux2(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[member,not_member]) , ! ,
    (SetType = set(Type) ; SetType = seq(Type)) ,
    unique_typed_id("i",Type,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",boolean,TempOut).
get_single_operator_constraint_aux2(Operator,TempIn1,TempIn2,TempOut,Operator,[]) :-
    unique_typed_id("i",integer,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",boolean,TempOut).
% set- and seq-expressions
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :-
    member(Operator,[general_union,general_intersection]) ,
    unique_typed_id("i",SetType,TempIn) ,
    SetType = set(InnerType) ,
    unique_typed_id("o",InnerType,TempOut).
get_single_operator_constraint_aux3(image-SetType,TempIn1,TempIn2,TempOut,image,[]) :-
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn1) ,
    unique_typed_id("i",set(A),TempIn2) ,
    unique_typed_id("o",set(B),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[domain_restriction,domain_subtraction]) ,
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(A),TempIn1) ,
    unique_typed_id("i",set(couple(A,B)),TempIn2) ,
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[range_restriction,range_subtraction]) ,
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn1) ,
    unique_typed_id("i",set(B),TempIn2) ,
    unique_typed_id("o",set(couple(A,B)),TempOut).
% TODO: function inverse ~
get_single_operator_constraint_aux3(reverse-SetType,TempIn,_,TempOut,reverse,[]) :-
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn) ,
    unique_typed_id("o",set(couple(B,A)),TempOut).
get_single_operator_constraint_aux3(domain-SetType,TempIn,_,TempOut,domain,[]) :-
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn) ,
    unique_typed_id("o",set(A),TempOut).
get_single_operator_constraint_aux3(range-SetType,TempIn,_,TempOut,range,[]) :-
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn) ,
    unique_typed_id("o",set(B),TempOut).
% ..
get_single_operator_constraint_aux3(cartesian_product-_,TempIn1,TempIn2,TempOut,cartesian_product,[]) :-
    unique_typed_id("i",set(A),TempIn1) ,
    unique_typed_id("i",set(B),TempIn2) ,
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[union,intersection,set_subtraction]) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(overwrite-SetType,TempIn1,TempIn2,TempOut,overwrite,[]) :-
    get_fn_types(SetType,integer,Type) ,
    unique_typed_id("i",seq(Type),TempIn1) ,
    unique_typed_id("i",set(couple(integer,Type)),TempIn2) ,
    unique_typed_id("o",seq(Type),TempOut).
get_single_operator_constraint_aux3(concat-SetType,TempIn1,TempIn2,TempOut,concat,[]) :-
    get_fn_types(SetType,integer,Type) ,
    unique_typed_id("i",seq(Type),TempIn1) ,
    unique_typed_id("i",seq(Type),TempIn2) ,
    unique_typed_id("o",seq(Type),TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :-
    member(Operator,[first,last]) ,
    unique_typed_id("i",SetType,TempIn) ,
    SetType = set(InnerType) ,
    unique_typed_id("o",InnerType,TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator,[]) :-
    member(Operator,[front,tail]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(insert_front-SetType,TempIn1,TempIn2,TempOut,insert_front,[]) :-
    SetType = set(couple(integer,InnerType)) ,
    unique_typed_id("i",InnerType,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(insert_tail-SetType,TempIn1,TempIn2,TempOut,insert_tail,[]) :-
    SetType = set(couple(integer,InnerType)) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",InnerType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_operator_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator,[]) :-
    member(Operator,[restrict_front,restrict_tail]) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).

% (INTEGER,INTEGER) :: INTEGER
get_single_arithmetic_constraint(ArithmeticOp,Constraint) :-
    unique_typed_id("i",integer,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(expr,ArithmeticOp,NArithmetic) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn1,[synthesis(NArithmetic,left_input)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(NArithmetic,right_input)],In2) ,
    add_texpr_infos(TempOut,[synthesis(NArithmetic,output)],Out) ,
    Node =.. [ArithmeticOp,In1,In2] ,
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).

get_fn_types(seq(Type),integer,Type).
get_fn_types(set(couple(A,B)),A,B).

% add prefix "l" to Var for connectivity constraint, add the type to the location variables
% information and adopt the given information which encodes the location variables reference,
% e.g. [synthesis(type,integer),synthesis(left_input,greater)]
get_location_var_from_ast_node(b(identifier(Name),Type,Info),b(identifier(LocationName),integer,TypedInfo)) :-
    ! , atom_concat(l,Name,LocationName) ,
    set_location_var_type(Info,Type,TypedInfo).
get_location_var_from_ast_node(b(Node,Type,Info),b(identifier(LocationName),integer,TypedInfo)) :-
    get_location_var_from_ast_node_aux(Node,Type,ValueCode) ,
    % set input/output enumeration as prefix to get unique name
    % output location names are like 'lovalue0boolean_true'
    ( member(synthesis(input(C)),Info)
    ->  number_codes(C,CCode)
    ;   member(synthesis(output(C)),Info) ,
        number_codes(C,NumberCodes) ,
        append([[111],[118,97,108,117,101],NumberCodes],CCode)) ,
    set_location_var_type(Info,Type,TypedInfo) ,
    append(CCode,ValueCode,Codes) ,
    atom_codes(LocationName,[108|Codes]).

get_location_var_from_ast_node_aux(unary_minus(b(integer(Value),integer,_)),integer,[45|ValueCode]) :-
    number_codes(Value,ValueCode).
get_location_var_from_ast_node_aux(integer(Value),integer,ValueCode) :-
    number_codes(Value,ValueCode).
get_location_var_from_ast_node_aux(_,set(_),ValueCode) :-
    atom_codes(set,ValueCode).
get_location_var_from_ast_node_aux(_,seq(_),ValueCode) :-
    atom_codes(seq,ValueCode).
get_location_var_from_ast_node_aux(Bool,boolean,ValueCode) :-
    atom_codes(Bool,ValueCode).

set_location_var_type(Info,_,Info) :-
    member(synthesis(ConstantSet,output),Info) ,
    delete_numbers_from_atom(ConstantSet,constant_set).
set_location_var_type(Info,_,[synthesis(type,integer)|Info]) :-
    member(synthesis(Operator,_),Info) ,
    delete_numbers_from_atom(Operator,Op) ,
    member(Op,[card,size,min,max]) , !.
set_location_var_type(Info,Type,[synthesis(type,Type)|Info]).

% enumerate used components for identifier nodes information so we definitely
% know which input/output locations belong together if a component is used multiple times
set_component_id(pred,Component,ComponentWithId) :-
    used_library(Predicates,Expressions) ,
    retract(used_library(_,_)) ,
    findall(Component,member((Component,_),Predicates),OpList) ,
    length(OpList,OpAmount) ,
    ComponentId is OpAmount + 1 ,
    atom_number_concat(Component,ComponentId,ComponentWithId) ,
    assert(used_library([(Component,ComponentWithId)|Predicates],Expressions)).
set_component_id(expr,Component,ComponentWithId) :-
    used_library(Predicates,Expressions) ,
    retract(used_library(_,_)) ,
    findall(Component,member((Component,_),Expressions),OpList) ,
    length(OpList,OpAmount) ,
    ComponentId is OpAmount + 1 ,
    atom_number_concat(Component,ComponentId,ComponentWithId) ,
    assert(used_library(Predicates,[(Component,ComponentWithId)|Expressions])).

% evaluate M := |I| + N
% the value of M determines the amount of lines in the program
get_line_amount((Predicates,Expressions),InputVars,M) :-
    length(InputVars,AmountIn) ,
    length(Predicates,N1) ,
    length(Expressions,N2) ,
    N is N1 + N2 ,
    M is AmountIn + N.

% set the location variables of the program parameters and output given by I/O-Examples
% each input parameters are defined in the first lines of the program
set_input_location_vars(InputValueNodes,InputLocations) :-
    set_input_location_vars(InputValueNodes,0,[],InputLocationsList) ,
    conjunct_predicates(InputLocationsList,InputLocations).
set_input_location_vars([],_,Acc,Acc).
set_input_location_vars([InputValueNode|R],C,Acc,InputLocations) :-
    get_location_var_from_ast_node(InputValueNode,InputLocationNode) ,
    InputLocation = b(equal(InputLocationNode,b(integer(C),integer,[])),pred,[]) ,
    C1 is C + 1 ,
    set_input_location_vars(R,C1,[InputLocation|Acc],InputLocations).
% we dispose output values in descending order starting at the last line of the program
set_output_location_vars(OutputValueNodes,M,OutputLocations) :-
    M1 is M - 1 ,
    set_output_location_vars(OutputValueNodes,M1,[],OutputLocationsList) ,
    conjunct_predicates(OutputLocationsList,OutputLocations).
set_output_location_vars([],_,Acc,Acc).
set_output_location_vars([OutputValueNode|R],C,Acc,OutputLocations) :-
    get_location_var_from_ast_node(OutputValueNode,OutputLocationNode) ,
    OutputLocation = b(equal(OutputLocationNode,b(integer(C),integer,[])),pred,[]) ,
    C1 is C - 1 ,
    set_output_location_vars(R,C1,[OutputLocation|Acc],OutputLocations).

% enumerate input nodes in the info to use within the connectivity constraint, the corresponding
% machine variable names have already been added to each node's info
transition_values_to_ast(InputValues,StateType,InputNodes) :-
    transition_values_to_ast(InputValues,StateType,0,InputNodes).
transition_values_to_ast([],_,_,[]).
transition_values_to_ast([Value|R],input,C,[b(Node,Type,[synthesis(input(C))|Info])|T]) :-
    value_to_ast(Value,b(Node,Type,Info)) ,
    C1 is C + 1 ,
    transition_values_to_ast(R,input,C1,T).
transition_values_to_ast([Value|R],output,C,[b(Node,Type,[synthesis(output(C))|Info])|T]) :-
    value_to_ast(Value,b(Node,Type,Info)) ,
    C1 is C + 1 ,
    transition_values_to_ast(R,output,C1,T).

inverse_operator(greater,less_equal).
inverse_operator(member,not_member).
inverse_operator(not_member,member).
inverse_operator(equal,not_equal).
inverse_operator(not_equal,equal).
inverse_operator(greater_equal,less).
inverse_operator(subset,not_subset).
inverse_operator(subset_strict,not_subset_strict).

atom_number_concat(Atom,Number,Concatenation) :-
    number_codes(Number,NumberCodes) ,
    atom_codes(Atom,AtomCodes) ,
    append(AtomCodes,NumberCodes,ConcatCodes) ,
    atom_codes(Concatenation,ConcatCodes).

% Get an output ĺocation variable for a specific location value.
get_output_location_for_location_value(LocationVars,SolutionBindings,LocationValue,OutputLocationVar) :-
    get_binding_for_name_or_value(SolutionBindings,OutputLocationName,LocationValue) ,
    atom_concat(lo,_,OutputLocationName) , ! ,
    member(OutputLocationVar,LocationVars) , OutputLocationVar = b(identifier(OutputLocationName),integer,_).
% Return a node for the location value in case there is no explicit operators' output mapped, thus, it is an input parameter.
% We know the first input parameter is in the first line of the program etc. which will not change, therefore we
% do not need the explicit location variable name but only the location value.
get_output_location_for_location_value(_,_,int(Value),b(integer(Value),integer,[])).

accumulate_used_location_var(Info,LocationVars,Acc,[b(identifier(Id),integer,IdInfo)|NewAcc]) :-
    member(synthesis(Operator,output),Info) ,
    member(b(identifier(Id),integer,IdInfo),LocationVars) ,
    member(synthesis(Operator,output),IdInfo) ,
    accumulate_used_location_var_aux(Operator,LocationVars,Acc,NewAcc).
accumulate_used_location_var(_,_,Acc,Acc).

% add input location variables since the program asts contain explicit values
accumulate_used_location_var_aux(Operator,LocationVars,Acc,NewAcc) :-
    findall(LocationVar,(member(LocationVar,LocationVars) , LocationVar = b(_,_,Info) ,
                         member(synthesis(Operator,LocationType),Info) ,
                         (LocationType = left_input ; LocationType = right_input)),InputLocationVars) ,
    append(Acc,InputLocationVars,NewAcc).

:- spec_pre(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),var,var]).
:- spec_invariant(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),list(ast),list(atom)]).
:- spec_post(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),var,var],
                                                   [ast,list(ast),list(ast),list(atom),list(ast),list(atom)]).
get_used_location_vars_and_op_names(b(Node,_,Info),LocationVars,LocationAcc,OpAcc,UsedLocationVars,UsedOperatorNames) :-
    Node =.. [_,LeftAST,RightAST] ,
    is_ast(LeftAST) , is_ast(RightAST) ,
    accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,LocationAcc1,OpAcc1) ,
    get_used_location_vars_and_op_names(LeftAST,LocationVars,LocationAcc1,OpAcc1,LocationAcc2,OpAcc2) ,
    get_used_location_vars_and_op_names(RightAST,LocationVars,LocationAcc2,OpAcc2,UsedLocationVars,UsedOperatorNames) , !.
    %findall(ParameterLocation,(member(ParameterLocation,LocationVars) ,
    %                           ParameterLocation = b(_,_,Info) , member(synthesis(input(_)),Info)),ParameterLocations) ,
    %append(TempUsedLocationVars,ParameterLocations,UsedLocationVars).
get_used_location_vars_and_op_names(b(Node,_,Info),LocationVars,LocationAcc,OpAcc,UsedLocationVars,UsedOperatorNames) :-
    Node =.. [_,AST] ,
    accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,NewLocationAcc,NewOpAcc) ,
    (is_ast(AST)
    ->  get_used_location_vars_and_op_names(AST,LocationVars,NewLocationAcc,NewOpAcc,UsedLocationVars,UsedOperatorNames)
    ;   UsedLocationVars = NewLocationAcc , UsedOperatorNames = NewOpAcc).
get_used_location_vars_and_op_names(_,_,_,_,Acc,Acc).

accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,NewLocationAcc,[Op|OpAcc]) :-
    member(synthesis(Op,output),Info) ,
    \+member(Op,OpAcc) ,
    member(b(identifier(_),integer,IdInfo),LocationVars) ,
    member(synthesis(Op,output),IdInfo) ,
    accumulate_used_location_var_aux(Op,LocationVars,LocationAcc,NewLocationAcc).
accumulate_used_location_var(_,_,LocationAcc,OpAcc,LocationAcc,OpAcc).

:- spec_pre(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),var]).
:- spec_invariant(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),list(ast)]).
:- spec_post(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),var],
                                                          [atom,list(ast),list(ast),list(ast)]).
get_input_location_nodes_for_operator_name(_,[],Acc,Acc).
get_input_location_nodes_for_operator_name(OperatorName,[LocationVar|T],Acc,InputLocations) :-
    LocationVar = b(identifier(_),integer,Info) ,
    member(synthesis(OperatorName,LocationType),Info) ,
    LocationType \= output , ! ,
    get_input_location_nodes_for_operator_name(OperatorName,T,[LocationVar|Acc],InputLocations).
get_input_location_nodes_for_operator_name(OperatorName,[_|T],Acc,InputLocations) :-
    get_input_location_nodes_for_operator_name(OperatorName,T,Acc,InputLocations).

%get_output_location_node_for_operator_name(OperatorName,[OutputLocationNode|_],OutputLocationNode) :-
%    OutputLocationNode = b(identifier(_),integer,Info) ,
%    member(synthesis(OperatorName,output),Info) , !.
%get_output_location_node_for_operator_name(OperatorName,[_|T],OutputLocationNode) :-
%    get_output_location_node_for_operator_name(OperatorName,T,OutputLocationNode).


% Exclude trivial assignments like add, minus, multiplication with 0 as well as multiplication or division with 1 (in the denominator).
% Those computations are constant values so we want the synthesis to use constants for that.
exclude_trivial_assignments([],Acc,Exclusion) :-
    conjunct_predicates(Acc,Exclusion).
exclude_trivial_assignments([InputVar|T],Acc,Exclusion) :-
    InputVar = b(_,_,Info) ,
    (member(synthesis(OperatorName,left_input),Info) ;  member(synthesis(OperatorName,right_input),Info)),
    delete_numbers_from_atom(OperatorName,Operator) ,
    exclude_trivial_assignments_aux(InputVar,Operator,SingleExclusion) ,
    exclude_trivial_assignments(T,[SingleExclusion|Acc],Exclusion).
exclude_trivial_assignments_aux(InputVar,Operator,b(not_equal(InputVar,b(integer(0),integer,[])),pred,[])) :-
    member(Operator,[add,minus]).
exclude_trivial_assignments_aux(InputVar,multiplication,SingleExclusion) :-
    SingleExclusion = b(conjunct(b(not_equal(InputVar,b(integer(1),integer,[])),pred,[]),
                                 b(not_equal(InputVar,b(integer(0),integer,[])),pred,[])),pred,[]).
exclude_trivial_assignments_aux(InputVar,div,SingleExclusion) :-
    % only restrict right input for division, i.e., the denominator
    get_texpr_info(InputVar,Info) ,
    member(synthesis(_,right_input),Info) ,
    SingleExclusion = b(not_equal(InputVar,b(integer(1),integer,[])),pred,[]).
exclude_trivial_assignments_aux(InputVar,Operator,SingleExclusion) :-
    % no empty set/seq for union, concat, set_subtraction, etc.
    member(Operator,[union,concat,set_subtraction]) ,
    SingleExclusion = b(greater(b(card(InputVar),integer,[]),b(integer(0),integer,[])),pred,[]).
exclude_trivial_assignments_aux(_,_,b(truth,pred,[])).


exclude_solution_for_further_search(UsedOperatorNames,LocationVars,SolutionBindings,Exclusion) :-
    maplist(exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings),UsedOperatorNames,ExclusionList) ,
    disjunct_predicates(ExclusionList,TempExclusion) ,
    exclude_solution_for_further_search_aux(TempExclusion,Exclusion).

exclude_solution_for_further_search_aux(b(falsity,_,_),b(truth,pred,[])) :- !.
exclude_solution_for_further_search_aux(Exclusion,Exclusion).

% constants: Inequality with the input location it is mapped to or with the explicit location value
%            in case the constant is the program output.
exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings,OperatorName,Exclusion) :-
    atom_concat(constant,_,OperatorName) , ! ,
    get_location_var_by_info_term(LocationVars,synthesis(OperatorName,output),ConstantLocation) ,
    ConstantLocation = b(identifier(ConstantLocationName),integer,_) ,
    get_binding_for_name_or_value(SolutionBindings,ConstantLocationName,LocationValue) ,
    get_binding_for_name_or_value(SolutionBindings,MappedInputLocationName,LocationValue) ,
    MappedInputLocationName \= ConstantLocationName ,
    exclude_output_to_input_mapping_for_constant_aux(LocationVars,LocationValue,ConstantLocation,MappedInputLocationName,Exclusion).
% Exclude the mapping of a given operators input locations to its corresponding output locations for the given solution.
exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings,OperatorName,Exclusion) :-
    get_input_location_nodes_for_operator_name(OperatorName,LocationVars,[],InputLocations) , InputLocations \= [] ,
    maplist(exclude_specific_input_output_mapping(LocationVars,SolutionBindings),InputLocations,InputOutputExclusions) ,
    disjunct_predicates(InputOutputExclusions,Exclusion).

exclude_output_to_input_mapping_for_constant_aux(LocationVars,_,ConstantLocation,MappedInputLocationName,Exclusion) :-
    member(MappedInputLocationVar,LocationVars) , MappedInputLocationVar = b(identifier(MappedInputLocationName),_,_) ,
    create_texpr(not_equal(ConstantLocation,MappedInputLocationVar),pred,[],Exclusion).
exclude_output_to_input_mapping_for_constant_aux(_,LocationValue,ConstantLocation,_,Exclusion) :-
    create_texpr(not_equal(ConstantLocation,b(value(LocationValue),integer,[])),pred,[],Exclusion).

get_location_var_by_info_term(LocationVars,SynthesisTerm,LocationVar) :-
    member(LocationVar,LocationVars) , LocationVar = b(identifier(_),integer,Info) ,
    member(SynthesisTerm,Info).

get_location_var_node_by_name(LocationVarName,LocationVars,LocationVarNode) :-
    member(LocationVarNode,LocationVars) ,
    LocationVarNode = b(identifier(LocationVarName),integer,_).

exclude_specific_input_output_mapping(LocationVars,SolutionBindings,b(identifier(Id),integer,Info),Exclusion) :-
    get_binding_for_name_or_value(SolutionBindings,Id,LocationValue) ,
    get_output_location_for_location_value(LocationVars,SolutionBindings,LocationValue,OutputLocationVar) ,
    create_texpr(not_equal(b(identifier(Id),integer,Info),OutputLocationVar),pred,[],Exclusion).
exclude_specific_input_output_mapping(_,_,_,b(falsity,pred,[])).

% Using a component several times, e.g. with internal names add1,add2, we don't want to
% swap both components when searching for a further solution.
get_prevent_component_swap_constraint(UsedLVars,SolutionBindings,LocationVars,Constraint) :-
    get_input_location_vars_for_outputs(UsedLVars,LocationVars,OutputInputsTuples) ,
    maplist(exclude_swapping_for_component(SolutionBindings,LocationVars),OutputInputsTuples,Exclusions) ,
    conjunct_predicates(Exclusions,Constraint).

% Maybe consider constants too?
get_input_location_vars_for_outputs(UsedLVars,LocationVars,OutputInputsTuples) :-
    % create tuples of component and an empty list to use as an accumulator
    findall((Component,[]),(member(OutLVar,UsedLVars) , get_texpr_info(OutLVar,Info) ,
                            member(synthesis(Component,output),Info)),Acc) ,
    get_input_location_vars_for_outputs_aux(LocationVars,Acc,OutputInputsTuples).
get_input_location_vars_for_outputs_aux([],Acc,Acc).
get_input_location_vars_for_outputs_aux([LocationVar|T],Acc,OutputInputsTuples) :-
    get_texpr_info(LocationVar,Info) ,
    (member(synthesis(Component,left_input),Info) ; member(synthesis(Component,right_input),Info)) ,
    member((Component,Inputs),Acc) ,
    delete(Acc,(Component,Inputs),TempAcc) ,
    NewTuple = (Component,[LocationVar|Inputs]) ,
    get_input_location_vars_for_outputs_aux(T,[NewTuple|TempAcc],OutputInputsTuples).
get_input_location_vars_for_outputs_aux([_|T],Acc,OutputInputsTuples) :-
    get_input_location_vars_for_outputs_aux(T,Acc,OutputInputsTuples).

% Given a tuple of an output location var and the corresponding component's inputs.
% We search for the output locations that are mapped to the input locations.
% Then, we search all equivalent component inputs and exclude mapping the output locations to
% the equivalent input location.
% Thus, we exclude swapping two equivalent operators when searching for a further solution for a given
% synthesized program.
exclude_swapping_for_component(SolutionBindings,LocationVars,(OutLVar,InLVars),ExclusionConstraint) :-
    maplist(get_lvar_name_and_input_type,InLVars,InLVarNames,InLVarTypes) ,
    % get the location values the inputs of OutLVar are mapped to
    maplist(get_binding_for_name_or_value(SolutionBindings),InLVarNames,InLVarValues) ,
    % create tuples of (lvar,lvalue)
    zip(InLVarNames,InLVarValues,VarValueTuples) ,
    zip(InLVarNames,InLVarTypes,VarTypeTuples) ,
    % then, get the output locations those inputs are mapped to
    maplist(get_corresponding_mapped_location(SolutionBindings),VarValueTuples,CorrespondingLocations) ,
    % and create tuples of output locations and the type of the input they are mapped to, i.e.,
    % either left_input or right_input
    findall((OutLVarName,InputType),(member((InLVar,OutLVarName),CorrespondingLocations) ,
                                     member((InLVar,InputType),VarTypeTuples)),OutLVarInTypeTuples) ,
    exclude_swapping_for_component_aux(LocationVars,(OutLVar,OutLVarInTypeTuples),ExclusionConstraint).
exclude_swapping_for_component_aux(LocationVars,(OutLVarName,OutLVarInTypeTuples),ExclusionConstraint) :-
    get_equivalent_input_location_nodes(OutLVarName,LocationVars,EquivalentInputLocations) ,
    % find pairs of output and input location that correspond (are mapped to) the same input type, i.e.
    % either left_input or right_input
    findall(Exclusion,(member((CorOutLVarName,InputType),OutLVarInTypeTuples) ,
                       get_location_var_node_by_name(CorOutLVarName,LocationVars,OutLVarNode) ,
                       member((InLVarNode,InputType),EquivalentInputLocations) ,
                       create_texpr(not_equal(OutLVarNode,InLVarNode),pred,[],Exclusion)),Exclusions) ,
    conjunct_predicates(Exclusions,ExclusionConstraint).

% Get all input locations that are equivalent to the component that LocationVarName belongs to.
% Don't consider exactly the same location var.
get_equivalent_input_location_nodes(LocationVarName,LocationVars,InputLocations) :-
    delete_numbers_from_atom(LocationVarName,RawName) ,
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,LocationVars,[],InputLocations).
get_equivalent_input_location_nodes_aux(_,_,[],Acc,Acc).
get_equivalent_input_location_nodes_aux(LocationVarName,RawName,[LocationVar|T],Acc,InputLocations) :-
    get_texpr_info(LocationVar,Info) ,
    member(synthesis(LVarName,InputType),Info) ,
    (InputType = left_input ; InputType = right_input) ,
    delete_numbers_from_atom(LVarName,RawLVarName) ,
    RawLVarName = RawName ,
    % not exactly the same location
    LVarName \= LocationVarName ,
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,T,[(LocationVar,InputType)|Acc],InputLocations).
get_equivalent_input_location_nodes_aux(LocationVarName,RawName,[_|T],Acc,InputLocations) :-
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,T,Acc,InputLocations).


% Given a tuple of location var name and mapped value, we search for the other location var that
% is mapped to this location value (line of the program).
get_corresponding_mapped_location(SolutionBindings,(LVarName,LVarValue),(LVarName,LVarName2)) :-
    get_binding_for_name_or_value(SolutionBindings,LVarName2,LVarValue) ,
    delete_numbers_from_atom(LVarName2,RawName) ,
    RawName \= constant ,
    LVarName \= LVarName2.

get_lvar_name_and_input_type(b(identifier(LVarName),integer,Info),LVarName,InputType) :-
    get_input_type_from_info(Info,InputType).

get_input_type_from_info(Info,left_input) :-
    member(synthesis(_,left_input),Info).
get_input_type_from_info(Info,right_input) :-
    member(synthesis(_,right_input),Info).
