% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(system_call,[system_call/4, system_call/5,
                       system_call_keep_open/7, system_call_keep_open_no_pipes/4,
                       get_writable_compiled_filename/3,
                       get_writable_compiled_filename/4, generate_temporary_cspm_file/2,
                       get_temporary_filename/2,
                       get_command_path/2]).
:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'Tools for calling underlying system procedures.').

:- use_module(library(process)).
:- use_module(library(file_systems)).
:- use_module(error_manager).

:- use_module(tools,[ajoin/2]).
:- use_module(tools_meta,[safe_on_exception/3]).

system_call_keep_open(Command,Options,Process,STDIn,STDOut,STDErr,Env) :-
    safe_on_exception(E,
          process:process_create(Command, Options,
		       [process(Process),stdout(pipe(STDOut)),stdin(pipe(STDIn)),stderr(pipe(STDErr)),environment(Env)]),
		  ( ajoin(['Could not execute command "',Command,'" due to exception: '],Msg),
		    add_error(system_call,Msg,E),fail)).

system_call_keep_open_no_pipes(Command,Options,Process,Env) :-
    safe_on_exception(E,
          process:process_create(Command, Options,
		       [process(Process),environment(Env)]),
		  ( ajoin(['Could not execute command "',Command,'" due to exception: '],Msg),
		    add_error(system_call,Msg,E),fail)).

system_call(Command,Options,ErrorTextAsCodeList,ExitCode) :-
    safe_on_exception(E,
          process:process_create(Command, Options,
		       [process(Process),stderr(pipe(JStderr))]),
		  (ajoin(['Could not execute command "',Command,'" due to exception: '],Msg),
		   add_error(system_call,Msg,E),fail)),

	read_all(JStderr,Command,stderr,ErrorTextAsCodeList),
	process_wait(Process,ExitCode).

system_call(Command,Options,OutputTextAsCodeList,ErrorTextAsCodeList,ExitCode) :-
    %StreamOptions = [encoding('UTF-8')], % we could use pipe(JStdout,StreamOptions), ... below
    safe_on_exception(E,
          process:process_create(Command, Options,
		       [process(Process),stdout(pipe(JStdout)),stderr(pipe(JStderr))]),
		  (ajoin(['Could not execute command "',Command,'" due to exception: '],Msg),
		   add_error(system_call,Msg,E),fail)),
        read_all(JStdout,Command,stdout,OutputTextAsCodeList),
	read_all(JStderr,Command,stderr,ErrorTextAsCodeList),
	process_wait(Process,ExitCode).

get_command_path(CmdName,CmdPath) :-
	absolute_file_name(path(CmdName),
				CmdPath,
				[access(exist),extensions(['.exe','']),solutions(all),file_errors(fail)]),!.
get_command_path(CmdName,_) :-
    add_error_fail(get_command_path,'Could not get path to command: ',CmdName).


:- use_module(library(lists)).

% read all characters from a stream
read_all(S,Command,Pipe,Text) :-
   on_exception(error(_,E), read_all2(S,Lines),
        (ajoin(['Error reading ',Pipe,' for "',Command,'" due to exception: '],Msg),
          add_error(system_call,Msg,E),fail)),  % E could be system_error('SPIO_E_ENCODING_INVALID')
	append(Lines,Text).
read_all2(S,Text) :-
	read_line(S,Line),
	( Line==end_of_file ->
	    Text=[""],close(S)
	; otherwise ->
	    Text = [Line, "\n" | Rest],
	    read_all2(S,Rest)).

:- use_module(tools,[get_tail_filename/2]).
generate_temporary_cspm_file(CSPFile,CSPFileNameTemp):-
   get_writable_compiled_filename(CSPFile,'.tmp',CSPFileNameTemp).

:- use_module(debug,[debug_println/2]).
get_writable_compiled_filename(SourceFile,Extension,WritableFile) :-
  get_writable_compiled_filename(SourceFile,Extension,WritableFile,_).
get_writable_compiled_filename(SourceFile,Extension,WritableFile,IsTemporaryFile) :-
   atom_concat(SourceFile,Extension,WFile),
   debug_println(9,generating_writable_file(SourceFile,Extension)),
   ((file_exists(WFile),file_property(WFile,writable))
     ->  debug_println(9,file_writable(WFile)), IsTemporaryFile=false,
         WritableFile=WFile
     ;   get_tail_filename(WFile,TFile),
         safe_on_exception(_E,
             (open(WFile,write,S1,[]),close(S1),WritableFile=WFile,
              IsTemporaryFile=false),
             (IsTemporaryFile=true,
              get_temporary_filename(TFile,WritableFile)) )
   ).

get_temporary_filename(PreferredFileName,WritableFile) :-
          debug_println(9,trying_to_get_temporary_filename(PreferredFileName)),
          open(temp(PreferredFileName),write,S1,[if_exists(generate_unique_name)]),
          stream_property(S1,file_name(WritableFile)),
          debug_println(9,temporary_file(WritableFile)),
          close(S1).
