% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(tools_meta,[safe_time_out/3,
                      safe_time_out_or_virtual_time_out/3,
                      call_residue/2,
                      safe_on_exception/3, safe_on_exception_silent/3,
                      reraise_important_exception/1,
                      det_call_cleanup/2,
                      frozen_member/2,
                      catch_matching/3,
                      safe_numbervars/3,
                      translate_term_into_atom/2, translate_term_into_atom_with_max_depth/2
     ]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'A utility on timeouts safe_time_out seperated out from tools.pl to avoid cyclic module dependencies.').

:- use_module(library(timeout),[time_out/3]).
:- meta_predicate safe_time_out(0,*,*).
safe_time_out(Call,TO,Res) :-
   (TO >= 2147483646
     -> (TO = 2147483646 -> true % special value to turn time_out off; it has an overhead
            ; print('### Warning: TIME_OUT value too high (>2147483646): '), print(TO),nl,
              print('### Calling goal without TIME_OUT (use 2147483646 to turn TIME_OUT off silently)'),nl),
         call(Call), Res=success
    ; TO < 1 -> print('### Warning: TIME_OUT value too small: '), print(TO),nl,
                time_out(Call,1,Res)
    ; time_out(Call,TO,Res)).

:- meta_predicate safe_time_out_or_virtual_time_out(0,*,*).
% catches virtual-time_out exceptions and returns them as normal time_out result
safe_time_out_or_virtual_time_out(Call,TO,Res) :-
   on_exception(enumeration_warning(_,_,_,_,_),safe_time_out(Call,TO,Res),Res=time_out).

/* from File: sp4_compatibility_mappings.pl */
/* Created: 08/05/2007 by Michael Leuschel */

:- meta_predicate call_residue(0,*).

call_residue(X,Residue) :- call_residue_vars(X,V),filter_residue_vars(V,Residue).


filter_residue_vars([],[]).
filter_residue_vars([H|T],Res) :-
  frozen(H,FH),
  (FH=true -> Res=RT ; Res = [FH|RT]),
  filter_residue_vars(T,RT).

% --------------------------

:- meta_predicate safe_on_exception(*,0,0).
% use if you want to catch any exception; ensures time_out not treated and passed on
safe_on_exception(E,Call,ExcCode) :-
  on_exception(E,call(Call), (print(exception(E)),nl,
                        reraise_important_exception(E),ExcCode)).
:- meta_predicate safe_on_exception_silent(*,0,0).
safe_on_exception_silent(E,Call,ExcCode) :-
  on_exception(E,call(Call), (reraise_important_exception(E),ExcCode)).

% reraise important exceptions
reraise_important_exception(time_out) :- !, raise_exception(time_out).
reraise_important_exception(_).

% --------------------------


:- meta_predicate det_call_cleanup(0,0).
% a simplified version of call_cleanup; only really works for deterministic predicates
% or if it is ok to call CleanUpCall multiple times on success of Call
% is much faster, as call_cleanup has an overhead of about 300 Prolog instructions
% but this seems even slower ?? add_transitions__with_timeout_fail_loop

det_call_cleanup(Call,CleanUpCall) :-
    on_exception(E,
       if(Call,CleanUpCall,CleanUpCall),
       (CleanUpCall,throw(E))).

% --------------------------

%:- meta_predicate frozen_member(*,0). % without meta_predicate we can call frozen_member with a variable for Goal
% check if Goal is attached as a pending co-routine to Var
frozen_member(Var,Goal) :- var(Var), frozen(Var,Frozen),
    frozen_mem_aux(Frozen,Goal).

frozen_mem_aux((A,B),Goal) :- !, (frozen_mem_aux(A,Goal) ; frozen_mem_aux(B,Goal)).
frozen_mem_aux(Goal,Goal).


% --------------------------

% like catch/3, but it does not fail if an exception occurs that is not
% unifiable with the second argument. Instead it re-throws the original
% exception.
:- meta_predicate catch_matching(0,*,0).
catch_matching(Call,Exception,Handler) :-
    catch(Call, E, (E=Exception -> Handler ; throw(E))).

% --------------------------

safe_numbervars(Term,Start,End) :-
  on_exception(E,numbervars(Term,Start,End),
               (print('Exception during numbervars: '),print(E),nl,nl,reraise_important_exception(E))).

% --------------------------


:- use_module(library(codesio), [write_term_to_codes/3]).
translate_term_into_atom(CTerm,Atom) :-
           copy_term(CTerm,Term), safe_numbervars(Term,0,_),
           write_term_to_codes(Term,Temp,[quoted(true),numbervars(true)]),
                       name(Atom,Temp).
translate_term_into_atom_with_max_depth(Term,Atom) :- atomic(Term),!,Atom=Term.
translate_term_into_atom_with_max_depth(Term,Atom) :-
                %copy_term(CTerm,Term),safe_numbervars(Term,0,_),
                %write_term_to_codes(Term,Temp,[quoted(true),numbervars(true),max_depth(20)]),
                write_term_to_codes(Term,Temp,[quoted(true),numbervars(true),max_depth(20)]),
                atom_codes(Atom,Temp).
