% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(typing_tools,[
     is_infinite_type/1, % check if a type is infinite
     valid_ground_type/1, % check if a type is a valid ground type
     contains_infinite_type/1,
     any_value_for_type/2,
     normalize_type/2
]).

%:- use_module(bsyntaxtree, [get_texpr_types/2, syntaxtraversion/6,  find_typed_identifier_uses/3]).
:- use_module(library(lists), [maplist/3]).

% check if a type is infinite for ProB animation; deferred sets are not counted as infinite here!
is_infinite_type(integer).
is_infinite_type(string).
is_infinite_type(freetype(_)).
is_infinite_type(set(X)) :- is_infinite_type(X).
is_infinite_type(seq(_)).
is_infinite_type(couple(X,Y)) :- (is_infinite_type(X) -> true ; is_infinite_type(Y)).
is_infinite_type(record(Fields)) :-
   ((member(field(_,Type),Fields), is_infinite_type(Type)) -> true).

contains_infinite_type([H|T]) :-
  (is_infinite_type(H) -> true ; contains_infinite_type(T)).

:- use_module(error_manager).

valid_ground_type(X) :- var(X),!, add_error(valid_ground_type,'Variable as type: ',X),fail.
valid_ground_type(X) :- valid_ground_type_aux(X),!.
valid_ground_type(X) :- add_error(valid_ground_type,'Illegal type term: ',X),fail.

%:- use_module(b_global_sets,[b_global_set/1]).

valid_ground_type_aux(any) :- format(user_error,'! Type contains *any*~n',[]).
valid_ground_type_aux(integer).
valid_ground_type_aux(string).
valid_ground_type_aux(boolean).
valid_ground_type_aux(global(G)) :- atom(G).
 %bmachine:b_get_machine_set(G).
 %b_global_sets:b_global_set(G).
valid_ground_type_aux(freetype(FT)) :- ground(FT).
valid_ground_type_aux(set(X)) :- valid_ground_type(X).
valid_ground_type_aux(seq(X)) :- valid_ground_type(X).
valid_ground_type_aux(couple(X,Y)) :- valid_ground_type(X), valid_ground_type(Y).
valid_ground_type_aux(record(Fields)) :- valid_fields(Fields).

valid_fields([]).
valid_fields([field(Name,_)|_]) :- \+ atom(Name),!,
   add_error(valid_ground_type,'Illegal record field name: ',Name),fail.
valid_fields([field(_Name,Type)|T]) :-
   valid_ground_type(Type),
   valid_fields(T).


% --------------

:- use_module(kernel_freetypes,[get_freeval_type/3]).
:- use_module(b_global_sets,[b_get_fd_type_bounds/3]).
any_value_for_type(T,V) :- if(inst2(T,V),true, (nl,print('*** instantiate_to_any_value failed: '), print(T:V),nl,nl)).
%inst2(_,V) :- ground(V),!.
%inst2(Type,V) :- print(inst2(Type,V)),nl,trace,fail.
inst2(boolean,V) :- !, inst3(V,pred_true).
inst2(pred,V) :- !, inst3(V,pred_true).
inst2(integer,V) :- !, V=int(X),inst3(X,0).
inst2(set(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(seq(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(couple(TA,TB),V) :- !, V=(A,B), inst2(TA,A), inst2(TB,B).
inst2(string,V) :- !, V=string(X), inst3(X,'').
inst2(global(T),V) :- !, V=fd(X,T), %(T=='MODES' -> trace ; true),
  b_global_sets:b_get_fd_type_bounds(T,LowBnd,_UpBnd),
  inst3(X,LowBnd).
inst2(record(_F),_V) :- !, true. % TO DO
inst2(constant([A]),V) :- !, V=term(A).
inst2(freetype(FT),V) :- !, V=freeval(FT,CASE,Val),
  (nonvar(CASE) -> (get_freeval_type(FT,CASE,Type) -> any_value_for_type(Type,Val)) % ensure no backtracking
    ; nonvar(Val) -> print((nonvar_value(V))),nl, get_freeval_type(FT,CASE,Type), any_value_for_type(Type,Val)
    ; (get_freeval_type(FT,C,Type) -> CASE=C, any_value_for_type(Type,Val))
  ).
inst2(X,V) :- print(cannot_inst2(X,V)),nl.

inst3(X,V) :- (var(X) -> X=V ; true).

% ----------------------

% replace seq(X) by set(couple(integer,X))

normalize_type(X,Y) :-
  (normalize_type_aux(X,Y) -> true ; add_internal_error('Illegal type: ',normalize_type(X,Y)),Y=X).

normalize_type_aux(V,R) :- var(V),!,R=V.
normalize_type_aux(any,any).
normalize_type_aux(pred,pred).
normalize_type_aux(subst,subst).
normalize_type_aux(boolean,boolean).
normalize_type_aux(integer,integer).
normalize_type_aux(string,string).
normalize_type_aux(global(F),global(F)).
normalize_type_aux(freetype(F),freetype(F)).
normalize_type_aux(set(X),set(N)) :- normalize_type_aux(X,N).
normalize_type_aux(seq(X),set(couple(integer,N))) :- normalize_type_aux(X,N).
normalize_type_aux(couple(X,Y),couple(NX,NY)) :- normalize_type_aux(X,NX),normalize_type_aux(Y,NY).
normalize_type_aux(record(F),record(NF)) :- normalize_type_fields(F,NF).
normalize_type_aux(op(A,B),op(NA,NB)) :- maplist(normalize_type_aux,A,NA), maplist(normalize_type_aux,B,NB).

normalize_type_fields([],[]).
normalize_type_fields([field(N,T)|R],[field(N,NT)|RN]) :- normalize_type_aux(T,NT),normalize_type_fields(R,RN).
